# Copyright (c) 2016 Boulder Real Time Technologies, Inc.
# All rights reserved 

# This software module is wholly owned by Boulder Real Time 
# Technologies, Inc. Any use of this software module without
# express written permission from Boulder Real Time Technologies,
# Inc. is prohibited.

# the unit test needs to be run with bqpy

if __name__ == '__main__':
    import unittest
    import os
    import sys
    sys.path.append(os.environ['ANTELOPE'] + "/data/python")

    import signal

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    from antelope import _bqplot
    from antelope import stock
    from antelope.buvector import *

    class PfSubs(object):
        def __init__ (self):
                return

        @classmethod
        def pfstring2py (self, pfstring, pyd=None):
                pf = stock.ParameterFile()
                pf.pfcompile (pfstring)
                pyf = pf.pf2dict()
                if pyd != None:
                        pyd.update(pyf)
                else:
                        pyd = pyf
                return (pyd)

        @classmethod
        def py2pfstring (self, pyd):
                pf = stock.ParameterFile()
                pf['foo'] = pyd
                pfstr = pf.pf2string()
                pfstr = pfstr[10:-2]
                return pfstr

    class Mycallback(object):
            instance = -1
            instances = {}

            def __init__ (self, vp, pl, vecpl):
                    Mycallback.instance += 1
                    Mycallback.instances[Mycallback.instance] = self
                    self.vp = vp
                    self.pl = pl
                    self.vecpl = vecpl

            @classmethod
            def mycallback (self, pfstring):
                    argsd = PfSubs.pfstring2py (pfstring)
                    instance = int(argsd['instance'])
                    obj = Mycallback.instances[instance]

                    if argsd['bqglyph']['state'] == 'pan':
                            if argsd['bqevent']['type'] == 'mousemove':
                                    (x, y) = (float(argsd['bqevent']['pt'][0]), float(argsd['bqevent']['pt'][1]))
                                    (xw, yw) = _bqplot._bqplot_getwcoords (obj.vp, x, y)
                                    index = int(argsd['bqglyph']['index'])
                                    buvector_put (obj.vecpl, index, xw, yw)
                                    if index == 0:
                                            buvector_put (obj.vecpl, 4, xw, yw)
                                    _bqplot._bqplot_configure (obj.pl, "vector", obj.vecpl)
                                    _bqplot._bqplot_setitemstale (obj.pl)

                    return 'ok'

    class Testbqplot_fixture(object):

        tempdir = '/tmp/python_bqplottest_' + str(os.getuid()) + str(os.getpid())

        def start(self):

            os.mkdir(self.tempdir)
            os.chdir(self.tempdir)

        def stop(self):

            os.chdir("/")
            os.rmdir(self.tempdir)

    class Testbqplot(unittest.TestCase):

        def test_procedure_bqplot_create(self):

            vp = _bqplot._bqplot_viewport ("toplevel")
            self.assertTrue(vp != None)
            print(vp)

            _bqplot._bqplot_configure (vp, \
                    "xleft",         -1.0, \
                    "xright",        5.0, \
                    "ybottom",        -1.1, \
                    "ytop",                1.1, \
                    "mleft",        80, \
                    "mright",        20, \
                    "mbottom",        50, \
                    "mtop",                30, \
                    "fill_frame",         "lightblue", \
                    "fill",                "#e0e0e0" )

            axes = _bqplot._bqplot_axes (vp)
            self.assertTrue(axes != None)
            print(axes)

            _bqplot._bqplot_configure (axes, \
                    "xlabel",                 "My X-stuff", \
                    "ylabel",                 "My Y-stuff", \
                    "xformat",                "%.1f", \
                    "yformat",                "%.1f", \
                    "linewidth_box",        0, \
                    "linewidth_tics",        2, \
                    "linewidth_tics_small",        0, \
                    "axis_style",                "sw", \
                    "tic_style",                "siwinoeo", \
                    "linewidth_grids",        2, \
                    "linewidth_grids_small",0, \
                    "color_grids",                "black", \
                    "color_grids_small",        "#c0c0c0" )

            text = _bqplot._bqplot_text (vp, "My text", "0.5v", "top+10")
            self.assertTrue(text != None)
            print(text)

            _bqplot._bqplot_configure (text, \
                    "color_text",        "red", \
                    "font",                "Times 14 Bold", \
                    "justification","s", \
                    "angle",        0 )
            
            polyline = _bqplot._bqplot_polyline (vp)
            self.assertTrue(polyline != None)
            print(polyline)
            
            _bqplot._bqplot_configure (polyline, \
                    "test",                "on", \
                    "color_outline","#400000ff", \
                    "linewidth",        "3" )
            
            _bqplot._bqplot_setlayer (vp, "layer1")
            _bqplot._bqplot_setlayer (axes, "layer1")
            _bqplot._bqplot_setlayer (text, "layer1")
            _bqplot._bqplot_setlayer (polyline, "layer1")
            
            vecpl = buvector_create ()
            buvector_append (vecpl, -1, 0.1, 0.6)
            buvector_append (vecpl, -1, 1.9, 0.4)
            buvector_append (vecpl, -1, 0.3, 0.9)
            buvector_append (vecpl, -1, 0.4, 0.8)
            buvector_append (vecpl, -1, 0.1, 0.6)
            
            polyline2 = _bqplot._bqplot_polyline (vp)
            self.assertTrue(polyline2 != None)
            print(polyline2)
            
            _bqplot._bqplot_configure (polyline2, \
                    "vector",         vecpl, \
                    "color_fill",         "#80ffff00", \
                    "color_outline","red", \
                    "linewidth",        2 )
            
            
            vecpp = buvector_create ()
            buvector_append (vecpp, -1, 0.1, 0.6, "Apt {o=black f=white n=red sz=30 sy=bb,1,2,3,4,5,6}")
            buvector_append (vecpp, -1, 1.9, 0.4, "Bpt {f=#400000b0 sz=30 sy=square}")
            buvector_append (vecpp, -1, 0.3, 0.9, "Cpt {f=white o=#00ff00 sz=20 sy=star}")
            buvector_append (vecpp, -1, 0.4, 0.8, "Dpt {f=black o=orange sz=15}")
            
            polypoint = _bqplot._bqplot_polypoint (vp)
            self.assertTrue(polypoint != None)
            print(polypoint)

            mycallback = Mycallback (vp, polyline2, vecpl)
            _bqplot._bqplot_register_callback (polypoint, "Mycallback.mycallback", '%d' % mycallback.instance)
            
            _bqplot._bqplot_configure (polypoint, \
                    "vector",        vecpp, \
                    "symbol",        "triangle",  \
                    "color_fill",        "white", \
                    "color_outline","green", \
                    "font",                "Helvetica 12 Bold", \
                    "color_text",        "brown", \
                    "linewidth",        2, \
                    "showtext",        1, \
                    "size",                50, \
                    "interaction",        "move" )
            
            _bqplot._bqplot_setlayer (polyline2, "layer2")
            _bqplot._bqplot_setlayer (polypoint, "layer2")

            _bqplot._bqplot_geometry (800, 800, 0)
            _bqplot._bqplot_show ()
            _bqplot._bqplot_qtmainloop ()

    server = Testbqplot_fixture()
    server.start()
    suite = unittest.makeSuite(Testbqplot)
    runner = unittest.TextTestRunner()
    runner.run(suite)
    server.stop()

    _bqplot._bqplot_pymainloop ()


