# Copyright (c) 2016 Boulder Real Time Technologies, Inc.
# All rights reserved 

# This software module is wholly owned by Boulder Real Time 
# Technologies, Inc. Any use of this software module without
# express written permission from Boulder Real Time Technologies,
# Inc. is prohibited.

import os
import sys
sys.path.append(os.environ['ANTELOPE'] + "/data/python")

from antelope import _bqplot
from antelope import datascope
from antelope import _datascope
from antelope.pfsubs import *
from antelope.ev import *

def bqplot_frame(parent = "toplevel"):
    """Make a new Qt frame widget"""

    return _bqplot._bqplot_frame(parent)

def bqplot_gridlayout(parent):
    """Make a new Qt grid layout"""

    return _bqplot._bqplot_gridlayout(parent)

def bqplot_gridlayout_addwidget(layout, widget, fromrow, fromcolumn, rowspan, columnspan):
    """Add a widget to a Qt grid layout"""

    return _bqplot._bqplot_gridlayout_addwidget(layout, widget, fromrow, fromcolumn, rowspan, columnspan)

def bqplot_gridlayout_columnstretch(layout, column, stretch):
    """Set a Qt grid layout column stretch factor"""

    return _bqplot._bqplot_gridlayout_columnstretch(layout, column, stretch)

def bqplot_gridlayout_rowstretch(layout, row, stretch):
    """Set a Qt grid layout row stretch factor"""

    return _bqplot._bqplot_gridlayout_rowstretch(layout, row, stretch)

def bqplot_splitter(parent, orientation='horizontal'):
    """Make a new Qt splitter frame"""

    return _bqplot._bqplot_splitter(parent, orientation)

def bqplot_splitter_addwidget(splitter, widget, index):
    """Add a widget to a Qt splitter frame"""

    return _bqplot._bqplot_splitter_addwidget(splitter, widget, index)

def bqplot_gridlayout_addwidget(layout, widget, fromrow, fromcolumn, rowspan, columnspan):
    """Add a widget to a Qt grid layout"""

    return _bqplot._bqplot_gridlayout_addwidget(layout, widget, fromrow, fromcolumn, rowspan, columnspan)

def bqplot_button(parent):
    """Make a new Qt push button"""

    return _bqplot._bqplot_button(parent)

def bqplot_button_text(button):
    """Set Qt push button text"""

    return _bqplot._bqplot_button_text(button)

def bqplot_viewport(parent = "toplevel"):
    """Make a new bqplot viewport item"""

    return _bqplot._bqplot_viewport(parent)

def bqplot_getwcoords(master, x, y):
    """Convert a bqplot viewport item pixel coordinates to world coordinates"""

    return _bqplot._bqplot_getwcoords(master, x, y)

def bqplot_axes(master):
    """Make a new bqplot axes item"""

    return _bqplot._bqplot_axes(master)

def bqplot_text(master, text, x, y):
    """Make a new bqplot text item"""

    return _bqplot._bqplot_text(master, text, x, y)

def bqplot_polyline(master):
    """Make a new bqplot polyline item"""

    return _bqplot._bqplot_polyline(master)

def bqplot_polypoint(master):
    """Make a new bqplot polypoint item"""

    return _bqplot._bqplot_polypoint(master)

def bqplot_map(master):
    """Make a new bqplot map item"""

    return _bqplot._bqplot_map(master)

def bqplot_mapevents(master):
    """Make a new bqplot mapevents item"""

    return _bqplot._bqplot_mapevents(master)

def bqplot_textpopup(master):
    """Make a new bqplot textpopup item"""

    return _bqplot._bqplot_textpopup(master)

def bqplot_configure(master, *args):
    """Configure a bqplot item"""

    return _bqplot._bqplot_configure(master, *args)

def bqplot_setlayer(item, level):
    """Set layer for a bqplot item"""

    return _bqplot._bqplot_setlayer(item, level)

def bqplot_itemshow(item):
    """Show a bqplot item"""

    return _bqplot._bqplot_itemshow(item)

def bqplot_itemhide(item):
    """Hide a bqplot item"""

    return _bqplot._bqplot_itemhide(item)

def bqplot_getviewport(item):
    """Get parent viewport for a bqplot item"""

    return _bqplot._bqplot_getviewport(item)

def bqplot_setitemstale(item):
    """Set a bqplot item stale flag"""

    return _bqplot._bqplot_setitemstale(item)

def bqplot_register_callback(item, proc, client_string='0', proc2=None, client_string2='0'):
    """Register callback for a bqplot item"""

    if proc2 == None:
        return _bqplot._bqplot_register_callback(item, proc, client_string)
    else:
        return _bqplot._bqplot_register_callback(item, proc, client_string, proc2, client_string2)

def bqplot_map_setdefaults(map):
    """Set default parameters for a bqplot map item"""

    return _bqplot._bqplot_map_setdefaults(map)

def bqplot_map_gettaskbar(map):
    """Get taskbar from a bqplot map item"""

    return _bqplot._bqplot_map_gettaskbar(map)

def bqplot_mapevents_setdefaults(mapevents):
    """Set default parameters for a bqplot mapevents item"""

    return _bqplot._bqplot_mapevents_setdefaults(mapevents)

def bqplot_mapevents_setdefaults_nort(mapevents):
    """Set default parameters for a bqplot mapevents item with no rt processing"""

    return _bqplot._bqplot_mapevents_setdefaults_nort(mapevents)

def bqplot_mapevents_seteventview(mapevents, ev):
    """Set eventview for a bqplot mapevents item"""

    return _bqplot._bqplot_mapevents_seteventview(mapevents, ev)

def bqplot_mapevents_refresheventview(mapevents):
    """Refresh eventview for a bqplot mapevents item"""

    return _bqplot._bqplot_mapevents_refresheventview(mapevents)

def bqplot_mapevents_getmap(mapevents):
    """Get map from a bqplot mapevents item"""

    return _bqplot._bqplot_mapevents_getmap(mapevents)

def bqplot_mapevents_gettaskbar(mapevents):
    """Get taskbar from a bqplot mapevents item"""

    return _bqplot._bqplot_mapevents_gettaskbar(mapevents)

def bqplot_geometry(*args):
    """Set geometry of bqplot toplevel window"""

    return _bqplot._bqplot_geometry(*args)

def bqplot_getscreen(screen):
    """Get screen size of display"""

    return _bqplot._bqplot_getscreen(screen)

def bqplot_show(handle="toplevel"):
    """Show Qt widget"""

    return _bqplot._bqplot_show(handle)

def bqplot_update(item="toplevel"):
    """Update bqplot item"""

    return _bqplot._bqplot_update(item)

def bqplot_qtmainloop():
    """Start QT event processing for a bqplot toplevel window"""

    return _bqplot._bqplot_qtmainloop()

def bqplot_pymainloop():
    """Enter python main loop"""

    return _bqplot._bqplot_pymainloop()

def bqplot_quit():
    """Send quit to QT server"""

    return _bqplot._bqplot_quit()

def bqplot_setstylesheet(parent, stylesheet):
    """Set Qt stylesheet"""

    return _bqplot._bqplot_setstylesheet(parent, stylesheet)

class Bqplot (object):

    @classmethod
    def check_color (self, color):
        ret = _bqplot._bqplot_check_color (color)
        if ret == 1:
            return 1
        return 0

class ViewportItem (object):
    instances = {}

    def __init__ (self, item, handle):
        self.item_object = item
        self.item_handle = handle
        self.layer = None
        ViewportItem.instances[handle] = item

    def __del__ (self):
        _bqplot._bqplot_delete (self.item_handle)

    def gethandle (self):
        return self.item_handle

    def getitem (self, handle):
        return ViewportItem.instances[handle]

    def configure (self, *args, **kwargs):
        if args != None:
            if len(args) == 1:
                return _bqplot._bqplot_configure_pfstring(self.item_handle, args[0])
            outargs = list(args)
        else :
            outargs = list()
        if kwargs != None:
            for key in list(kwargs.keys()):
                outargs.append (key)
                outargs.append (kwargs[key])
        if len(outargs) < 1:
            return
        return _bqplot._bqplot_configure(self.item_handle, *outargs)

    def setlayer (self, layer):
        self.layer = _bqplot._bqplot_setlayer(self.item_handle, layer)
        return self.layer

    def setlayervisibility (self, visibility=1):
        if self.layer == None:
            raise Exception ('ViewportItem.setlayervisibility: no layer set for this item')
            return
        return _bqplot._bqplot_setlayervisibility(self.layer, visibility)

    def totop (self):
        return _bqplot._bqplot_totop(self.item_handle)

    def tobottom (self):
        return _bqplot._bqplot_tobottom(self.item_handle)

    def layertotop (self):
        if self.layer == None:
            raise Exception ('ViewportItem.setlayertotop: no layer set for this item')
            return
        return _bqplot._bqplot_layertotop(self.layer)

    def layertobottom (self):
        if self.layer == None:
            raise Exception ('ViewportItem.setlayertobottom: no layer set for this item')
            return
        return _bqplot._bqplot_layertobottom(self.layer)

    def itemshow (self):
        return _bqplot._bqplot_itemshow(self.item_handle)

    def itemhide (self):
        return _bqplot._bqplot_itemhide(self.item_handle)

    def setitemstale (self):
        return _bqplot._bqplot_setitemstale(self.item_handle)

    def setitemname (self, name):
        return _bqplot._bqplot_itemsetname(self.item_handle, name)

    def update (self):
        return _bqplot._bqplot_update(self.item_handle)

    def getviewport (self):
        vphandle = _bqplot._bqplot_getviewport(self.item_handle)
        if not vphandle in ViewportItem.instances:
            ViewportItem.instances[vphandle] = Viewport (None, vphandle)
        return ViewportItem.instances[vphandle]

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.item_handle, proc, client_string)

class QtBasic (object):
    instances = {}

    def __init__ (self, parent, handle=None):
        if parent == None or isinstance (parent, GraphicsEngine):
            parent = 'none'
        self.parent = parent
        self.handle = handle

    def setinstance (self, handle):
        Widget.instances[handle] = self

    def gethandle (self):
        return self.handle

    def delete (self):
        return _bqplot._bqplot_delete(self.handle)

    def update (self):
        return _bqplot._bqplot_update(self.handle)

    def getgeometry (self):
        return _bqplot._bqplot_getgeometry(self.handle)

class Widget (QtBasic):
    def __init__ (self, parent, handle=None):
        super (Widget, self).__init__ (parent, handle)

    def delete (self):
        _bqplot._bqplot_delete (self.handle)

    def show (self):
        return _bqplot._bqplot_show(self.handle)

    def hide (self):
        return _bqplot._bqplot_hide(self.handle)

    def totop (self):
        return _bqplot._bqplot_totop(self.handle)

    def tobottom (self):
        return _bqplot._bqplot_tobottom(self.handle)

    def maximumsize (self, w, h):
        return _bqplot._bqplot_maximumsize(self.handle, w, h)

    def maximumwidth (self, w):
        return _bqplot._bqplot_maximumsize(self.handle, w, -1)

    def maximumheight (self, h):
        return _bqplot._bqplot_maximumsize(self.handle, -1, h)

    def minimumsize (self, w, h):
        return _bqplot._bqplot_minimumsize(self.handle, w, h)

    def minimumwidth (self, w):
        return _bqplot._bqplot_minimumsize(self.handle, w, -1)

    def minimumheight (self, h):
        return _bqplot._bqplot_minimumsize(self.handle, -1, h)

class ConfigWidget (Widget):
    def __init__ (self, parent, handle=None):
        super (ConfigWidget, self).__init__ (parent, handle)

    def configure (self, *args, **kwargs):
        if args != None:
            outargs = list(args)
        else :
            outargs = list()
        if kwargs != None:
            for key in kwargs.keys():
                outargs.append (key)
                outargs.append (kwargs[key])
        if len(outargs) < 1:
            return
        return _bqplot._bqplot_configure(self.handle, *outargs)

class MainWindow (Widget):
    def __init__ (self, parent, close_window=None, handle=None):
        super (MainWindow, self).__init__ (parent, handle)
        if self.handle == None:
            if close_window == None:
                if self.parent == 'none':
                    self.handle = _bqplot._bqplot_mainwindow('none')
                else:
                    self.handle = _bqplot._bqplot_mainwindow(self.parent.gethandle())
            else:
                if self.parent == 'none':
                    self.handle = _bqplot._bqplot_mainwindow('none', close_window)
                else:
                    self.handle = _bqplot._bqplot_mainwindow(self.parent.gethandle(), close_window)
        super (MainWindow, self).setinstance (self.handle)

    def getframe (self):
        frame_handle = _bqplot._bqplot_mainwindow_getframe(self.handle)
        return Frame (self, None, frame_handle)

    def geometry (self, *args):
        if len(args) == 2:
            return _bqplot._bqplot_geometry(self.handle, args[0], args[1], 0)
        elif len(args) == 3:
            return _bqplot._bqplot_geometry(self.handle, *args)
        elif len(args) == 4:
            return _bqplot._bqplot_geometry(self.handle, args[0], args[1], args[2], args[3], 0)
        elif len(args) == 5:
            return _bqplot._bqplot_geometry(self.handle, *args)

    def position (self, x, y):
        return _bqplot._bqplot_position(self.handle, x, y)

    def size (self, w, h, fixed=0):
        return _bqplot._bqplot_geometry(self.handle, w, h, fixed)

    def screengeometry (self, width, aspect, screen, fixed=0):
        (s, x, y, w, h) = _bqplot._bqplot_getscreen (screen)
        if width > 0.9*w:
            width = int(0.9*w)
        height = int(width / aspect)
        if height > 0.9*h:
            height = int(0.9*h)
        width = int(aspect*height)
        x += w - width
        y += h - height
        return _bqplot._bqplot_geometry (self.handle, x, y, width, height, fixed)

    def title (self, title):
        return _bqplot._bqplot_window_title(self.handle, title)

    def adddockwidget (self, dockwidget, area):
        return _bqplot._bqplot_mainwindow_adddockwidget(self.handle, dockwidget.gethandle(), area)

    def resizedocks (self, dockwidgets, sizes, orientation):
        dockwidget_handles = []
        if type(dockwidgets) is list or type(dockwidgets) is tuple:
            for dockwidget in dockwidgets:
                dockwidget_handles.append (dockwidget.gethandle())
        else:
            dockwidget_handles.append (dockwidgets.gethandle())
        return _bqplot._bqplot_mainwindow_resizedocks(self.handle, dockwidget_handles, sizes, orientation)

class Frame (Widget):
    def __init__ (self, parent, close_window=None, handle=None):
        super (Frame, self).__init__ (parent, handle)
        if self.handle == None:
            if close_window == None:
                if self.parent == 'none':
                    self.handle = _bqplot._bqplot_frame('none')
                else:
                    self.handle = _bqplot._bqplot_frame(self.parent.gethandle())
            else:
                if self.parent == 'none':
                    self.handle = _bqplot._bqplot_frame('none', close_window)
                else:
                    self.handle = _bqplot._bqplot_frame(self.parent.gethandle(), close_window)
        super (Frame, self).setinstance (self.handle)

    def seteventtransparency (self):
        return _bqplot._bqplot_frame_seteventtransparency(self.handle)

    def cleareventtransparency (self):
        return _bqplot._bqplot_frame_cleareventtransparency(self.handle)

    def geometry (self, *args):
        if len(args) == 2:
            return _bqplot._bqplot_geometry(self.handle, args[0], args[1], 0)
        elif len(args) == 3:
            return _bqplot._bqplot_geometry(self.handle, *args)
        elif len(args) == 4:
            return _bqplot._bqplot_geometry(self.handle, args[0], args[1], args[2], args[3], 0)
        elif len(args) == 5:
            return _bqplot._bqplot_geometry(self.handle, *args)

    def position (self, x, y):
        return _bqplot._bqplot_position(self.handle, x, y)

    def size (self, w, h, fixed=0):
        return _bqplot._bqplot_geometry(self.handle, w, h, fixed)

    def screengeometry (self, width, aspect, screen, fixed=0):
        (s, x, y, w, h) = _bqplot._bqplot_getscreen (screen)
        if width > 0.9*w:
            width = int(0.9*w)
        height = int(width / aspect)
        if height > 0.9*h:
            height = int(0.9*h)
        width = int(aspect*height)
        x += w - width
        y += h - height
        return _bqplot._bqplot_geometry (self.handle, x, y, width, height, fixed)

    def title (self, title):
        return _bqplot._bqplot_window_title(self.handle, title)

    def setmodality (self, modality):
        return _bqplot._bqplot_frame_setmodality(self.handle, modality)

class Dialog (Widget):
    def __init__ (self, parent, handle=None):
        super (Dialog, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_dialog(self.parent.gethandle())
        super (Dialog, self).setinstance (self.handle)

    def title (self, title):
        return _bqplot._bqplot_window_title(self.handle, title)

class GridLayout (QtBasic):
    def __init__ (self, parent, handle=None):
        super (GridLayout, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_gridlayout(self.parent.gethandle())
        super (GridLayout, self).setinstance (self.handle)

    def addwidget (self, item, fromrow, fromcolumn, rowspan, columnspan):
        return _bqplot._bqplot_gridlayout_addwidget(self.handle, item.gethandle(), fromrow, fromcolumn, rowspan, columnspan)

    def columnstretch (self, column, stretch):
        return _bqplot._bqplot_gridlayout_columnstretch(self.handle, column, stretch)

    def rowstretch (self, row, stretch):
        return _bqplot._bqplot_gridlayout_rowstretch(self.handle, row, stretch)

    def columnwidth (self, column, width):
        return _bqplot._bqplot_gridlayout_columnwidth(self.handle, column, width)

    def rowheight (self, row, height):
        return _bqplot._bqplot_gridlayout_rowheight(self.handle, row, height)

    def spacing (self, hspacing, vspacing):
        return _bqplot._bqplot_gridlayout_spacing(self.handle, hspacing, vspacing)

    def hspacing (self, hspacing):
        return _bqplot._bqplot_gridlayout_spacing(self.handle, hspacing, -1)

    def vspacing (self, vspacing):
        return _bqplot._bqplot_gridlayout_spacing(self.handle, -1, vspacing)

    def margins (self, left, top, right, bottom):
        return _bqplot._bqplot_gridlayout_margins(self.handle, left, top, right, bottom)

class FormLayout (QtBasic):
    def __init__ (self, parent, handle=None):
        super (FormLayout, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_formlayout(self.parent.gethandle())
        super (FormLayout, self).setinstance (self.handle)

    def addrow (self, item, text):
        return _bqplot._bqplot_formlayout_addrow(self.handle, item.gethandle(), text)

    def spacing (self, hspacing, vspacing):
        return _bqplot._bqplot_formlayout_spacing(self.handle, hspacing, vspacing)

    def margins (self, left, top, right, bottom):
        return _bqplot._bqplot_formlayout_margins(self.handle, left, top, right, bottom)

class Splitter (Widget):
    def __init__ (self, parent, orientation='horizontal', handle=None):
        super (Splitter, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_splitter(self.parent.gethandle(), orientation)
        super (Splitter, self).setinstance (self.handle)

    def addwidget (self, item, index):
        return _bqplot._bqplot_splitter_addwidget(self.handle, item.gethandle(), index)

    def stretch (self, index, factor):
        return _bqplot._bqplot_splitter_stretch(self.handle, index, factor)

    def setsizes (self, sizes):
        return _bqplot._bqplot_splitter_setsizes(self.handle, sizes)

    def geometry (self, *args):
        if len(args) == 2:
            return _bqplot._bqplot_geometry(self.handle, args[0], args[1], 0)
        elif len(args) == 3:
            return _bqplot._bqplot_geometry(self.handle, *args)
        elif len(args) == 4:
            return _bqplot._bqplot_geometry(self.handle, args[0], args[1], args[2], args[3], 0)
        elif len(args) == 5:
            return _bqplot._bqplot_geometry(self.handle, *args)

class DockWidget (Widget):
    def __init__ (self, parent, child, title, handle=None):
        super (DockWidget, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_dockwidget(self.parent.gethandle(), child.gethandle(), title)
        super (DockWidget, self).setinstance (self.handle)

    def setfloating (self, ans):
        return _bqplot._bqplot_dockwidget_setfloating(self.handle, ans)

class ScrollArea (Widget):
    def __init__ (self, parent, handle=None):
        super (ScrollArea, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_scrollarea(self.parent.gethandle())
        super (ScrollArea, self).setinstance (self.handle)

    def widget (self, widget):
        _bqplot._bqplot_scrollarea_widget (self.handle, widget.gethandle())

    def widgetresizable (self, ans):
        _bqplot._bqplot_scrollarea_widgetresizable (self.handle, ans)

    def enable_horizontalscrollbar (self, ans):
        _bqplot._bqplot_scrollarea_scrollbar (self.handle, 'horizontal', ans)

    def enable_verticalscrollbar (self, ans):
        print (self.handle, 'vertical', ans)
        _bqplot._bqplot_scrollarea_scrollbar (self.handle, 'vertical', ans)

    def geometry (self, *args):
        if len(args) == 2:
            return _bqplot._bqplot_geometry(self.handle, args[0], args[1], 0)
        elif len(args) == 3:
            return _bqplot._bqplot_geometry(self.handle, *args)
        elif len(args) == 4:
            return _bqplot._bqplot_geometry(self.handle, args[0], args[1], args[2], args[3], 0)
        elif len(args) == 5:
            return _bqplot._bqplot_geometry(self.handle, *args)

class Button (Widget):
    def __init__ (self, parent, handle=None):
        super (Button, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_button(self.parent.gethandle())
        super (Button, self).setinstance (self.handle)

    def text (self, text_string):
        return _bqplot._bqplot_button_text(self.handle, text_string)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class Label (Widget):
    def __init__ (self, parent, handle=None):
        super (Label, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_label(self.parent.gethandle())
        super (Label, self).setinstance (self.handle)

    def text (self, text_string):
        return _bqplot._bqplot_label_text(self.handle, text_string)

    def selectable (self, ans):
        return _bqplot._bqplot_label_selectable(self.handle, ans)

class Checkbox (Widget):
    def __init__ (self, parent, handle=None):
        super (Checkbox, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_checkbox(self.parent.gethandle())
        super (Checkbox, self).setinstance (self.handle)

    def text (self, *args):
        if len(args) == 1:
            return _bqplot._bqplot_checkbox_text(self.handle, args[0])
        else:
            return _bqplot._bqplot_checkbox_text(self.handle)

    def state (self, *args):
        if len(args) == 1:
            return _bqplot._bqplot_checkbox_state(self.handle, args[0])
        else:
            return _bqplot._bqplot_checkbox_state(self.handle)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class Combobox (Widget):
    def __init__ (self, parent, handle=None):
        super (Combobox, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_combobox(self.parent.gethandle())
        super (Combobox, self).setinstance (self.handle)

    def clear (self):
        return _bqplot._bqplot_combobox_clear(self.handle)

    def additems (self, items):
        return _bqplot._bqplot_combobox_additems(self.handle, items)

    def text (self, *args):
        if len(args) == 1:
            return _bqplot._bqplot_combobox_text(self.handle, args[0])
        else:
            return _bqplot._bqplot_combobox_text(self.handle)

    def index (self, *args):
        if len(args) == 1:
            return _bqplot._bqplot_combobox_index(self.handle, args[0])
        else:
            return _bqplot._bqplot_combobox_index(self.handle)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class LineEdit (Widget):
    def __init__ (self, parent, handle=None):
        super (LineEdit, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_lineedit(self.parent.gethandle())
        super (LineEdit, self).setinstance (self.handle)

    def text (self, *args):
        if len(args) == 1:
            return _bqplot._bqplot_lineedit_text(self.handle, args[0])
        else:
            return _bqplot._bqplot_lineedit_text(self.handle)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class Menubar (Widget):
    def __init__ (self, parent, handle=None):
        self.parent = parent
        if handle != None:
            self.handle = handle
        else:
            self.handle = _bqplot._bqplot_menubar(parent.gethandle())
        super (Menubar, self).setinstance (self.handle)

    def addmenu (self, title):
        menu_handle = _bqplot._bqplot_menubar_addmenu(self.handle, title)
        return Menu (self, menu_handle)

class Menu (Widget):
    def __init__ (self, parent, handle=None):
        self.parent = parent
        if handle != None:
            self.handle = handle
        else:
            self.handle = _bqplot._bqplot_menu(parent.gethandle())
        super (Menu, self).setinstance (self.handle)

    def addmenu (self, title):
        menu_handle = _bqplot._bqplot_menu_addmenu(self.handle, title)
        return Menu (self.parent, menu_handle)

    def additem (self, *argv):
        if len(argv) == 1:
            title = argv[0]
            action_handle = _bqplot._bqplot_menu_additem(self.handle, self.parent.gethandle(), title)
            return Action (self.parent, title, action_handle)
        elif len(argv) == 2:
            action_parent = argv[0]
            title = argv[1]
            action_handle = _bqplot._bqplot_menu_additem(self.handle, action_parent.gethandle(), title)
            return Action (self.parent, title, action_handle)
        else:
            raise Exception ('Menu.additem: must have one or two arguments')
            return

    def addwidget (self, *argv):
        if len(argv) == 1:
            widget = argv[0]
            action_handle = _bqplot._bqplot_menu_addwidget(self.handle, self.parent.gethandle(), widget.gethandle())
            return Action (self.parent, None, action_handle)
        elif len(argv) == 2:
            action_parent = argv[0]
            widget = argv[1]
            action_handle = _bqplot._bqplot_menu_addwidget(self.handle, action_parent.gethandle(), widget.gethandle())
            return Action (self.parent, None, action_handle)
        else:
            raise Exception ('Menu.addwidget: must have one or two arguments')
            return

    def addseparator (self):
        _bqplot._bqplot_menu_addseparator(self.handle)

    def popup (self):
        _bqplot._bqplot_menu_popup(self.handle)

class Action (Widget):
    def __init__ (self, parent, title, handle=None):
        super (Action, self).__init__ (parent, handle)
        if self.handle == None:
            if self.parent == 'none':
                self.handle = _bqplot._bqplot_lineedit('none')
            else:
                self.handle = _bqplot._bqplot_menu_additem(self.parent.gethandle(), self.parent.getparent().gethandle(), title)
        super (Action, self).setinstance (self.handle)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class Timer (Widget):
    def __init__ (self, parent, handle=None):
        super (Timer, self).__init__ (parent, handle)
        if self.handle == None:
            self.handle = _bqplot._bqplot_timer(self.parent.gethandle())
        super (Timer, self).setinstance (self.handle)

    def interval (self, interval):
        return _bqplot._bqplot_timer_setinterval(self.handle, interval)

    def stop (self):
        return _bqplot._bqplot_timer_stop(self.handle)

    def singleshot (self, issingleshot):
        if issingleshot:
            return _bqplot._bqplot_timer_setsingleshot(self.handle)
        else:
            return _bqplot._bqplot_timer_clearsingleshot(self.handle)

    def start (self, **kwargs):
        if kwargs == None:
            return _bqplot._bqplot_timer_start(self.handle)
        if 'singleshot' in kwargs:
            if kwargs['singleshot']:
                _bqplot._bqplot_timer_setsingleshot(self.handle)
            else:
                _bqplot._bqplot_timer_clearsingleshot(self.handle)
        if 'interval' in kwargs:
            _bqplot._bqplot_timer_setinterval(self.handle, kwargs['interval'])
        return _bqplot._bqplot_timer_start(self.handle)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class FileDialog (object):
    def __init__ (self, parent, caption, directory, filt):
        self.parent = parent
        self.list = _bqplot._bqplot_filedialog(parent.gethandle(), caption, directory, filt, 'onetime')

    def getlist (self):
        return self.list

class SaveFileDialog (object):
    def __init__ (self, parent, caption, directory, filt):
        self.parent = parent
        self.name = _bqplot._bqplot_filedialog(parent.gethandle(), caption, directory, filt, 'save')

    def getname (self):
        return self.name

class PageDialog (object):
    def __init__ (self, parent):
        self.parent = parent
        _bqplot._bqplot_pagedialog(parent.gethandle())

class PrintDialog (object):
    def __init__ (self, parent, widget, image=None):
        if image == None:
            image = 0
        else:
            image = 1
        self.parent = parent
        _bqplot._bqplot_printdialog(parent.gethandle(), widget.gethandle(), image)

class Print (object):
    def __init__ (self, widget, fname=None, image=None, orientation='protrait'):
        if image == None:
            image = 0
        _bqplot._bqplot_print(widget.gethandle(), image, orientation, fname)

class PfTreeview (Widget):
    def __init__ (self, parent, bupfhandle, **kwargs):
        self.parent = parent
        if kwargs != None and 'handle' in kwargs:
            self.handle = kwargs['handle']
        else:
            self.handle = _bqplot._bqplot_pftreeview(self.parent.gethandle(), bupfhandle.gethandle())
            if kwargs != None:
                self.configure (**kwargs)
        super (PfTreeview, self).setinstance (self.handle)

    def configure (self, **kwargs):
        outargs = list()
        if kwargs != None:
            for key in kwargs.keys():
                outargs.append (key)
                outargs.append (kwargs[key])
        if len(outargs) < 1:
            return
        return _bqplot._bqplot_configure(self.handle, *outargs)

    def expandall (self):
        _bqplot._bqplot_pftreeview_expandall(self.handle)

    def collapseall (self):
        _bqplot._bqplot_pftreeview_collapseall(self.handle)

    def redraw (self):
        _bqplot._bqplot_pftreeview_redraw(self.handle)

    def checkvalue (self, index, value):
        return _bqplot._bqplot_pftreeview_checkvalue(self.handle, index, value)

class Tableview (ConfigWidget):
    def __init__ (self, parent, *args, **kwargs):
        self.parent = parent
        if kwargs and 'handle' in kwargs:
            self.handle = kwargs['handle']
            del kwargs['handle']
        else:
            self.handle = _bqplot._bqplot_tableview(self.parent.gethandle())
        super (Tableview, self).setinstance (self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def set_dimensions (self, nrows, ncols):
        _bqplot._bqplot_tableview_setdimensions (self.handle, nrows, ncols)

    def set_recordtab (self, record_tab):
        _bqplot._bqplot_tableview_setrecordtab (self.handle, record_tab)

    def get_totalsize (self):
        return _bqplot._bqplot_tableview_gettotalsize (self.handle)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class DbTableview (ConfigWidget):
    def __init__ (self, parent, db, *args, **kwargs):
        self.parent = parent
        if kwargs and 'handle' in kwargs:
            self.handle = kwargs['handle']
            del kwargs['handle']
        else:
            self.handle = _bqplot._bqplot_dbtableview(self.parent.gethandle(), db._dbptr)
        super (DbTableview, self).setinstance (self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def get_totalsize (self):
        return _bqplot._bqplot_dbtableview_gettotalsize (self.handle)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class EVTableview (ConfigWidget):
    def __init__ (self, parent, source, *args, **kwargs):
        self.parent = parent
        if kwargs and 'handle' in kwargs:
            self.handle = kwargs['handle']
            del kwargs['handle']
        else:
            if isinstance (source, EVServer):
                self.handle = _bqplot._bqplot_evtableview_evserver(self.parent.gethandle(), source.gethandle())
            elif isinstance (source, EVTableview):
                self.handle = _bqplot._bqplot_evtableview_evtableview(self.parent.gethandle(), source.gethandle())
            else:
                raise Exception ('EVTableview: source must be either EVServer or EVTableview object')
                return
        super (EVTableview, self).setinstance (self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def get_totalsize (self):
        return _bqplot._bqplot_evtableview_gettotalsize (self.handle)

    def select_row (self, row):
        _bqplot._bqplot_evtableview_select_row (self.handle, row)

    def show_row (self, row):
        _bqplot._bqplot_evtableview_show_row (self.handle, row)

    def select_event (self, evid, nofeedback=0):
        _bqplot._bqplot_evtableview_select_event (self.handle, evid, nofeedback)

    def select_origin (self, orid, nofeedback=0):
        _bqplot._bqplot_evtableview_select_origin (self.handle, orid, nofeedback)

    def set_sort_column (self, column, reverse=0):
        _bqplot._bqplot_evtableview_sortbycolumn (self.handle, column, reverse)

    def set_sort_colname (self, colname, reverse=0):
        _bqplot._bqplot_evtableview_sortbycolname (self.handle, colname, reverse)

    def reset (self, nrows, ncols):
        _bqplot._bqplot_evtableview_reset (self.handle, nrows, ncols)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class EVEventsTableview (ConfigWidget):
    def __init__ (self, parent, source, *args, **kwargs):
        self.parent = parent
        if kwargs and 'handle' in kwargs:
            self.handle = kwargs['handle']
            del kwargs['handle']
        else:
            if isinstance (source, EVServer):
                self.handle = _bqplot._bqplot_eveventstableview_evserver(self.parent.gethandle(), source.gethandle())
            else:
                raise Exception ('EVEventsTableview: source must be EVServer object')
                return
        super (EVEventsTableview, self).setinstance (self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def get_totalsize (self):
        return _bqplot._bqplot_eveventstableview_gettotalsize (self.handle)

    def select_evid (self, evid, nofeedback=0):
        _bqplot._bqplot_eveventstableview_select_evid (self.handle, evid, nofeedback)

    def show_evid (self, evid):
        _bqplot._bqplot_eveventstableview_show_evid (self.handle, evid)

    def set_sort_colname (self, colname, reverse=0):
        _bqplot._bqplot_eveventstableview_sortbycolname (self.handle, colname, reverse)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class EVOriginsTableview (ConfigWidget):
    def __init__ (self, parent, source, *args, **kwargs):
        self.parent = parent
        if kwargs and 'handle' in kwargs:
            self.handle = kwargs['handle']
            del kwargs['handle']
        else:
            if isinstance (source, EVEventsTableview):
                self.handle = _bqplot._bqplot_evoriginstableview_eveventstableview(self.parent.gethandle(), source.gethandle())
            else:
                raise Exception ('EVOriginsTableview: source must be EVEventsTableview object')
                return
        super (EVOriginsTableview, self).setinstance (self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def get_totalsize (self):
        return _bqplot._bqplot_evoriginstableview_gettotalsize (self.handle)

    def select_orid (self, orid, nofeedback=0):
        _bqplot._bqplot_evoriginstableview_select_orid (self.handle, orid, nofeedback)

    def show_orid (self, orid):
        _bqplot._bqplot_evoriginstableview_show_orid (self.handle, orid)

    def set_sort_colname (self, colname, reverse=0):
        _bqplot._bqplot_evoriginstableview_sortbycolname (self.handle, colname, reverse)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class CommandConsole (ConfigWidget):
    def __init__ (self, parent, *args, **kwargs):
        self.parent = parent
        if kwargs and 'handle' in kwargs:
            self.handle = kwargs['handle']
            del kwargs['handle']
        else:
            self.handle = _bqplot._bqplot_commandconsole(self.parent.gethandle())
        super (CommandConsole, self).setinstance (self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def sendcommand (self, command, *args):
        echo_command = 1
        echo_output = 1
        if args != None and len(args) > 0:
            echo_command = args.pop(0)
            if len(args) > 0:
                echo_output = args.pop(0)
                if len(args) > 0:
                    raise Exception ('CommandConsole:sendcommand: Too many arguments')
                    return
        _bqplot._bqplot_commandconsole_sendcommand(self.handle, command, echo_command, echo_output)
        return

    def printlineoutput (self, line, error_output_flag, output_done_flag):
        command = 'NULL'
        _bqplot._bqplot_commandconsole_printlineoutput(self.handle, command, line, error_output_flag, output_done_flag)
        return

    def importfromfile (self, fname, *args):
        echo = 1
        if args != None and len(args) > 0:
            echo = args.pop(0)
            if len(args) > 0:
                raise Exception ('CommandConsole:importfromfile: Too many arguments')
                return
        return _bqplot._bqplot_commandconsole_importfromfile(self.handle, fname, echo)

    def getcommand (self, command_index):
        return _bqplot._bqplot_commandconsole_getcommand(self.handle, command_index)

    def registercallback (self, proc, client_string='0'):
        return _bqplot._bqplot_register_callback(self.handle, proc, client_string)

class Spreadsheet (Widget):
    def __init__ (self, parent, proc, client_string='0', **kwargs):
        self.parent = parent
        if kwargs and 'handle' in kwargs:
            self.handle = kwargs['handle']
        else:
            pfstring = None
            if kwargs:
                pfstring = PfSubs.py2pfstring (kwargs)
            self.handle = _bqplot._bqplot_spreadsheet(self.parent.gethandle(), proc, client_string, pfstring)
        super (Spreadsheet, self).setinstance (self.handle)

    def select_row (self, row):
        _bqplot._bqplot_select_row (self.handle, row)

    def set_sort_column (self, column, reverse=0):
        _bqplot._bqplot_spreadsheet_sortbycolumn (self.handle, column, reverse)

    def reset (self, nrows, ncols):
        _bqplot._bqplot_spreadsheet_reset (self.handle, nrows, ncols)

class DbSpreadsheet (Widget):
    def __init__ (self, parent, db, row0, collist, colhash, **kwargs):
        self.parent = parent
        if kwargs != None and 'handle' in kwargs:
            self.handle = kwargs['handle']
        else:
            self.handle = None
            self.nrows = db.query (datascope.dbRECORD_COUNT);
            pfhash = {}
            if kwargs != None:
                pfhash= kwargs
            if 'nrows' not in pfhash:
                pfhash['nrows'] = self.nrows
            self.db = db
            self.dbsort = db.sort([])
            self.collist = collist
            self.ncols = len(self.collist)
            self.sort_col = None
            self.sort_reverse = None
            pfhash['ncols'] = len(self.collist)
            self.colhash = colhash
            pfhash['cell_text_templates'] = []
            for name in collist:
                pfhash['cell_text_templates'].append (colhash[name]['text_template'])
            pfstring = PfSubs.py2pfstring (kwargs)
            self.handle = _bqplot._bqplot_spreadsheet(self.parent.gethandle(), self, 'config_callback', pfstring)
            self.selectioncallback_proc = None
        self.last_hover_row = None
        self.last_hover_column = None
        super (DbSpreadsheet, self).setinstance (self.handle)

    def selectioncallback (self, proc):
        self.selectioncallback_proc = proc

    def config_callback (self, pfstring):
        argsd = PfSubs.pfstring2py (pfstring)
        if argsd['request'][0] == 'data':
            (row, column) = argsd['request'][1].split()
            row = int(row)
            column = int(column)
            colname = self.collist[column]
            params = self.colhash[colname]
            if colname == 'row':
                if 'format' in params:
                    string = params['format'] % row
                else:
                    string = str(row)
                return string
            self.dbsort.record = row
            value = self.dbsort.ex_eval ( params['source'])
            if 'format' in params:
                string = params['format'] % value
            else:
                string = str(value)
            if 'null' in params:
                if string == params['null']:
                    string = ''
            return string
        elif argsd['request'][0] == 'headerdata':
            section = int(argsd['request'][1])
            if argsd['request'][2] == 'vertical':
                return ''
            colname = self.collist[section]
            return self.colhash[colname]['label']
        elif argsd['request'][0] == 'hover':
            (row, column) = argsd['request'][1].split()
            row = int(row)
            column = int(column)
            self.last_hover_row = row
            self.last_hover_column = column
            return
        elif argsd['request'][0] == 'accentuated':
            (row, column) = argsd['request'][1].split()
            row = int(row)
            column = int(column)
            ret = '0'
            if self.last_hover_row != None:
                if row == self.last_hover_row and column == self.last_hover_column:
                    ret = '1'
            return ret
        elif argsd['request'][0] == 'sort':
            self.sort_col = int(argsd['request'][1])
            self.sort_reverse = int(argsd['request'][2])
            colname = self.collist[self.sort_col]
            params = self.colhash[colname]
            if 'source_sort' in params:
                keys = (params['source_sort'],)
                self.dbsort.free()
                if self.sort_reverse:
                    self.dbsort = self.db.sort (keys, reverse=True)
                else:
                    self.dbsort = self.db.sort (keys)
                if self.handle != None:
                    _bqplot._bqplot_spreadsheet_reset (self.handle, self.nrows, self.ncols)
                return "sort_ok"
            return ""
        elif argsd['request'][0] == 'selection_changed':
            row = int(argsd['request'][1])
            if self.selectioncallback_proc != None:
                pfhash = {}
                pfhash['handle'] = argsd['handle']
                pfhash['row'] = str(row)
                pfstring = PfSubs.py2pfstring (pfhash)
                self.dbsort.record = row
                self.selectioncallback_proc (self.dbsort, pfstring)
        return

    def select_row (self, row):
        _bqplot._bqplot_select_row (self.handle, self.sorttab[row])

    def set_sort_column (self, colname, reverse=0):
        if colname in self.collist:
            icol = self.collist.index(colname)
            _bqplot._bqplot_spreadsheet_sortbycolumn (self.handle, icol, reverse)

    def reset (self, db):
        self.db = db
        self.nrows = db.query (datascope.dbRECORD_COUNT);
        self.dbsort.free()
        self.dbsort = None
        if self.sort_col != None:
            colname = self.collist[self.sort_col]
            params = self.colhash[colname]
            if 'source_sort' in params:
                keys = (params['source_sort'],)
                if self.sort_reverse:
                    self.dbsort = self.db.sort (keys, reverse=True)
                else:
                    self.dbsort = self.db.sort (keys)
        else:
            self.dbsort = db.sort([])
        _bqplot._bqplot_spreadsheet_reset (self.handle, self.nrows, self.ncols)

class EVEventsSpreadsheet (Widget):
    def __init__ (self, parent, evclient, collist, colhash, mapevents=None, font='courier', font_size=12, handle=None):
        if parent == None:
            self.handle = handle
        else:
            self.parent = parent
            self.evclient = evclient
            self.evclient_handle = evclient.gethandle()
            self.collist = collist
            self.colhash = colhash
            self.evclient.grab_lock()
            self.events = self.evclient.events_get ()
            self.last_modified_evid = self.evclient.last_modified_event_get_evid ()
            self.evclient.release_lock()
            self.last_modified_event_index = None
            if self.last_modified_evid != None:
                self.last_modified_event_index = ev_events_evid_to_index (self.events, self.last_modified_evid)
            self.selected_evid = self.last_modified_evid
            self.nrows = ev_events_number (self.events)
            self.ncols = len(collist)
            self.sort_col = -1
            self.sort_reverse = 0
            self.evid = -1
            self.event_spreadsheet = None
            self.mapevents = mapevents
            if self.mapevents != None:
                self.mapevents.registercallback (self, 'mapevents_glyphs_callback', 'mapevents_event_callback', 'ev_callback')
            self.selectioncallback_proc = []

            pfhash = {}
            pfhash['nrows'] = str(self.nrows)
            pfhash['ncols'] = str(self.ncols)
            pfhash['font'] = font
            pfhash['font_size'] = str(font_size)
            pfhash['default_cell_width'] = '100'
            pfhash['cell_text_templates'] = []
            for name in collist:
                pfhash['cell_text_templates'].append (colhash[name]['text_template'])

            pfstring = PfSubs.py2pfstring (pfhash)
            self.handle = None
            self.handle = _bqplot._bqplot_spreadsheet(parent.gethandle(), self, 'config_callback', pfstring)
            if self.selected_evid != None:
                row = ev_events_evid_to_row (self.events, self.selected_evid)
                if row == None:
                    self.selected_evid = None
                else:
                    _bqplot._bqplot_spreadsheet_select_row (self.handle, row)
        super (EVEventsSpreadsheet, self).setinstance (self.handle)

    def set_sort_column (self, colname, reverse=0):
        if colname in self.collist:
            icol = self.collist.index(colname)
            _bqplot._bqplot_spreadsheet_sortbycolumn (self.handle, icol, reverse)

    def register_event_spreadsheet (self, event_spreadsheet):
        # print "registering event_spreadsheet ", event_spreadsheet
        self.event_spreadsheet = event_spreadsheet

    def sort (self):
        if self.sort_col == -1:
            if self.selected_evid != None:
                row = ev_events_evid_to_row (self.events, self.selected_evid)
                if row == None:
                    self.selected_evid = None
                else:
                    if self.handle != None:
                        _bqplot._bqplot_spreadsheet_select_row (self.handle, row)
            return
        colname = self.collist[self.sort_col]
        params = self.colhash[colname]
        queries_sort = []
        function = None
        if 'function' in params:
            function = params['function']
        if 'source_sort' in params:
            queries_sort = params['source_sort'].split(';')
        if function == None:
            query = queries_sort[0]
            ev_events_set_sorttab (self.evclient_handle, self.events, query, self.sort_reverse)
        else:
            l = function.split (' ')
            if len(l) == 1:
                function = '$' + l[0]
                function_args = ''
            else:
                function = '$' + l[0]
                function_args = ' '.join(l[1:])
            ev_events_set_sorttab (self.evclient_handle, self.events, function, function_args, self.sort_reverse)
        if self.selected_evid != None:
            row = ev_events_evid_to_row (self.events, self.selected_evid)
            if row == None:
                self.selected_evid = None
            else:
                if self.handle != None:
                    return _bqplot._bqplot_spreadsheet_select_row (self.handle, row)

    def selectioncallback (self, proc):
        self.selectioncallback_proc.append (proc)

    def register_ev_callback (self):
        ev_client_register_callback (self.evclient_handle, self, "ev_callback")

    def ev_callback (self, pfstring):
        argsd = PfSubs.pfstring2py(pfstring)

        if argsd['events_changed'] == '0' \
                and argsd['last_modified_event_changed'] == '0':
            return

        #print "EVEventsSpreadsheet.ev_callback",
        #if argsd['database_reset'] != '0':
        #    print "database_reset",
        #if argsd['events_changed'] != '0':
        #    print "events_changed",
        #if argsd['last_modified_event_changed'] != '0':
        #    print "last_modified_event_changed",
        #print

        events_tmp = None
        detections_tmp = None
        last_modified_event_tmp = None

        self.evclient.grab_lock()

        if argsd['events_changed'] != '0':
            events_tmp = self.evclient.events_get()
        if argsd['last_modified_event_changed'] != '0':
            last_modified_evid_tmp = self.evclient.last_modified_event_get_evid ()

        self.evclient.release_lock()

        if argsd['events_changed'] != '0':
            # print "events changes:", self.events, events_tmp
            # ev_print(self.events, events_tmp)
            ev_events_free(self.events)
            self.events = events_tmp
            if self.event_spreadsheet != None:
                self.event_spreadsheet.events = self.events
            self.nrows = ev_events_number (self.events)
            _bqplot._bqplot_spreadsheet_reset (self.handle, self.nrows, self.ncols)
            self.sort()
            if self.event_spreadsheet != None:
                self.event_spreadsheet.select_evid ( self.selected_evid )
            if self.selected_evid != None:
                row = ev_events_evid_to_row (self.events, self.selected_evid)
                if row == None:
                    self.selected_evid = None
                else:
                    _bqplot._bqplot_spreadsheet_select_row (self.handle, row)
                    for selectioncallback in self.selectioncallback_proc:
                        pfhash = {}
                        pfhash['handle'] = argsd['handle']
                        pfhash['row'] = str(row)
                        pfhash['evid'] = str(self.selected_evid)
                        index = ev_events_evid_to_index (self.events, self.selected_evid)
                        pfhash['index'] = str(index)
                        pfstring = PfSubs.py2pfstring (pfhash)
                        selectioncallback (pfstring)

        if argsd['last_modified_event_changed'] != '0':
            # print "last Modified event changes:", self.last_modified_evid, last_modified_evid_tmp
            if self.last_modified_evid != last_modified_evid_tmp:
                self.last_modified_evid = last_modified_evid_tmp
                self.selected_evid = self.last_modified_evid
                row = ev_events_evid_to_row (self.events, self.selected_evid)
                if self.event_spreadsheet != None:
                    self.event_spreadsheet.select_evid ( self.selected_evid )
                if row == None:
                    self.selected_evid = None
                else:
                    _bqplot._bqplot_spreadsheet_select_row (self.handle, row)
                    for selectioncallback in self.selectioncallback_proc:
                        pfhash = {}
                        pfhash['handle'] = argsd['handle']
                        pfhash['row'] = str(row)
                        pfhash['evid'] = str(self.selected_evid)
                        index = ev_events_evid_to_index (self.events, self.selected_evid)
                        pfhash['index'] = str(index)
                        pfstring = PfSubs.py2pfstring (pfhash)
                        selectioncallback (pfstring)

    def config_callback (self, pfstring):
        argsd = PfSubs.pfstring2py (pfstring)
        if argsd['request'][0] == 'data':
            (row, column) = argsd['request'][1].split()
            row = int(row)
            if row >= self.nrows:
                return
            index = ev_events_get_index (self.events, row)
            column = int(column)
            colname = self.collist[column]
            params = self.colhash[colname]
            sources = []
            if 'source' in params:
                sources = params['source'].split(';')
            function = None
            if 'function' in params:
                function = params['function']
            values = []
            if len(sources) > 0:
                for i in range(len(sources)):
                    query = sources[i]
                    query = query.replace ('%index', '%d' % index)
                    value = ev_events_get_value (self.evclient_handle, self.events, query)
                    if value == None:
                        values = []
                        break
                    values.append(value)
            elif function != None:
                value = ev_events_get_value_function (self.evclient_handle, self.events, function, row)
                if value != None:
                    values.append(value)
            if len(values) < 1:
                return ''
            if 'format' in params:
                string = params['format'] % tuple(values)
            else:
                string = str(value)
            if 'null' in params:
                if string == params['null']:
                    string = ''
            return string
        if argsd['request'][0] == 'headerdata':
            section = int(argsd['request'][1])
            if argsd['request'][2] == 'vertical':
                return ''
            colname = self.collist[section]
            return self.colhash[colname]['label']
        if argsd['request'][0] == 'sort':
            self.sort_col = int(argsd['request'][1])
            self.sort_reverse = int(argsd['request'][2])
            self.sort()
            return "sort_ok"
        if argsd['request'][0] == 'selection_changed':
            row = int(argsd['request'][1])
            evid = ev_events_get_evid (self.events, row)
            # print('event selection_changed', evid)
            if self.mapevents != None:
                self.mapevents.configure ( "show_evid", evid )
            if self.event_spreadsheet != None:
                self.event_spreadsheet.select_evid ( evid )
            for selectioncallback in self.selectioncallback_proc:
                pfhash = {}
                pfhash['handle'] = argsd['handle']
                pfhash['row'] = str(row)
                pfhash['evid'] = str(evid)
                index = ev_events_get_index (self.events, row)
                pfhash['index'] = str(index)
                pfstring = PfSubs.py2pfstring (pfhash)
                selectioncallback (pfstring)
        return

    def mapevents_glyphs_callback (self, pfstring):
        argsd = PfSubs.pfstring2py (pfstring)
        if 'bqglyph' in argsd:
            if argsd['bqglyph']['type'] == 'event':
                if argsd['display_mode'] == 'initial' and argsd['bqevent']['type'] == 'leave' and self.selected_evid != None:
                    evid = self.selected_evid
                else:
                    evid = int(argsd['bqglyph']['evid'])
                if evid != self.evid:
                    self.evid = evid
                    row = ev_events_evid_to_row (self.events, self.evid)
                    if self.handle != None:
                        _bqplot._bqplot_spreadsheet_show_row (self.handle, row)

    def mapevents_event_callback (self, pfstring):
        argsd = PfSubs.pfstring2py (pfstring)
        if argsd['display_mode'] == 'selected':
            evid = int(argsd['evid'])
            self.selected_evid = evid
            # print 'event selection_changed', evid
            if self.mapevents != None:
                self.mapevents.configure ( "show_evid", evid )
            if self.event_spreadsheet != None:
                self.event_spreadsheet.select_evid ( evid )
            for selectioncallback in self.selectioncallback_proc:
                pfhash = {}
                pfhash['handle'] = argsd['handle']
                row = ev_events_evid_to_row (self.events, self.selected_evid)
                pfhash['row'] = str(row)
                pfhash['evid'] = str(evid)
                index = ev_events_evid_to_index (self.events, self.selected_evid)
                pfhash['index'] = str(index)
                pfstring = PfSubs.py2pfstring (pfhash)
                selectioncallback (pfstring)

class EVEventSpreadsheet (Widget):
    def __init__ (self, parent, evclient, collist, colhash, events, evid, spreadsheet_events=None, mapevents=None, font='courier', font_size=12, handle=None):
        if parent == None:
            self.handle = handle
        else:
            self.parent = parent
            self.evclient = evclient
            self.evclient_handle = evclient.gethandle()
            self.collist = collist
            self.colhash = colhash
            self.events = events
            self.orid = -1
            self.nrows = 0
            self.ncols = len(collist)
            self.sort_col = -1
            self.sort_reverse = 0
            self.selectioncallback_proc = []
            self.mapevents = mapevents
            self.spreadsheet_events = spreadsheet_events
            if self.mapevents != None:
                self.mapevents.registercallback (self, 'mapevents_callback')
            if self.spreadsheet_events != None:
                spreadsheet_events.register_event_spreadsheet (self)

            pfhash = {}
            pfhash['nrows'] = 0
            pfhash['ncols'] = str(self.ncols)
            pfhash['font'] = font
            pfhash['font_size'] = str(font_size)
            pfhash['default_cell_width'] = '100'
            pfhash['cell_text_templates'] = []
            pfhash['accentuated_foreground'] = 'red'
            pfhash['accentuated_selected_foreground'] = 'red'
            for name in collist:
               pfhash['cell_text_templates'].append (colhash[name]['text_template'])

            pfstring = PfSubs.py2pfstring (pfhash)
            self.handle = None
            if parent == "toplevel":
                self.handle = _bqplot._bqplot_spreadsheet("toplevel", self, 'config_callback', pfstring)
            else:
                self.handle = _bqplot._bqplot_spreadsheet(parent.gethandle(), self, 'config_callback', pfstring)

            self.select_evid (evid)
        super (EVEventSpreadsheet, self).setinstance (self.handle)

    def set_sort_column (self, colname, reverse=0):
        if colname in self.collist:
            icol = self.collist.index(colname)
            _bqplot._bqplot_spreadsheet_sortbycolumn (self.handle, icol, reverse)

    def select_evid (self, evid):
        # print "event select_evid ", evid
        self.evid = evid
        if self.evid == None:
            self.nrows = 0
            self.events_index = None
            self.selected_orid = None
            self.prefor = None
            self.origins = None
        else:
            self.events_index = ev_events_evid_to_index (self.events, self.evid)
            query = 'events[%d].norigins' % self.events_index
            self.nrows = ev_events_get_value (self.evclient_handle, self.events, query)
            query = 'events[%d].prefor' % self.events_index
            self.selected_orid = ev_events_get_value (self.evclient_handle, self.events, query)
            self.prefor = self.selected_orid
            self.origins = ev_events_origins_get (self.events, self.events_index, 0)
        _bqplot._bqplot_spreadsheet_reset (self.handle, self.nrows, self.ncols)
        self.sort()
        if self.selected_orid != None:
            row = ev_origins_orid_to_row (self.origins, self.selected_orid)
            if row == None:
                self.selected_orid = None
            else:
                _bqplot._bqplot_spreadsheet_select_row (self.handle, row)

    def selectioncallback (self, proc):
        self.selectioncallback_proc.append (proc)

    def sort (self):
        if self.sort_col == -1:
            if self.origins != None:
                row = ev_origins_orid_to_row (self.origins, self.selected_orid)
                if row == None:
                    self.selected_orid = None
                else:
                    if self.handle != None:
                        _bqplot._bqplot_spreadsheet_select_row (self.handle, row)
            return
        if self.nrows == 0:
            return
        colname = self.collist[self.sort_col]
        params = self.colhash[colname]
        queries_sort = []
        function = None
        if 'function' in params:
            function = params['function']
        if 'source_sort' in params:
            queries_sort = params['source_sort'].split(';')
        if function == None:
            query = queries_sort[0]
            ev_origins_set_sorttab (self.evclient_handle, self.origins, query, self.sort_reverse)
        else:
            l = function.split (' ')
            if len(l) == 1:
                function = '$' + l[0]
                function_args = ''
            else:
                function = '$' + l[0]
                function_args = ' '.join(l[1:])
            ev_origins_set_sorttab (self.evclient_handle, self.origins, function, function_args, self.sort_reverse)
        row = ev_origins_orid_to_row (self.origins, self.selected_orid)
        if row == None:
            self.selected_orid = None
        else:
            if self.handle != None:
                _bqplot._bqplot_spreadsheet_select_row (self.handle, row)
    
    def config_callback (self, pfstring):
        argsd = PfSubs.pfstring2py (pfstring)
        if argsd['request'][0] == 'data':
            (row, column) = argsd['request'][1].split()
            row = int(row)
            if row >= self.nrows:
                return
            index = ev_origins_get_index (self.origins, row)
            column = int(column)
            colname = self.collist[column]
            params = self.colhash[colname]
            sources = []
            if 'source' in params:
                sources = params['source'].split(';')
            function = None
            if 'function' in params:
                function = params['function']
            values = []
            if len(sources) > 0:
                for i in range(len(sources)):
                    query = 'events[%d].%s' % (self.events_index, sources[i])
                    query = query.replace ('%index', '%d' % index)
                    # print query
                    value = ev_events_get_value (self.evclient_handle, self.events, query)
                    if value == None:
                        values = []
                        break
                    values.append(value)
            elif function != None:
                value = ev_events_get_value_function (self.evclient_handle, self.events, function, index, 0)
                if value != None:
                    values.append(value)
            if len(values) < 1:
                return ''
            if 'format' in params:
                string = params['format'] % tuple(values)
            else:
                string = str(value)
            if 'null' in params:
                if string == params['null']:
                    string = ''
            return string
        if argsd['request'][0] == 'headerdata':
            section = int(argsd['request'][1])
            if argsd['request'][2] == 'vertical':
                return ''
            colname = self.collist[section]
            return self.colhash[colname]['label']
        if argsd['request'][0] == 'sort':
            self.sort_col = int(argsd['request'][1])
            self.sort_reverse = int(argsd['request'][2])
            self.sort()
            return "sort_ok"
        if argsd['request'][0] == 'accentuated':
            (row, column) = argsd['request'][1].split()
            accentuated = '0'
            row = int(row)
            if row >= self.nrows:
                return accentuated
            orid = ev_origins_get_orid (self.origins, row)
            if orid == self.prefor:
                accentuated = '1'
            return accentuated
        if argsd['request'][0] == 'selection_changed':
            row = int(argsd['request'][1])
            orid = ev_origins_get_orid (self.origins, row)
            # print('origin selection_changed', orid)
            if self.mapevents != None:
                self.mapevents.configure ( "show_orid", orid )
            for selectioncallback in self.selectioncallback_proc:
                pfhash = {}
                pfhash['handle'] = argsd['handle']
                pfhash['handle_origins'] = self.origins
                pfhash['row'] = str(row)
                pfhash['orid'] = str(orid)
                pfhash['evid'] = str(self.evid)
                index = ev_origins_get_index (self.origins, row)
                pfhash['index'] = str(index)
                pfstring = PfSubs.py2pfstring (pfhash)
                selectioncallback (pfstring)
            return "selection_ok"
    def mapevents_callback (self, pfstring):
        argsd = PfSubs.pfstring2py (pfstring)
        if 'bqglyph' in argsd:
            if argsd['bqglyph']['type'] == 'origin':
                orid = int(argsd['bqglyph']['orid'])
                if orid != self.orid:
                    self.orid = orid
                    row = ev_origins_orid_to_row (self.origins, self.orid)
                    if self.handle != None:
                        _bqplot._bqplot_spreadsheet_show_row (self.handle, row)

class Viewport (ViewportItem):
    def __init__ (self, parent, *args, **kwargs):
        self.parent = parent
        self.handle = _bqplot._bqplot_viewport(parent.gethandle())
        super (Viewport, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def delete (self):
        return _bqplot._bqplot_delete(self.item_handle)

    def show (self):
        return _bqplot._bqplot_show(self.item_handle)

    def hide (self):
        return _bqplot._bqplot_hide(self.item_handle)

    def getwcoords (self, x, y):
        return _bqplot._bqplot_getwcoords(self.handle, x, y)

    def getwcoordbb (self):
        return _bqplot._bqplot_viewport_getwcoordbb(self.handle)

    def getdimensions (self):
        return _bqplot._bqplot_viewport_getdimensions(self.handle)

    def setcursor (self, cursor_name):
        return _bqplot._bqplot_viewport_setcursor(self.handle, cursor_name)

    def showdisplaylist (self):
        return _bqplot._bqplot_viewport_showdisplaylist(self.handle)

class Axes (ViewportItem):
    def __init__ (self, master, *args, **kwargs):
        self.master = master
        self.handle = _bqplot._bqplot_axes(master.gethandle())
        super (Axes, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

class Datagrid (ViewportItem):
    def __init__ (self, master, *args, **kwargs):
        self.master = master
        self.handle = _bqplot._bqplot_datagrid(master.gethandle())
        super (Datagrid, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

class Text (ViewportItem):
    def __init__ (self, master, text, x, y, *args, **kwargs):
        self.master = master
        self.handle = _bqplot._bqplot_text(master.gethandle(), text, x, y)
        super (Text, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

class Textpopup (ViewportItem):
    def __init__ (self, master, *args, **kwargs):
        self.master = master
        self.handle = _bqplot._bqplot_textpopup(master.gethandle())
        super (Textpopup, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

class Polyline (ViewportItem):
    def __init__ (self, master, *args, **kwargs):
        self.master = master
        self.handle = _bqplot._bqplot_polyline(master.gethandle())
        super (Polyline, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def configure (self, *args, **kwargs):
        if args != None:
            outargs = list(args)
        else :
            outargs = list()
        if kwargs != None:
            for key in kwargs.keys():
                outargs.append (key)
                outargs.append (kwargs[key])
        if len(outargs) < 1:
            return
        outargs2 = list()
        i = 0
        while i < len(outargs):
            key = outargs[i]
            i += 1
            value = outargs[i]
            i += 1
            if key == 'points':
                new_value = ''
                if type(value) is list or type(value) is tuple:
                    for point in value:
                        point_str = ''
                        if type(point) is list or type(point) is tuple:
                            for entry in point:
                                if point_str != '':
                                    point_str += " "
                                point_str += str(entry)
                        else:
                            point_str = str(value)
                        if point_str[-1] != "\n":
                            point_str += "\n"
                        new_value += point_str
                else:
                    point_str = str(value)
                    if point_str[-1] != "\n":
                        point_str += "\n"
                    new_value += point_str
                value = new_value
            if key != 'trace':
                outargs2.append (key)
                outargs2.append (value)

        if len(outargs2) > 0:
            ret = _bqplot._bqplot_configure(self.item_handle, *outargs2)
            if ret == None:
                return None

        i = 0
        while i < len(outargs):
            key = outargs[i]
            i += 1
            value = outargs[i]
            i += 1
            if key == 'trace':
                ret = _bqplot._bqplot_polyline_setsource_trace (self.item_handle, value._dbptr)
                if ret == None:
                    return None

        return 0

class Polypoint (ViewportItem):
    def __init__ (self, master, *args, **kwargs):
        self.master = master
        self.handle = _bqplot._bqplot_polypoint(master.gethandle())
        super (Polypoint, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

class Taskbar (ViewportItem):
    def __init__ (self, handle, *args, **kwargs):
        self.handle = handle
        super (Taskbar, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

class Map (ViewportItem):
    def __init__ (self, master, *args, **kwargs):
        if isinstance (master, str):
            self.handle = master
        elif isinstance (master, Viewport):
            self.master = master
            self.handle = _bqplot._bqplot_map(master.gethandle())
        else:
            self.master = None
            self.handle = _bqplot._bqplot_map(master)
        super (Map, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def setdefaults(self):
        return _bqplot._bqplot_map_setdefaults(self.handle)

    def gettaskbar(self):
        tbhandle = _bqplot._bqplot_map_gettaskbar(self.handle)
        if not tbhandle in ViewportItem.instances:
            ViewportItem.instances[tbhandle] = Taskbar (tbhandle)
        return ViewportItem.instances[tbhandle]

class Mapevents (ViewportItem):
    def __init__ (self, *args, **kwargs):
        argsl = list(args)
        if isinstance (argsl[0], Viewport):
            self.events_vp = argsl[0].gethandle()
            argsl.pop(0)
            if len(argsl) < 1 or isinstance (argsl[0], Viewport) == False:
                self.handle = _bqplot._bqplot_mapevents(self.events_vp)
            else:
                self.event_vp = argsl[0].gethandle()
                argsl.pop(0)
                if len(argsl) < 1 or isinstance (argsl[0], Viewport) == False:
                    self.handle = _bqplot._bqplot_mapevents(self.events_vp, self.event_vp)
                else:
                    self.origins_vp = argsl[0].gethandle()
                    argsl.pop(0)
                    self.handle = _bqplot._bqplot_mapevents(self.events_vp, self.event_vp, self.origins_vp)
        else:
            raise Exception ('Mapevents: first argument must be viewport')
            return
        args = argsl
        super (Mapevents, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def setdefaults(self, nort=False):
        if nort:
            return _bqplot._bqplot_mapevents_setdefaults_nort(self.handle)
        else:
            return _bqplot._bqplot_mapevents_setdefaults(self.handle)

    def getmap(self):
        maphandle = _bqplot._bqplot_mapevents_getmap(self.handle)
        if not maphandle in ViewportItem.instances:
            ViewportItem.instances[maphandle] = Map (maphandle)
        return ViewportItem.instances[maphandle]

    def gettaskbar(self):
        tbhandle = _bqplot._bqplot_mapevents_gettaskbar(self.handle)
        if not tbhandle in ViewportItem.instances:
            ViewportItem.instances[tbhandle] = Taskbar (tbhandle)
        return ViewportItem.instances[tbhandle]

    def seteventview (self, ev):
        if isinstance (ev, EVServer):
            return _bqplot._bqplot_mapevents_seteventview(self.handle, ev.gethandle())
        else:
            return _bqplot._bqplot_mapevents_seteventview(self.handle, ev)

    def refresheventview (self):
        return _bqplot._bqplot_mapevents_refresheventview(self.handle)

    def link_eveventstableview (self, eveventstableview):
        return _bqplot._bqplot_mapevents_link_eveventstableview(self.handle, eveventstableview.gethandle())

    def registercallback (self, proc, client_string='0', client_string2=None, client_string3=None):
        if client_string2 == None:
            return _bqplot._bqplot_register_callback(self.handle, proc, client_string)
        elif client_string3 == None:
            return _bqplot._bqplot_register_callback(self.handle, proc, client_string, proc, client_string2)
        else:
            return _bqplot._bqplot_register_callback(self.handle, proc, client_string, proc, client_string2, proc, client_string3)

class Traceview (ViewportItem):
    def __init__ (self, master, *args, **kwargs):
        self.master = master
        cc = ''
        make_taskbar = 1
        vptaskbar = ''
        args = list(args)
        if len(args) > 0 and isinstance (args[0], CommandConsole):
            cc = args.pop(0).gethandle()
            if len(args) > 0 and isinstance (args[0], bool):
                if args.pop(0):
                    make_taskbar = 1
                else:
                    make_taskbar = 0;
                if len(args) > 0 and isinstance (args[0], Viewport):
                    vptaskbar = args.pop(0).gethandle()
        self.handle = _bqplot._bqplot_traceview(master.gethandle(), cc, make_taskbar, vptaskbar)
        super (Traceview, self).__init__ (self, self.handle)
        if args != None and kwargs != None:
            self.configure (*args, **kwargs)

    def createtraces (self, source, *args):
        if isinstance (source, EVServer):
            stachan_accept = '.* .*'
            if len(args) == 1:
                stachan_accept = args[0]
            _bqplot._bqplot_traceview_createtraces_evserver(self.handle, source.gethandle(), stachan_accept)
        else:
            raise Exception ('Traceview: source must be EVServer object')
            return

    def setcommandconsole (self, cc):
        if isinstance (cc, CommandConsole):
            _bqplot._bqplot_traceview_setcommandconsole(self.handle, cc.gethandle())
        else:
            raise Exception ('Traceview: cc must be CommandConsole object')
            return

    def link_eveventstableview (self, eveventstableview):
        return _bqplot._bqplot_traceview_link_eveventstableview(self.handle, eveventstableview.gethandle())

    def sendcommand (self, command, *args):
        provide_output = 0
        send_to_cc = 1
        nofeedback = -1
        if len(args) > 0:
            provide_output = args.pop(0)
            if len(args) > 0:
                send_to_cc = args.pop(0)
                if len(args) > 0:
                    nofeedback = args.pop(0)
                    if len(args) > 0:
                        raise Exception ('Traceview:sendcommand: Too many arguments')
                        return
        return _bqplot._bqplot_traceview_sendcommand (self.handle, command, provide_output, send_to_cc, nofeedback)

class GraphicsEngine (object):
    def __init__ (self):
        return

    def gethandle (self):
        return 'root'

    def getscreen (self, screen):
        return _bqplot._bqplot_getscreen(screen)

    def quit (self):
        return _bqplot._bqplot_quit()

    def stylesheet (self, stylesheet_text):
        return _bqplot._bqplot_setstylesheet('root', stylesheet_text)

    def qtmainloop (self):
        return _bqplot._bqplot_qtmainloop()

    def pymainloop (self):
        return _bqplot._bqplot_pymainloop()

if __name__ == '__main__':
    import unittest
    import os
    import sys
    sys.path.append(os.environ['ANTELOPE'] + "/data/python")

    import signal

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    from antelope.buvector import *
    from antelope import stock

    class PfSubs(object):
        def __init__ (self):
                return

        @classmethod
        def pfstring2py (self, pfstring, pyd=None):
                pf = stock.ParameterFile()
                pf.pfcompile (pfstring)
                pyf = pf.pf2dict()
                if pyd != None:
                        pyd.update(pyf)
                else:
                        pyd = pyf
                return (pyd)

        @classmethod
        def py2pfstring (self, pyd):
                pf = stock.ParameterFile()
                pf['foo'] = pyd
                pfstr = pf.pf2string()
                pfstr = pfstr[10:-2]
                return pfstr

    class Mycallback(object):
        instance = -1
        instances = {}

        def __init__ (self, vp, pl, vecpl):
            Mycallback.instance += 1
            Mycallback.instances[Mycallback.instance] = self
            self.vp = vp
            self.pl = pl
            self.vecpl = vecpl

        @classmethod
        def mycallback (self, pfstring):
            argsd = PfSubs.pfstring2py (pfstring)
            instance = int(argsd['client_string'])
            obj = Mycallback.instances[instance]
    
            if argsd['bqglyph']['state'] == 'pan':
                if argsd['bqevent']['type'] == 'mousemove':
                    (x, y) = (float(argsd['bqevent']['pt'][0]), float(argsd['bqevent']['pt'][1]))
                    (xw, yw) = obj.vp.getwcoords (x, y)
                    index = int(argsd['bqglyph']['index'])
                    buvector_put (obj.vecpl, index, xw, yw)
                    if index == 0:
                        buvector_put (obj.vecpl, 4, xw, yw)
                    obj.pl.configure ("vector", obj.vecpl)
                    obj.pl.setitemstale ()
    
            return 'ok'

    class Testbqplot_fixture(object):

        tempdir = '/tmp/python_bqplottest_' + str(os.getuid()) + str(os.getpid())

        def start(self):

            os.mkdir(self.tempdir)
            os.chdir(self.tempdir)

        def stop(self):

            os.chdir("/")
            os.rmdir(self.tempdir)

    class Testbqplot(unittest.TestCase):

        def test_procedure_bqplot_test(self):
    
            ge = GraphicsEngine()

            mainwindow = MainWindow(ge)

            menubar = Menubar (mainwindow)
            file_menu = menubar.addmenu ("&File")
            quit_item = file_menu.additem ("&Quit")
            quit_item.registercallback ( lambda self: ge.quit() )
            file_menu.addseparator ()
            page_item = file_menu.additem ("&Page Setup...")
            page_item.registercallback ( lambda self: PageDialog(mainwindow.getframe()) )
            print_item = file_menu.additem ("&Print...")
            print_item.registercallback ( lambda self: PrintDialog(mainwindow, mainwindow, 1) )
            view_menu = menubar.addmenu ("&View")

            vp = Viewport (mainwindow)
            self.assertTrue(vp != None)

            vp.configure ( \
                    "xleft",         -1.0, \
                    "xright",        5.0, \
                    "ybottom",        -1.1, \
                    "ytop",                1.1, \
                    "mleft",        80, \
                    "mright",        20, \
                    "mbottom",        50, \
                    "mtop",                30, \
                    "fill_frame",         "lightblue", \
                    "fill",                "#e0e0e0" )

            axes = Axes (vp)
            self.assertTrue(axes != None)

            axes.configure ( \
                    "xlabel",                 "My X-stuff", \
                    "ylabel",                 "My Y-stuff", \
                    "xformat",                "%.1f", \
                    "yformat",                "%.1f", \
                    "linewidth_box",        0, \
                    "linewidth_tics",        2, \
                    "linewidth_tics_small",        0, \
                    "axis_style",                "sw", \
                    "tic_style",                "siwinoeo", \
                    "linewidth_grids",        2, \
                    "linewidth_grids_small",0, \
                    "color_grids",                "black", \
                    "color_grids_small",        "#c0c0c0" )

            text = Text (vp, "My text", "0.5v", "top+10")
            self.assertTrue(text != None)

            text.configure ( \
                    "color_text",        "red", \
                    "font",                "Times 14 Bold", \
                    "justification","s", \
                    "angle",        0 )
            
            polyline = Polyline (vp)
            self.assertTrue(polyline != None)
            
            polyline.configure ( \
                    "test",                "on", \
                    "color_outline","#400000ff", \
                    "linewidth",        "3" )
            
            layer1_name = axes.setlayer ("layer1")
            text.setlayer ("layer1")
            polyline.setlayer ("layer1")

            layer1_on_item = view_menu.additem ("&Show layer1")
            layer1_on_item.registercallback ( lambda self: vp.setlayervisibility (layer1_name, 1) )
            layer1_off_item = view_menu.additem ("&Hide layer1")
            layer1_off_item.registercallback ( lambda self: vp.setlayervisibility (layer1_name, 0) )
            
            vecpl = buvector_create ()
            buvector_append (vecpl, -1, 0.1, 0.6)
            buvector_append (vecpl, -1, 1.9, 0.4)
            buvector_append (vecpl, -1, 0.3, 0.9)
            buvector_append (vecpl, -1, 0.4, 0.8)
            buvector_append (vecpl, -1, 0.1, 0.6)
            
            polyline2 = Polyline (vp)
            self.assertTrue(polyline2 != None)
            
            polyline2.configure ( \
                    "vector",         vecpl, \
                    "color_fill",         "#80ffff00", \
                    "color_outline","red", \
                    "linewidth",        2 )
            
            
            vecpp = buvector_create ()
            buvector_append (vecpp, -1, 0.1, 0.6, "Apt {o=black f=white n=red sz=30 sy=bb,1,2,3,4,5,6}")
            buvector_append (vecpp, -1, 1.9, 0.4, "Bpt {f=#400000b0 sz=30 sy=square}")
            buvector_append (vecpp, -1, 0.3, 0.9, "Cpt {f=white o=#00ff00 sz=20 sy=star}")
            buvector_append (vecpp, -1, 0.4, 0.8, "Dpt {f=black o=orange sz=15}")
            
            polypoint = Polypoint (vp)
            self.assertTrue(polypoint != None)

            mycallback = Mycallback (vp, polyline2, vecpl)
            polypoint.registercallback ("Mycallback.mycallback", '%d' % mycallback.instance)

            polypoint.configure ( \
                    "vector",        vecpp, \
                    "symbol",        "triangle",  \
                    "color_fill",        "white", \
                    "color_outline","green", \
                    "font",                "Helvetica 12 Bold", \
                    "color_text",        "brown", \
                    "linewidth",        2, \
                    "size",                50, \
                    "interaction",        "move" )
            
            layer2_name = polyline2.setlayer ("layer2")
            polypoint.setlayer ("layer2")

            layer2_on_item = view_menu.additem ("&Show layer2")
            layer2_on_item.registercallback ( lambda self: vp.setlayervisibility (layer2_name, 1) )
            layer2_off_item = view_menu.additem ("&Hide layer2")
            layer2_off_item.registercallback ( lambda self: vp.setlayervisibility (layer2_name, 0) )

            mainwindow.geometry(1000, 1000, 0)
            mainwindow.show ()
            ge.qtmainloop ()
            ge.pymainloop()

    server = Testbqplot_fixture()
    server.start()
    suite = unittest.makeSuite(Testbqplot)
    runner = unittest.TextTestRunner()
    runner.run(suite)
    server.stop()

    bqplot_pymainloop()
