# Copyright (c) 2016 Boulder Real Time Technologies, Inc.
# All rights reserved 

# This software module is wholly owned by Boulder Real Time 
# Technologies, Inc. Any use of this software module without
# express written permission from Boulder Real Time Technologies,
# Inc. is prohibited.

import os
import sys
sys.path.append(os.environ['ANTELOPE'] + "/data/python")

from antelope import _ev
from antelope import datascope
from antelope.pfsubs import *


def ev_server_create(dbname, age_ref_now=1, max_events=-1):
    """Make an event view server object"""

    return _ev._ev_server_create(dbname, age_ref_now, max_events)


def ev_server_events_auto_refresh(evserver_handle, interval):
    """Automatic server refresh events list"""

    return _ev._ev_server_events_auto_refresh(evserver_handle, interval)

def ev_server_events_auto_checkdb(evserver_handle, interval):
    """Automatic check database"""

    return _ev._ev_server_events_auto_checkdb(evserver_handle, interval)


def ev_server_release_lock(evclient_handle):
    """Release server mutex on events vector"""

    return _ev._ev_server_release_lock(evclient_handle)


def ev_server_grab_lock(evclient_handle):
    """Grab server mutex on events vector"""

    return _ev._ev_server_grab_lock(evclient_handle)


def ev_server_events_refresh(evclient_handle):
    """Cause events list refresh in server"""

    return _ev._ev_server_events_refresh(evclient_handle)


def ev_server_events_number(evclient_handle):
    """Return current number of events in server events list"""

    return _ev._ev_server_events_number(evclient_handle)


def ev_server_events_size(evclient_handle):
    """Return current byte size of events vector in server events list"""

    return _ev._ev_server_events_size(evclient_handle)


def ev_server_events_get(evclient_handle):
    """Get all current events from server"""

    return _ev._ev_server_events_get(evclient_handle)


def ev_server_detections_number(evclient_handle):
    """Return current number of detections in server events list"""

    return _ev._ev_server_detections_number(evclient_handle)


def ev_server_detections_size(evclient_handle):
    """Return current byte size of detections vector in server events list"""

    return _ev._ev_server_detections_size(evclient_handle)


def ev_server_detections_get(evclient_handle):
    """Get all current detections from server"""

    return _ev._ev_server_detections_get(evclient_handle)


def ev_server_last_modified_event_get(evclient_handle):
    """Get most recently modified event from server"""

    return _ev._ev_server_last_modified_event_get(evclient_handle)


def ev_server_last_modified_event_get_evid(evclient_handle):
    """Get evid of most recently modified event from server"""

    return _ev._ev_server_last_modified_event_get_evid(evclient_handle)

def ev_server_most_recent_event_get(evclient_handle):
    """Get most recent event from server"""

    return _ev._ev_server_most_recent_event_get(evclient_handle)


def ev_server_most_recent_event_get_evid(evclient_handle):
    """Get evid of most recent event from server"""

    return _ev._ev_server_most_recent_event_get_evid(evclient_handle)


def ev_server_expanded_events_number(evclient_handle):
    """Return current number expanded of events in server events list"""

    return _ev._ev_server_expanded_events_number(evclient_handle)


def ev_server_expanded_events_size(evclient_handle):
    """Return current byte size of expanded events vector in server events list"""

    return _ev._ev_server_expanded_events_size(evclient_handle)


def ev_client_create(evserver_handle):
    """Make an event view client object"""

    return _ev._ev_client_create(evserver_handle)


def ev_client_register_callback(evclient_handle, callback, instance):
    """Define event view client object callback"""

    return _ev._ev_client_register_callback(evclient_handle, callback, instance)


def ev_client_get_dbptr(evclient_handle, object, table_name):
    """Get datascope database pointer"""

    return _ev._ev_client_get_dbptr(evclient_handle, object, table_name)


def ev_events_number(evevents_handle):
    """Return number of events"""

    return _ev._ev_events_number(evevents_handle)


def ev_events_size(evevents_handle):
    """Return byte size of events vector"""

    return _ev._ev_events_size(evevents_handle)


def ev_events_free(evevents_handle):
    """Free an events vector object"""

    return _ev._ev_events_free(evevents_handle)


def ev_events_get_index(evevents_handle, row):
    """Get index from a row for an event view"""

    return _ev._ev_events_get_index(evevents_handle, row)


def ev_events_get_row(evevents_handle, index):
    """Get for from an index for an event view"""

    return _ev._ev_events_get_row(evevents_handle, index)


def ev_events_get_value(evclient_handle, evevents_handle, query):
    """Get a value from an event view"""

    return _ev._ev_events_get_value(evclient_handle, evevents_handle, query)


def ev_events_get_value_function(evclient_handle, evevents_handle, function, index, index_is_row=1):
    """Get a value from an event view"""

    return _ev._ev_events_get_value_function(evclient_handle, evevents_handle, function, index, index_is_row)


def ev_events_get_evid(evevents_handle, index, index_is_row=1):
    """Get evid from an events view vector index or row"""

    return _ev._ev_events_get_evid(evevents_handle, index, index_is_row)


def ev_events_set_sorttab(evclient_handle, evevents_handle, query=None, reverse_flag=0):
    """Set sorting tabs for all events"""

    if query is not None:
        return _ev._ev_events_set_sorttab(evclient_handle, evevents_handle, query, reverse_flag)
    else:
        return _ev._ev_events_set_sorttab(evclient_handle, evevents_handle)


def ev_events_evid_to_index(evevents_handle, evid):
    """Return events vector index corresponding to evid"""

    return _ev._ev_events_evid_to_index(evevents_handle, evid)


def ev_events_evid_to_row(evevents_handl, evid):
    """Return events vector row corresponding to evid"""

    return _ev._ev_events_evid_to_row(evevents_handl, evid)


def ev_events_origins_get(evevents_handl, index, index_is_row=1):
    """Get origins vector from event"""

    return _ev._ev_events_origins_get(evevents_handl, index, index_is_row)


def ev_origins_get_index(evorigins_handle, row):
    """Get index from a row for an origins view"""

    return _ev._ev_origins_get_index(evorigins_handle, row)


def ev_origins_get_row(evorigins_handle, index):
    """Get for from an index for an origins view"""

    return _ev._ev_origins_get_row(evorigins_handle, index)


def ev_origins_get_orid(evorigins_handle, index, index_is_row=1):
    """Get orid from an origins view vector index or row"""

    return _ev._ev_origins_get_orid(evorigins_handle, index, index_is_row)


def ev_origins_get_value(evclient_handle, evorigins_handle, query):
    """Get a value from an origins list"""

    return _ev._ev_origins_get_value(evclient_handle, evorigins_handle, query)


def ev_origins_orid_to_index(evorigins_handl, orid):
    """Return origins vector index corresponding to orid"""

    return _ev._ev_origins_orid_to_index(evorigins_handl, orid)


def ev_origins_orid_to_row(evorigins_handl, orid):
    """Return origins vector row corresponding to orid"""

    return _ev._ev_origins_orid_to_row(evorigins_handl, orid)


def ev_origins_free(evorigins_handle):
    """Free an origins vector object"""

    return _ev._ev_origins_free(evorigins_handle)


def ev_origins_set_sorttab(evclient_handle, evorigins_handle, query=None, reverse_flag=0):
    """Set sorting tabs for all origins"""

    if query is not None:
        return _ev._ev_origins_set_sorttab(evclient_handle, evorigins_handle, query, reverse_flag)
    else:
        return _ev._ev_origins_set_sorttab(evclient_handle, evorigins_handle)


def ev_event_get(evevents_handle, index, index_is_row=1):
    """Get an event object"""

    return _ev._ev_event_get(evevents_handle, index, index_is_row)


def ev_event_get_dict(evevents_handle, index, index_is_row=1):
    """Get an event object"""

    return _ev._ev_event_get_dict(evevents_handle, index, index_is_row)


def ev_event_free(evevent_handle):
    """Free an event object"""

    return _ev._ev_event_free(evevent_handle)


def ev_detections_number(evdetections_handle):
    """Return number of detections"""

    return _ev._ev_detections_number(evdetections_handle)


def ev_detections_size(evdetections_handle):
    """Return byte size of detections vector"""

    return _ev._ev_detections_size(evdetections_handle)


def ev_detections_free(evdetections_handle):
    """Free a detections vector object"""

    return _ev._ev_detections_free(evdetections_handle)


def ev_detection_get(evdetections_handle, index):
    """Get a detection"""

    return _ev._ev_detection_get(evdetections_handle, index)


def ev_print(object1, object2=None):
    """Call event view diagnostic print methods"""

    if object2 == None:
        return _ev._ev_print(object1)
    else:
        return _ev._ev_print(object1, object2)


class EVServer (object):
    instances = {}

    def __init__(self, dbname, age_ref_now=0, max_events=-1, foo=1):
        self.handle = _ev._ev_server_create (dbname, age_ref_now, max_events, foo)
        if self.handle == None:
            raise Exception ('ev_server_create error.')
        EVServer.instances[self.handle] = self

    def gethandle(self):
        return self.handle

    def set_maxevents(self, maxevents=-1):
        _ev._ev_server_set_maxevents(self.handle, maxevents)

    def refresh(self, force=0):
        _ev._ev_server_events_refresh(self.handle, force)

    def auto_refresh(self, interval):
        _ev._ev_server_events_auto_refresh(self.handle, interval)

    def auto_checkdb(self, interval):
        _ev._ev_server_events_auto_checkdb(self.handle, interval)


class EVClient(object):
    instances = {}

    def __init__(self, evserver):
        self.handle = _ev._ev_client_create (evserver.gethandle())
        EVClient.instances[self.handle] = self

    def gethandle(self):
        return self.handle

    def grab_lock(self):
        _ev._ev_server_grab_lock(self.handle)

    def server_events_size(self):
        return _ev._ev_server_events_size(self.handle)

    def server_detections_size(self):
        return _ev._ev_server_detections_size(self.handle)

    def server_expanded_events_size(self):
        return _ev._ev_server_expanded_events_size(self.handle)

    def server_events_number(self):
        return _ev._ev_server_events_number(self.handle)

    def server_detections_number(self):
        return _ev._ev_server_detections_number(self.handle)

    def server_expanded_events_number(self):
        return _ev._ev_server_expanded_events_number(self.handle)

    def release_lock(self):
        _ev._ev_server_release_lock(self.handle)

    def events_get(self):
        return _ev._ev_server_events_get(self.handle)

    def detections_get(self):
        return _ev._ev_server_detections_get(self.handle)

    def last_modified_event_get(self):
        return _ev._ev_server_last_modified_event_get(self.handle)

    def last_modified_event_get_evid(self):
        return _ev._ev_server_last_modified_event_get_evid(self.handle)

    def most_recent_event_get(self):
        return _ev._ev_server_most_recent_event_get(self.handle)

    def most_recent_event_get_evid(self):
        return _ev._ev_server_most_recent_event_get_evid(self.handle)

    def events_get_index(self, evevents_handle, row):
        return _ev._ev_events_get_index(evevents_handle, row)

    def events_get_row(self, evevents_handle, index):
        return _ev._ev_events_get_row(evevents_handle, index)

    def get_value(self, evevents_handle, query):
        return _ev._ev_events_get_value(self.handle, evevents_handle, query)

    def registercallback(self, callback, client_string='0'):
        return _ev._ev_client_register_callback(self.handle, callback, client_string)

    def refresh(self, force=0):
        return _ev._ev_server_events_refresh(self.handle, force)