/* Copyright (c) 2015 Boulder Real Time Technologies, Inc. */
/* All rights reserved */

/* This software module is wholly owned by Boulder Real Time 
	Technologies, Inc. Any use of this software module without
	express written permission from Boulder Real Time Technologies,
	Inc. is prohibited. */

#ifndef __BQ__
#define __BQ__

#include <stddef.h>
#include <cstdarg>
#include <functional>

#include <QtWidgets>
#include <QPrinter>
#include <QPrintDialog>
#include <QPageSetupDialog>

#include "brttutil.h"
#include "coords.h"
#include "stock.h"
#include "BUConfigure.h"
#include "BUMapData.h"
#include "EV.h"
#include "BUPf.h"

#define BQ_ClientData 	BU_ClientData
#define BQ_ConfigSpec 	BU_ConfigureSpec
#define BQ_CustomOption BU_CustomOption
#define BQ_Offset 	BU_Offset
#define BQ_CONFIGURE_OK 	BU_CONFIGURE_OK
#define BQ_CONFIGURE_ERROR 	BU_CONFIGURE_ERROR
#define BQ_CONFIGURE_NULL_OK 	BU_CONFIGURE_NULL_OK
#define BQ_CONFIGURE_DONT_SET_DEFAULT 	BU_CONFIGURE_DONT_SET_DEFAULT
#define BQ_CONFIGURE_OPTION_SPECIFIED 	BU_CONFIGURE_OPTION_SET
#define BQ_CONFIGURE_BOOLEAN	"boolean"
#define BQ_CONFIGURE_INT	"int"
#define BQ_CONFIGURE_DOUBLE	"double"
#define BQ_CONFIGURE_STRING	"string"
#define BQ_CONFIGURE_PTR	"ptr"
#define BQ_CONFIGURE_CUSTOM	"custom"
#define BQ_CONFIGURE_COLOR	"color"
#define BQ_CONFIGURE_FONT	"font"
#define BQ_CONFIGURE_END	"end"
#define BQ_SetName		BU_SetName
#define BQ_GetName		BU_GetName
#define BQ_GetObject		BU_GetObject
#define BQ_ParseBoolean		BU_ParseBoolean

#define BQ_register_error	bu_register_error
#define BQ_complain		bu_complain
#define BQ_die			bu_die
#define BQ_clear_register	bu_clear_register
#define BQ_elog_string		bu_elog_string
#define BQ_ptimes		bu_ptimes

#define BQ_QString2char		toLocal8Bit().data()

#define BQ_ANCHOR_SOUTHWEST		0
#define BQ_ANCHOR_SOUTHCENTER	1
#define BQ_ANCHOR_SOUTHEAST		2
#define BQ_ANCHOR_EASTCENTER	3
#define BQ_ANCHOR_CENTER		4
#define BQ_ANCHOR_NORTHEAST		5
#define BQ_ANCHOR_NORTHCENTER	6
#define BQ_ANCHOR_NORTHWEST		7
#define BQ_ANCHOR_WESTCENTER	8

#define BQ_FLOAT_NULL		(2.12345e30)
#define BQ_FLOAT_NULL_STR	"2.12345e30"
#define BQ_DEFAULT_CAP_STYLE	CapButt
#define BQ_DEFAULT_JOIN_STYLE	JoinRound
#define BQ_DEFAULT_LINE_STYLE	LineSolid
#define BQ_X2PS_FACTOR		(10.0)
#define BQ_MINDX		"200"
#define BQ_MINDY		"100"

#define BQ_MODE_LIN		(0)	/* linear plot mode */
#define BQ_MODE_LOG		(1)	/* logarithmic plot mode */
#define BQ_MODE_TIME	(2)	/* time plot mode */

#define BQ_POINT_MODE_WORLD_COORDINATES	(0)	/* plot point expressed in world coordinates */
#define BQ_POINT_MODE_PIXEL_COORDINATES	(1)	/* plot point expressed in pixel coordinates */
#define BQ_POINT_MODE_VIEWPORT_COORDINATES (2)	/* plot point expressed in normalized viewport coordinates */

#define BQ_MAP_TRAN_NONE	(0)	/* No map coordinate transformations */
#define BQ_MAP_TRAN_EDP		(1)	/* Equal distance projection map coordinate transformations */
#define BQ_MAP_TRAN_GLOBE	(2)	/* Equal distance projection map coordinate transformations */
#define BQ_MAP_TRAN_MERC	(3)	/* Mercator projection map coordinate transformations (wrapping longitude) */
#define BQ_MAP_TRAN_GEOG	(4)	/* geographic map coordinates (wrapping longitude) */

#define BQ_QEVENT_TYPE_MOUSEPRESS	1
#define BQ_QEVENT_TYPE_MOUSEMOVE	2
#define BQ_QEVENT_TYPE_MOUSERELEASE	3
#define BQ_QEVENT_TYPE_KEYPRESS		4
#define BQ_QEVENT_TYPE_KEYRELEASE	5
#define BQ_QEVENT_TYPE_WHEEL		6
#define BQ_QEVENT_TYPE_ENTER		7
#define BQ_QEVENT_TYPE_LEAVE		8
#define BQ_QEVENT_TYPE_RESIZE		9
#define BQ_QEVENT_TYPE_PAINT		10

#define BQ_QEVENT_STATE_INITIAL		1
#define BQ_QEVENT_STATE_PAN		2
#define BQ_QEVENT_STATE_ZOOMIN		3
#define BQ_QEVENT_STATE_ZOOMOUT		4
#define BQ_QEVENT_STATE_RLATLON		5
#define BQ_QEVENT_STATE_DELTIME		6

#define BQ_QEVENT_RETURN_OK		1
#define BQ_QEVENT_RETURN_DONT_PROPAGATE	2

#define BQ_INTERACTION_MODE_NONE	0
#define BQ_INTERACTION_MODE_SHOW	1
#define BQ_INTERACTION_MODE_SHOW_WITHOUT_SIZE_INCREASE	2
#define BQ_INTERACTION_MODE_MOVE	3
#define BQ_INTERACTION_MODE_SELECTED	4

#define BQ_XPIXEL(O,X) ((X)-O->viewport_parent->vars.xleft)*(O->viewport_parent->vars.w - O->viewport_parent->vars.mleft - O->viewport_parent->vars.mright)/(O->viewport_parent->vars.xright-O->viewport_parent->vars.xleft)
#define BQ_XPIXELV(V,X) ((X)-V->vars.xleft)*(V->vars.w - V->vars.mleft - V->vars.mright)/(V->vars.xright-V->vars.xleft)

#define	BQ_MERCY(lat)	(((lat) >= 90.0) ? (1.e5) : (((lat) <= -90.0) ? (-1.e5) \
			: (180.0*log(tan((0.5*M_PI*(lat)/180.0)+0.25*M_PI))/M_PI)))

#define BQ_IMERCLAT(y)	(180.0*(2.0 * (atan(exp((y)*M_PI/180.0)) - 0.25*M_PI))/M_PI)

#define INT(x)  (int)((x)<0.0?((x)-0.5):((x)+0.5))
#define SAMP(x,y)       (int)((x)<0.0?((x)/(y)-1.0):((x)/(y)))

struct BQ_Color {
	QString color_name;
	QColor color;
};

struct BQ_Font {
	BQ_Font () {
		memset (this, 0, BQ_Offset (BQ_Font, font_name));
		invalid = 1;
	}

	int size;
	int weight;
	int italic;
	int invalid;
	QString font_name;
	QString family;
	QFont font;
};

extern int BQ_ParseColor (QString value, BQ_Color &color);
extern void BQ_ClearColor (BQ_Color &color);
extern int BQ_ParseFont (QString value, BQ_Font &font);
extern int BQ_SetFontSize (int size, BQ_Font &font);

extern int bqplot_debug;

struct BQ_XY {
	BQ_XY () {
		x = BQ_FLOAT_NULL;
		y = BQ_FLOAT_NULL;
	}
	BQ_XY (double x, double y) {
		this->x = x;
		this->y = y;
	}

	double x, y;
};

struct BQ_XYL {
	BQ_XYL () {
		x = BQ_FLOAT_NULL;
		y = BQ_FLOAT_NULL;
		clientData = NULL;
	}
	BQ_XYL (double x, double y, QString label) {
		this->x = x;
		this->y = y;
		this->label = label;
	}
	BQ_XYL (double x, double y, QString label, BQ_ClientData clientData) {
		this->x = x;
		this->y = y;
		this->label = label;
		this->clientData = clientData;
	}

	double x, y;
	QString label;
	BQ_ClientData clientData;
};

class BQConfigure : public BUConfigure {
public:
	BQConfigure ();
};

class BQViewportItem;
class BQViewport;
class BQLayer;
class BQLogo;
class BQGraphicsContext;
class BQGlyph;
class BQShape;
class BQGlyphs;
class BQGlyphsGraphicsContext;
class BQAxes;
class BQText;
class BQTextPopup;
class BQTaskbar;
class BQPolyline;
class BQPolypoint;
class BQMap;
class BQNavigator;
class BQMapevents;
class BQMainWindow;
class BQTraceview;
class BQTraceviewCommands;
class BQTrace;
class BQTableview;
class BQEVTableview;
class BQEVEventsTableview;
class BQEVOriginsTableview;

class BQViewportItem {
	friend class BQViewport;
	friend class BQLayer;

public:
	BQViewportItem (BQViewport *master, QWidget *widget, QString class_name, int register_item=1);
	~BQViewportItem ();
	QString getFullLayerName (QString name);
	QString setLayer (QString name);
	void setItemStale();
	void setItemName(QString name);
	void setAntialias(bool antialias);
	void showNames (QString *string, QString *lname=NULL);
	void removeLayer ();
	bool hasLayer();
	BQLayer *getLayer();
	QString getItemName();
	QString getLayerName();
	BQViewport *getViewportParent();
	void showItem();
	void hideItem();
	bool isShown();
	void processItem();
	void ignoreItem();
	void moveViewport (BQViewport *viewport);

	int order;
	int item_process;
	int item_display;

protected:
	virtual int processEvent (int type, QEvent *event);
	virtual int drawViewportItem (QPainter *painter) = 0;
	virtual void setVisibility () = 0;
	QPainter *getPainter(QString *layer_name = NULL);

	QWidget *widget;
	BQViewport *viewport_parent;
	QString layer_name;
	QString class_name;
	QString item_name;
	bool antialias;
	static int next_order;

private:
	int noderegister;
};

class BQLayer : public BQViewportItem {
	friend class BQViewportItem;

public:
	BQLayer (BQViewport *master, QString name);
	void setLayerSize (int width, int height, int *stale);
	void clearLayer();
	void setStale();
	void maskColorTransparent();
	void maskColorTransparent (QColor *mask_color);
	void setShift (int xshift, int yshift);
	QColor *getMaskColor();
	QPixmap *getPixmap();

	int stale;

	static BQLayer *getLayer (QString name);
	static void setLayerVisibility (QString name, int visibility=1);
	static int getRatio ();

protected:
	int drawViewportItem (QPainter *painter);
	void removeItem (BQViewportItem *item);
	void addItem (BQViewportItem *item);
	void setVisibility ();
	void setRatio (QWidget *widget);

private:
	static QHash<QString, BQLayer *> layers;
	static int ratio;

	int xshift=0, yshift=0;
	int w, h;
	QPixmap *pixmap;
	QColor *mask_color;
	QList<BQViewportItem *> display_list;
	int visibility;
};

class BQEvent {
public:
	BQEvent (int type, QEvent *event, QWidget *widget=NULL);

	void print();

	int type;
	int button;
	int modifiers;
	QString key;
	double delta;
	double deltax;
	QPointF pt;
	QPointF gpt;
};

class BQEVCallbackEvent : public QEvent {
public:
	BQEVCallbackEvent () : QEvent ( registeredType() ) { }
	static QEvent::Type eventType;

	EVResponse response;
	BQMapevents *bqme=NULL;
	BQTraceview *bqtr=NULL;
	Array<std::string,int> changed;

private:
	static QEvent::Type registeredType();
};

typedef void (BQ_paintCallback) (void *item, int type, QResizeEvent *event, BQ_ClientData client_data);
typedef void (BQ_eventCallback) (void *item, BQEvent *event, BQ_ClientData client_data);
typedef void (BQ_resizeCallback) (void *item, QResizeEvent *event, BQ_ClientData client_data);

struct BqViewportPaintCallback {
	BqViewportPaintCallback () {
		memset (this, 0, sizeof(BqViewportPaintCallback));
	}

	std::function<BQ_paintCallback> callback_proc;
	void *item;
	BQ_ClientData client_data;
	int stale;
};

struct BqViewportEventCallback {
	BqViewportEventCallback () {
		memset (this, 0, sizeof(BqViewportEventCallback));
	}

	std::function<BQ_eventCallback> callback_proc;
	void *item;
	BQ_ClientData client_data;
};

struct BqViewportResizeCallback {
	BqViewportResizeCallback () {
		memset (this, 0, sizeof(BqViewportResizeCallback));
	}

	std::function<BQ_resizeCallback> callback_proc;
	void *item;
	BQ_ClientData client_data;
};

struct BqViewport {
	BqViewport () {
		memset (this, 0, BQ_Offset (BqViewport, viewport_name));
		cursor_opacity = 0x80;
	}

	int w, h;
	int spc_w, spc_h;
	int last_w, last_h;
	int mleft, mright, mbottom, mtop;
	double opacity;
	int preserve_xscale;
	double xscale_anchor;
	int preserve_yscale;
	double yscale_anchor;
	int center_yscale;
	double constant_aspect;
	double xleft, xright, ybottom, ytop;
	double spc_xleft, spc_xright, spc_ybottom, spc_ytop;
	int xoffset, yoffset;
	int xgain_anchor, ygain_anchor;
	double xgain, ygain;
	int xmode, ymode;
	double xmin, xmax, ymin, ymax;
	int map_tran;
	double latr, lonr, range;
	double latl, lonl;
	double bblatmin, bblatmax;
	double bblonmin, bblonmax;
	int qevent_state;
	QPointF mouse_start;
	QPointF mouse_start_wc;
	QRubberBand *rubberband;
	BQLogo *logo;
	QResizeEvent *resize_event;
	int mouse_timer_event;
	int wheel_timer_event;
	int default_event_processing;
	QWidget *parent;
	int make_logo;
	BQViewportItem *item_grab;
	QGridLayout *layout;
	QGridLayout *layout_parent;
	int debug;
	int disable_layers;
	unsigned char cursor_opacity;

	QString viewport_name;
	BQ_Color background;
    BQ_Color background_frame;
	QList<BQViewportItem *> display_list;
	QList<BqViewportPaintCallback> paint_callbacks;
	QList<BqViewportEventCallback> event_callbacks;
	QList<BqViewportResizeCallback> resize_callbacks;
};

class BQViewport : public QWidget, public BQViewportItem, public BQConfigure {
public:
	BQViewport(QWidget *parent, int make_parent_layout=1, int make_logo=1);
	~BQViewport();

	BqViewport vars;

	static int viewport_next_instance;

	void registerItem (BQViewportItem *item);
	void deregisterItem (BQViewportItem *item);
	void removeItem (BQViewportItem *item);
	void addItemAfter (BQViewportItem *item, BQViewportItem *after_item);
	void moveItemAfter (BQViewportItem *item, BQViewportItem *after_item);
	void moveItemToTop (BQViewportItem *item);
	void moveItemToBottom (BQViewportItem *item);
	void showDisplayList (const char *prefix=NULL);
	void grabEventStream (BQViewportItem *item);
	void ungrabEventStream ();
	void setPaintCallback (std::function<BQ_paintCallback> paintCallback, void *item, BQ_ClientData client_data);
	void removePaintCallback (std::function<BQ_paintCallback> paintCallback);
	void setEventCallback (std::function<BQ_eventCallback> eventCallback, void *item, BQ_ClientData client_data);
	void setResizeCallback (std::function<BQ_resizeCallback> resizeCallback, void *item, BQ_ClientData client_data);

	void setWcoordsX (double xleft, double xright);
	void setWcoordsY (double ybottom, double ytop);
	void setScales ();
	int getScales (int *w, int *h,
				double *xstart, double *xend, double *xscale, 
				double *ystart, double *yend, double *yscale);
	int getWcoords (double x, double y, double *xcoord, double *ycoord, int no_map_tran=0, int no_pscale=0, int vp_input=0);
	QPainterPath *getWcoordsBB (int nside, int *np=NULL, int *sp=NULL);
	int getPcoords (double xcoord, double ycoord, double *x, double *y, int no_map_tran=0, int no_pscale=0);
	int getPcoordX (double xcoord, double *x);
	int getPcoordY (double ycoord, double *y);
	int translateAndGain ();
	void bufferDraws (QPainter *painter, int fill, int flush, int w, int h, double fx1, double fy1, double fx2, double fy2);
	int getPixels (char *xstr, char *ystr, double *xpixels, double *ypixels, void *relative=NULL);
	int getPixels (char *str, int *pixels);
	void drawText (QPainter *painter, 
			BQ_Font *font, double x, double y, char *just, double angle, char *text,
			int show_not_symbol=0, int background_width=0, int background_height=0, 
			BQ_Color *text_color=NULL, BQ_Color *background_color=NULL, BQ_Color *background_outline_color=NULL,
			QPainterPath **path=NULL, double *xbb=NULL, double *ybb=NULL, int *wbb=NULL, int *hbb=NULL,
			int clip=1);
	int polyDraw (QPainter *painter, int npts, double *xpts, double *ypts, int fill, int complete, int ignoretran);
	int drawXTimeAxis (QPainter *painter, BQ_Font *font, 
			BQ_Color *linecolor, int linewidth, int linestyle,
			BQ_Color *fontcolor, char *label_time_format, char *label_date_format);
	void computeMapBB ();
	void setDefaultCursor (int force=0);
	void zoom (double xw, double yw, double xfactor, double factor);
	void center (double xw, double yw);
	int processEvent (int type, QEvent *event);
	void setStale ();
	QPainterPath *wrapLongitudes (int npts, double *lons, double *lats, int np, int sp);
	QPainterPath *wrapLongitudes (QPainterPath *path, int np, int sp);

	static int getScales (BqViewport *vp, int *w, int *h,
				double *xstart, double *xend, double *xscale, 
				double *ystart, double *yend, double *yscale);
	static int getWcoords (BqViewport *vp, double x, double y, double *xcoord, double *ycoord, int no_map_tran=0, int no_pscale=0, int vp_input=0);
	static int getPcoords (BqViewport *vp, double xcoord, double ycoord, double *x, double *y, int no_map_tran=0, int no_pscale=0);
	static int getPcoordX (BqViewport *vp, double xcoord, double *x);
	static int getPcoordY (BqViewport *vp, double ycoord, double *y);
	static int translateAndGain (BqViewport *vp);
	static QPainterPath *wrapLongitudes (BqViewport *vp, int npts, double *lons, double *lats, int np, int sp);
	static QPainterPath *wrapLongitudes (BqViewport *vp, QPainterPath *path, int np, int sp);
	static void parseAnchor (char *anchor, double *jx, double *jy);
	static void parseAnchor (int anchor, double *jx, double *jy);

protected:
	void paintEvent (QPaintEvent *event);
	void mousePressEvent (QMouseEvent *event);
	void mouseMoveEvent (QMouseEvent *event);
	void mouseReleaseEvent (QMouseEvent *event);
	void wheelEvent (QWheelEvent *event);
	void keyPressEvent (QKeyEvent *event);
	void keyReleaseEvent (QKeyEvent *event);
	void resizeEvent (QResizeEvent *event);
	void enterEvent (QEvent *event);
	void leaveEvent (QEvent *event);
	bool event(QEvent *event);

	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	void paintCallback ();
	void eventCallback (int type, QEvent *event);
	void resizeCallback ();
	void processevent_pvt (int type, QEvent *event);
	void limitBB (double *x, double *y);
	QRectF boundingRect() const;
	static int setFpoints (int fnpts, int *fpoints_size, QPointF ** fpoints);
	int getXPixel (char *xstr, double *xpixels, void *relative);
	int getYPixel (char *ystr, double *ypixels, void *relative);
	double last_update_time = 0.0;
};

#define BQ_GRAPHICS_CONTEXT_TYPE_CLEAR_PEN		0
#define BQ_GRAPHICS_CONTEXT_TYPE_COLOR_PEN		1
#define BQ_GRAPHICS_CONTEXT_TYPE_LINESTYLE_PEN		2
#define BQ_GRAPHICS_CONTEXT_TYPE_JOINSTYLE_PEN		3
#define BQ_GRAPHICS_CONTEXT_TYPE_CAPSTYLE_PEN		4
#define BQ_GRAPHICS_CONTEXT_TYPE_CLEAR_BRUSH		5
#define BQ_GRAPHICS_CONTEXT_TYPE_COLOR_BRUSH		6
#define BQ_GRAPHICS_CONTEXT_TYPE_LINEWIDTH		7
#define BQ_GRAPHICS_CONTEXT_TYPE_LINEWIDTHF		8
#define BQ_GRAPHICS_CONTEXT_TYPE_OPACITY		9
#define BQ_GRAPHICS_CONTEXT_TYPE_LABEL			10
#define BQ_GRAPHICS_CONTEXT_TYPE_FONT_LABEL		11
#define BQ_GRAPHICS_CONTEXT_TYPE_COLOR_LABEL		12
#define BQ_GRAPHICS_CONTEXT_TYPE_SIZE_LABEL		13
#define BQ_GRAPHICS_CONTEXT_TYPE_SYMBOL			14
#define BQ_GRAPHICS_CONTEXT_TYPE_SIZE			15
#define BQ_GRAPHICS_CONTEXT_TYPE_LINESTYLEC1_PEN	16
#define BQ_GRAPHICS_CONTEXT_TYPE_LINESTYLEC2_PEN	17

#define BQ_GRAPHICS_CONTEXT_CODING_TYPE_NONE		0
#define BQ_GRAPHICS_CONTEXT_CODING_TYPE_SCALAR		1
#define BQ_GRAPHICS_CONTEXT_CODING_TYPE_COLOR_HSL	2

#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_DEFAULT		(0)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_POINT		(1)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_CROSS		(2)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_X		(3)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_TRIANGLE	(4)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_SQUARE		(5)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_CIRCLE		(6)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_STAR		(7)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_DIAMOND		(8)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_BEACHBALL	(9)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_SHAPE		(10)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_VERTICAL_LINE (11)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_HORIZONTAL_LINE (12)

struct BqGraphicsContextCoding {
	BqGraphicsContextCoding () {
		memset (this, 0, sizeof (BqGraphicsContextCoding));
	}

	double in_0;
	double out_0;
	double out_scale;
	double out_min;
	double out_max;
};

struct BqGraphicsContextItemParams {
	BqGraphicsContextItemParams () {
		memset (this, 0, BQ_Offset (BqGraphicsContextItemParams, name));
	}
	~BqGraphicsContextItemParams () {
		QColor *color;
		QFont *font;
		if (free_flag) return;
		if (coding_type == BQ_GRAPHICS_CONTEXT_CODING_TYPE_COLOR_HSL) {
			color = (QColor *) coding_out;
			if (color) delete color;
		}
		switch (type) {
		case BQ_GRAPHICS_CONTEXT_TYPE_COLOR_PEN:
		case BQ_GRAPHICS_CONTEXT_TYPE_COLOR_BRUSH:
		case BQ_GRAPHICS_CONTEXT_TYPE_COLOR_LABEL:
			color = (QColor *) value;
			if (color) delete color;
			break;
		default:
			break;
		}
	}

	int type;
	void *value;
	int coding_type;
	void *coding_out;
	BqGraphicsContextCoding *coding;
	int free_flag;
	QString name;
};

struct BqGraphicsContextItem {
	BqGraphicsContextItem () {
		memset (this, 0, sizeof(BqGraphicsContextItem));
		coding_input = BQ_FLOAT_NULL;
	}
	~BqGraphicsContextItem () {
		if (item && item->free_flag) delete item;
	}

	BqGraphicsContextItemParams *item;
	float coding_input;
};

typedef QList<BqGraphicsContextItem> BqGraphicsContext;

class BQGraphicsContext {
public:
	BQGraphicsContext ();
	BQGraphicsContext (BQGraphicsContext *source);
	void copy (BQGraphicsContext *source);
	BqGraphicsContextItem *setGcItem (QString name, int type, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, QColor param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, QFont *param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, Qt::PenJoinStyle style, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, Qt::PenStyle style, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, long param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, double param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, QString param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, char *param, int replace=0);
	int parseColorCoding (QString string, QColor **constant_color, BqGraphicsContextCoding **color_coding);
	BqGraphicsContext *getGc (QString name);
	BqGraphicsContextItem *getGcItem (QString name);
	void setGcItemCoding (QString name, int coding_type, BqGraphicsContextCoding coding);
	void setGcItemCoding (QString name, int coding_type, BqGraphicsContextCoding *coding = NULL);
	void drawGcItem (QString name, QPainter *painter);
	void drawGc (QString name, QPainter *painter);
	void drawPainterPath (QString name, QPainter *painter, QPainterPath &path);

	static BqGraphicsContext *copy (BqGraphicsContext *gc);
	static void getValue (const BqGraphicsContextItem *item, void **value);
	static void getValue (const BqGraphicsContextItem *item, int *value);
	static void getValue (const BqGraphicsContextItem *item, long *value);
	static void getValue (const BqGraphicsContextItem *item, double *value);
	static void setGcCoding (BqGraphicsContextCoding *coding, double in1, double out1, double in2, double out2, double outmin=1.0, double outmax=-1.0);
	static void setGcItemCoding (BqGraphicsContextItem *item, int coding_type, BqGraphicsContextCoding coding);
	static void setGcItemCoding (BqGraphicsContextItem *item, int coding_type, BqGraphicsContextCoding *coding = NULL);
	static void drawGcItem (const BqGraphicsContextItem *item, QPen *pen, QBrush *brush);
	static void drawGcItem (BqGraphicsContextItem *item, QPainter *painter);
	static void makeGcPenBrush (BqGraphicsContext *gc, QPen *pen, QBrush *brush);
	static void drawGc (BqGraphicsContext *gc, QPainter *painter);
	static void drawGcPainter (BqGraphicsContext *gc, QPainter *painter);
	static void drawGcPainter (BqGraphicsContext *gc, QPen *pen, QBrush *brush, QPainter *painter);
	static void drawPainterPath (BqGraphicsContext *gc, QPainter *painter, QPainterPath &path);
	static BqGraphicsContextItem *getSymbol (int symbol);
	static void initContextItem (BqGraphicsContextItem *item);

	QHash<QString, BqGraphicsContext *> mapgc;
	QHash<QString, BqGraphicsContextItem *> mapgcitem;

private:
	BqGraphicsContext *get_gc (QString name);
	BqGraphicsContextItem *get_item (QString name, int set = 1, int replace=0);
	int parseCodings (QString string, int ncodings, BqGraphicsContextCoding *color_coding);
	static void makeStaticSymbols();
	static void setStaticSymbol(QString name, int symbol_type);

	static QHash<QString, BqGraphicsContextItem *> *symbols;
};

struct BqShape {
	BqShape () {
		memset (this, 0, BQ_Offset (BqShape, paths));
	}

	QPainterPath *outline;
	QList<QPainterPath *> paths;
	QRectF bounding_rect;
};

class BQShape {
public:
	BQShape();

	static BqShape *getPoint();
	static BqShape *getCross();
	static BqShape *getX();
	static BqShape *getCircle();
	static BqShape *getSquare();
	static BqShape *getTriangle();
	static BqShape *getDiamond();
	static BqShape *getStar();
	static BqShape *getVerticalLine();
	static BqShape *getHorizontalLine();
	static BqShape *getBeachball (double m[6]);
};

#define BQ_GLYPH_GRAPHICS_CONTEXT_SYMBOL	(-5)
#define BQ_GLYPH_GRAPHICS_CONTEXT_SIZE		(-4)

typedef int (BQ_glyphsGraphicsContextCallback) (BQGlyph *glyph, QString definitions_name, QString coding_name, double *value, BQ_ClientData client_data);

struct BqGlyphsGcDefinition {
	BqGlyphsGcDefinition () {
		memset (this, 0, BQ_Offset (BqGlyphsGcDefinition, coding));
	}

	BqGraphicsContextItem *gcitem;
	QString coding;
	QString value;
};

struct BqGlyphsCodingSlot {
	BqGlyphsCodingSlot () {
		memset (this, 0, sizeof (BqGlyphsCodingSlot));
	}

	int gc;
	int gci;
};

struct BqGlyphsDefinition {
	QString name;
	BqGraphicsContextItem symbol;
	BqGlyphsGcDefinition size;
	QList<BqGlyphsGcDefinition> fill_color_items;
	QList<BqGlyphsGcDefinition> outline_linewidth_items;
	QList<BqGlyphsGcDefinition> outline_color_items;
	QList<BqGlyphsGcDefinition> opacity_items;
	QList<BqGraphicsContext *> glyph_gcs;
	QHash<QString, int> coding_defs;
	QList<QList<BqGlyphsCodingSlot>> coding_slots;
};

class BQGlyphsGraphicsContext : public BQGraphicsContext, public BQConfigure {
	friend class BQGlyphs;
	friend class BQGlyph;

public:
	BQGlyphsGraphicsContext();
	~BQGlyphsGraphicsContext();

	int getGcIndex (QString gcname);
	int getCodingIndex (int gcindex, QString codingname);
	void setCoding (int gcindex, int codingindex, double coding_input, BQGlyph *glyph);
	void getCoding (int gcindex, int codingindex, double *coding_input, BQGlyph *glyph);
	void setCodingCallback (std::function<BQ_glyphsGraphicsContextCallback> codingCallback, BQ_ClientData client_data);

	QHash<QString, BqGlyphsGcDefinition> gcitem_definitions;
	QHash<QString, BqGlyphsDefinition> gc_definitions;
	QList<BqGlyphsDefinition> gc_definitions_list;
	QHash<QString, int> gc_definitions_index;
	QList<QString> coding_names;

protected:
	void updateCoding (BQGlyph *glyph);
	BQ_ClientData codingClientData;
	std::function<BQ_glyphsGraphicsContextCallback> codingCallback;
};

struct BqPath {
	BqPath () {
		memset (this, 0, BQ_Offset (BqPath, pen));
	}

	BqGraphicsContext *gc;
	QPainterPath *path;
	QPen pen;
	QBrush brush;
};

struct BqGlyph {
	BqGlyph () {
		memset (this, 0, BQ_Offset (BqGlyph, gcs));
	}

	int index;
	void *shape;
	double xw, yw;
	double xp, yp;
	double opacity;
	int visible;
	int skip;
	int visible_label;
	char *label;
	int qevent_state;
	int freegc;
	BQGlyphs *glyphs;
	QPainterPath *label_path;
	int gcindex;
	BQ_ClientData clientData;
	int clientData_type;

	QList<BqGraphicsContext *> gcs;
	QPainterPath shape_path;
	QList<BqPath> paths;
	QList<QList<QPainterPath>> plot_paths;
	BqGraphicsContextItem symbol;
	BqGraphicsContextItem size;
};

class BQGlyph : public BQViewportItem {
	friend class BQGlyphs;
	friend class BQPolypoint;
	friend class BQGlyphsGraphicsContext;

public:
	BQGlyph(BQViewport *master, BQGlyphs *glyphs);
	~BQGlyph();
	void setGlyph (int symbol, double *params, QList<BqGraphicsContext *> gcs);
	void setGlyph (int symbol, double *params, BqGraphicsContext *gc);
	void setGlyph (QString gc_definitions_name, double *params);
	void setGlyph (int symbol, QString gc_definitions_name, double *params);
	void setGlyph (double *params);
	void setPosition (double xw, double yw);
	void getPosition (double *xw, double *yw);
	int getIndex ();
	int getQEventState ();
	void setSize (double size);
	void setCodingValue (QString name, double coding_value);
	double getCodingValue (QString name);
	bool contains (QEvent *event);
	void setClientData(BQ_ClientData client_data, int client_data_type=0);
	void getClientData(BQ_ClientData *client_data, int *client_data_type=NULL);
	BqGraphicsContextItem getGraphicsContextItem (int gcindex, int type);
	void showPaths();

	mutable BqGlyph vars;

protected:
	int processEvent (int type, QEvent *event);
	int drawViewportItem (QPainter *painter) {return (0);};
	void setVisibility () {};
	void draw(QPainter *painter);
	void drawLabel(QPainter *painter, BQ_XY &label_position);
	void drawSelected(QPainter *painter);
	void drawUnselected(QPainter *painter);
	void setCodingInput (int igc, int igci, double coding);
	void getCodingInput (int igc, int igci, double *coding);
	void setCodingInputs (int def_index, int coding_index, double coding);

private:
	void clearPaths ();
	int setTransformations (QList<BQ_XY> *label_positions = NULL) const;
	int processEventMove (int type, QEvent *event);
	int processEventShow (int type, QEvent *event);
	void setGlyph (int gc_definitions_index, double *params);
	void setGlyph (int symbol, int gc_definitions_index, double *params);
};

typedef int (BQ_glyphCallback) (BQGlyph *glyph, BQEvent *event, BQ_ClientData client_data);

struct BqGlyphCallback {
	BqGlyphCallback () {
		memset (this, 0, sizeof(BqGlyphCallback));
	}

	std::function<BQ_glyphCallback> callback_proc;
	BQ_ClientData client_data;
};

struct BqGlyphs {
	BqGlyphs () {
		memset (this, 0, BQ_Offset (BqGlyphs, glyphs));
		selection_opacity_reduction_factor = 0.5;
		selection_size_increase_factor = 1.5;
	}

	BqShape *shape_default;
	double sizep_default;
	BqGraphicsContext *gc_label;
	BqGraphicsContext *gc_label_background;
	BQGlyphsGraphicsContext *ggc;
	int selection_index;
	int interaction_mode;
	int xmode_default;
	int ymode_default;
	int shape_label_anchor_default;
	int label_anchor_default;
	int auto_cursor;
	double selection_opacity_reduction_factor;
	double selection_size_increase_factor;
	double xgain;
	double ygain;

	QList<BQGlyph *> glyphs;
	QList<BqGlyphCallback> glyphCallbacks;
};

class BQGlyphs : public BQViewportItem {
	friend class BQGlyph;
	friend class BQPolypoint;

public:
	BQGlyphs(BQViewport *master, BQGlyphsGraphicsContext *ggc = NULL, int register_item = 1);
	~BQGlyphs();
	void setShapeDefault (int symbol);
	void setSizeDefault (double sizep);
	void setXmodeDefault (int xmode);
	void setYmodeDefault (int ymode);
	void setShapeLabelAnchorDefault (int anchor);
	void setLabelAnchorDefault (int anchor);
	void setLabelGC (BqGraphicsContext *gc);
	void setLabelBackgroundGC (BqGraphicsContext *gc);
	BQGlyph *selectedGlyph();
	QList<BQGlyph *> items (QEvent *event);
	void setInteractionMode (int interaction_mode);
	void setGlyphCallback (std::function<BQ_glyphCallback> glyphCallback, BQ_ClientData client_data);
	void setGlyphCallbacks (QList<BqGlyphCallback> callbacks);
	void setGlyphs();
	void updateCoding();
	void setClientData (BQ_ClientData client_data);
	BQ_ClientData getClientData ();

	BqGlyphs vars;

protected:
	int drawViewportItem (QPainter *painter);
	int processEvent (int type, QEvent *event);
	void addGlyph (BQGlyph *glyph);

private:
	void setVisibility () {};
	BQGlyph *findClosest (QEvent *event, int selected_index = -1);

	BQ_ClientData clientData;
};

struct BqLogo {
	BqLogo () {
		memset (this, 0, sizeof (BqLogo));
	}

	double opacity;
	QImage *logo_image;
};

class BQLogo : public BQViewportItem, public BQConfigure {
public:
	BQLogo(BQViewport *master);
	void raiseLogo();

	BqLogo vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	void draw (QPainter *painter);
};

struct BqAxes {
	BqAxes () {
		memset (this, 0, BQ_Offset (BqAxes, foreground_labels));
	}

	int xvisible_tics;
	int yvisible_tics;
	char *x_label;
	char *y_label;
	char *x_format;
	char *y_format;
	char *x_time_format;
	char *x_date_format;
	char *axis_style;
	char *tic_style;
	char tic_style_west;
	char tic_style_north;
	char tic_style_east;
	char tic_style_south;
	double xincr_num;
	double xincr_sml;
	double yincr_num;
	double yincr_sml;
	int boxwidth;
	int xtimewidth;
	int xwidth_big;
	int xwidth_sml;
	double xticlength_sml;
	double xticlength_big;
	int ywidth_big;
	int ywidth_sml;
	double yticlength_sml;
	double yticlength_big;
	int mindx;
	int mindy;
	int style;
	int xvisible_grids;
	int yvisible_grids;
	int xwidth_big_grids;
	int xwidth_sml_grids;
	int ywidth_big_grids;
	int ywidth_sml_grids;

	BQ_Color foreground_labels;
	BQ_Color foreground_numbers;
	BQ_Color foreground_axes;
	BQ_Color x_time_axes;
	BQ_Font font_labels;
	BQ_Font font_numbers;
	BQ_Font font_supers;
	BQ_Color color_grids_x;
	BQ_Color color_grids_x_small;
	BQ_Color color_grids_y;
	BQ_Color color_grids_y_small;
};

class BQAxes : public BQViewportItem, public BQConfigure {
public:
	BQAxes(BQViewport *master);
	int processEvent (int type, QEvent *event);

	BqAxes vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	void chooseScale (double range, int min_ticks, int max_ticks, 
			double *divisions, int ndivisions, double *major, double *minor, int *power);
	char *chooseTimeFormat ( double major );
	void labelXTic (QPainter *painter, char *string, double x, double fy);
	void drawXTimeAxis (QPainter *painter, double t0, double t1);
	void drawXTicLabels (QPainter *painter, int w, double xstart, double xscale);
	void drawYTicLabels (QPainter *painter, int h, double ystart, double yscale);
	void drawBox (int w, int h, QPainter *painter);
	void drawTics (int w, int h, QPainter *painter);
	void drawXLabel (QPainter *painter);
	void drawYLabel (QPainter *painter);
	void drawXaxisStuff (QPainter *painter, 
		BQ_Color * color, int linewidth, int linestyle, double incr, double fy1, double fy2,
		BQ_Color * color_small, int linewidth_small, int linestyle_small, double incr_small,
		double fy1_small, double fy2_small, double mindx);
	void drawYaxisStuff (QPainter *painter, 
		BQ_Color *color, int linewidth, int linestyle, double incr, double fx1, double fx2,
		BQ_Color *color_small, int linewidth_small, int linestyle_small, double incr_small,
		double fx1_small, double fx2_small, double mindy);
	int draw (QPainter *painter);
}; 

typedef int (BQ_textCallback) (BQText *text, int type, QEvent *event, BQ_ClientData client_data);

struct BqTextCallback {
	BqTextCallback () {
		memset (this, 0, sizeof(BqTextCallback));
	}

	std::function<BQ_textCallback> callback_proc;
	BQ_ClientData client_data;
};

struct BqText {
	BqText () {
		memset (this, 0, BQ_Offset (BqText, foreground));
	}

	double opacity;
	int select;
	char *text;
	double fxbb, fybb;
	int wbb, hbb;
	double x, y;
	double angle;
	char *justification;
	void *relative;
	char *xpos;
	char *ypos;
	int width, height;
	int show_not_symbol;
	QPainterPath *path;
	BQ_ClientData textCallbackClientData;

	BQ_Color foreground;
	BQ_Color outline;
	BQ_Color background;
	BQ_Font font;
	QList<BqTextCallback> textCallbacks;
};

struct BqTextSpecialSymbol {
	BqTextSpecialSymbol () {
		memset (this, 0, sizeof (BqTextSpecialSymbol));
	}

	char *name;
	int size;
	int width;
	int height;
	QPixmap *pixmap;
};

class BQText : public BQViewportItem, public BQConfigure {
	friend class BQTaskbar;

public:
	BQText(BQViewport *master, QString textstring, QString x, QString y, int register_item = 1);
	~BQText();
	int processEvent (int type, QEvent *event);
	void getBB (double *x, double *y, int *w, int *h);
	static QPixmap *specialSymbol (char *name, int size);
	void setTextCallback (std::function<BQ_textCallback> textCallback, BQ_ClientData client_data);

	BqText vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init(char *textstring, char *x, char *y);
	void init(QString textstring, QString x, QString y);
	bool contains (QMouseEvent *event);
	int draw (QPainter *painter);
	static BqTextSpecialSymbol *makeSpecialSymbol (char *name, int size);
}; 

class BQTextPopupWidget : public QWidget {
public:
	BQTextPopupWidget(BQTextPopup *popup);

protected:
	void paintEvent (QPaintEvent *event);

	BQTextPopup *popup;
};

struct BqTextpopup {
	BqTextpopup () {
		memset (this, 0, BQ_Offset (BqTextpopup, color_background));
	}
	~BqTextpopup () {
		if (timer) delete timer;
		if (popup_widget) delete popup_widget;
		if (pixmap) delete pixmap;
	}

	double opacity;
	double timeout;
	QTimer *timer;
	BQTextPopupWidget *popup_widget;
	QPixmap *pixmap;

	BQ_Color color_background;
	QString text;
	QPoint position;
	QString justification;
};

class BQTextPopup : public BQViewportItem, public BQConfigure {
public:
	BQTextPopup(BQViewport *master);
	void raiseTextPopup();
	void setText (QString text);
	void setTimeout (double timeout);
	void setJustification (QString justification);
	void setPosition (QPoint pos);
	void stopTimer ();
	void deletePopup ();
	QPoint justify();
	static QPixmap *makePixmap(QString text, QColor *background);

	BqTextpopup vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
};

class BQPolylinePath {
public:
	BQPolylinePath (QPainterPath *path, int np=0, int sp=0) {
		this->path = path;
		this->np = np;
		this->sp = sp;
	}
	~BQPolylinePath () {
		if (this->path) delete this->path;
	}

	QPainterPath *path;
	int np;
	int sp;
};

struct BqPolyline {
	BqPolyline () {
		memset (this, 0, BQ_Offset (BqPolyline, fill));
	}
	~BqPolyline () {
		if (xpts) free (xpts);
		if (ypts) free (ypts);
	}

	int linewidth;
	int linestyle;
	double xmin, xmax, ymin, ymax;
	double gap_value;
	int npts;
	double *xpts;
	double *ypts;
	int xpts_size;
	int ypts_size;
	char *vechandle;
	char *thandle;
	BQPolylinePath *plpath;
	double xgain;
	double ygain;
	int xmode;
	int ymode;

	BQ_Color fill;
	BQ_Color outline;
	BQ_Color fill_black;
	BqGraphicsContext gc;
};

class BQPolyline : public BQViewportItem, public BQConfigure {
public:
	BQPolyline(BQViewport *master);
	void setGc (BqGraphicsContext gc);
	int setSourceTrace (Dbptr trace);
	int setSourceBugrid (QString name);

	BqPolyline vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	int draw (QPainter *painter);
};

struct BqDatagrid {
	BqDatagrid () {
		memset (this, 0, sizeof (BqDatagrid));
	}
	~BqDatagrid () {
		if (xgrid) free (xgrid);
		if (ygrid) free (ygrid);
		if (zgrid) free (zgrid);
		if (image) delete (image);
		if (image_buf) free (image_buf);
	}

	double gap_value;
	int nx, ny;
	double *xgrid;
	double *ygrid;
	double *zgrid;
	char *gridhandle;
	int zmode;
	double zmin, zmax;
	double zgain;
	double opacity;
	int invertx;
	int inverty;
	int zspec_dif;

	QImage *image;
	int image_width, image_height;
	unsigned char *image_buf;
	size_t image_buf_size;
};

class BQDatagrid : public BQViewportItem, public BQConfigure {
public:
	BQDatagrid(BQViewport *master);

	BqDatagrid vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	int draw (QPainter *painter);
	int makeImage ( );
};

struct BqPolypoint {
	BqPolypoint () {
		memset (this, 0, BQ_Offset (BqPolypoint, glyphCallbacks));
	}
	~BqPolypoint () {
		if (points) delete points;
	}

	int symbol_default;
	int xmode_default;
	int ymode_default;
	int shape_label_anchor_default;
	int label_anchor_default;
	int auto_cursor;
	double size_default;
	int showlabels;
	BQGraphicsContext *gc_default;
	BqGraphicsContextItem *color_fill_default;
	BqGraphicsContextItem *color_fill_node_default;
	BqGraphicsContextItem *color_outline_default;
	BqGraphicsContextItem *color_text_default;
	BqGraphicsContextItem *color_outline_text_default;
	BqGraphicsContextItem *color_fill_text_default;
	BqGraphicsContextItem *font_default;
	BqGraphicsContextItem *linestyle_default;
	BqGraphicsContextItem *joinstyle_default;
	BqGraphicsContextItem *capstyle_default;
	BqGraphicsContextItem *linewidth_default;
	BqGraphicsContextItem *linewidth_node_default;
	BqGraphicsContextItem *opacity_default;
	double linewidth_value_default;
	double linewidth_node_value_default;
	double opacity_value_default;
	double selection_size_increase_factor;
	double selection_opacity_reduction_factor;
	double xgain;
	double ygain;
	int interaction_mode;
	char *vechandle;
	BQGlyphs *points;

	QList<BqGlyphCallback> glyphCallbacks;
	QString layer_name;
	QString color_fill_name_default;
	QString color_fill_node_name_default;
	QString color_outline_name_default;
	QString color_text_name_default;
	QString color_outline_text_name_default;
	QString color_fill_text_name_default;
	QString font_name_default;
	QString linestyle_name_default;
	QString joinstyle_name_default;
	QString capstyle_name_default;
};

class BQPolypoint : public BQViewportItem, public BQConfigure {
public:
	BQPolypoint(BQViewport *master);
	void setGlyphs ();
	void setGlyphCallback (std::function<BQ_glyphCallback> glyphCallback, BQ_ClientData client_data);
	void setGlyphsAutoCursor (int auto_cursor);

	BqPolypoint vars;

protected:
	int drawViewportItem (QPainter *painter);
	int processEvent (int type, QEvent *event);
	void setVisibility () {};

private:
	void init();
};

struct BqTaskbarButtonStateParams;

struct BqTaskbarButtonParams {
	BqTaskbarButtonParams () {
		memset (this, 0, BQ_Offset (BqTaskbarButtonParams, description));
	}

	BQTaskbar *taskbar;
	int width, height;
	BQText *bq_text;
	int index;
	int current;

	QString description;
	QList<BqTaskbarButtonStateParams> states;
};

struct BqTaskbarButtonStateParams {
	BqTaskbarButtonStateParams () {
		memset (this, 0, BQ_Offset (BqTaskbarButtonStateParams, name));
	}

	double opacity;
	int show_not_symbol;
	int order;

	QString name;
	QString label;
	QString description;
	QList<QString> action;
	BQ_Color fill_color;
	BQ_Color background_color;
};

typedef int (BQ_taskbarCommandCallback) (QString command, BQ_ClientData client_data);

struct BqTaskbarCommandCallback {
	BqTaskbarCommandCallback () {
		memset (this, 0, sizeof(BqTaskbarCommandCallback));
	}

	std::function<BQ_taskbarCommandCallback> callback_proc;
	BQ_ClientData client_data;
};

struct BqTaskbar {
	BqTaskbar () {
		memset (this, 0, BQ_Offset (BqTaskbar, background));
	}

	int selection_index;
	BQTextPopup *popup;

	BQ_Color background;
	BQ_Font font;
	QString buttons_definitions_string;
	QHash<QString, BqTaskbarButtonParams *> buttons_definitions;
	QString buttons_string;
	QList<BqTaskbarButtonParams *> buttons_list;
};

class BQTaskbar : public BQViewportItem, public BQConfigure {
public:
	BQTaskbar(BQViewport *master);
	~BQTaskbar();

	int setState (QString request);
	void setState (BqTaskbarButtonParams *params, BqTaskbarButtonStateParams *state, QMouseEvent *mevent=NULL);
	int setStateIndex (QString name_button, QString name_state, 
			BqTaskbarButtonParams **params=NULL, BqTaskbarButtonStateParams **state=NULL);
	void setCommandCallback (std::function<BQ_taskbarCommandCallback> commandCallback, BQ_ClientData client_data=NULL);

	BqTaskbar vars;
	QList<BqTaskbarCommandCallback> command_callbacks;

protected:
	int drawViewportItem (QPainter *painter);
	int processEvent (int type, QEvent *event);
	void setVisibility () {};

private:
	void init();
	QList<BQText *> items (QMouseEvent *mevent);
	BQText *findClosest (QEvent *event);
};

class BUMapTransform {
public:
	BUMapTransform (double latr, double lonr);
	void edpDelxy2latlon (double delx, double dely, double *lat, double *lon);
	void edpLatlon2delxy (double lat, double lon, double *delx, double *dely);
	int globeDelxy2latlon (double delx, double dely, double *lat, double *lon);
	int globeLatlon2delxy (double lat, double lon, double *delx, double *dely);
	double latr, lonr, clat, slat;
};

class BQScaledata {
public:
	static const int disposition_noInput = -1;
	static const int disposition_notUsed = 0;
	static const int disposition_moveTo = 1;
	static const int disposition_lineTo = 2;

	BqViewport *vp;
	int map_tran;
	int plxmode;
	int plymode;
	int vpxmode;
	int vpymode;
	int w, h;
	double xstart, xend, xscale;
	double ystart, yend, yscale;
	BUMapTransform *tran;
	double xp_last, yp_last;
	double xpt, ypt;
	double xptp, yptp;
	double xinput, yinput;
	double xpts_last, ypts_last;
	int start, ndelta, idelta;
	double delx, dely;

	BQScaledata (BqViewport *viewport, int ignoretran=0,
							int xmode=BQ_POINT_MODE_WORLD_COORDINATES, int ymode=BQ_POINT_MODE_WORLD_COORDINATES);
	~BQScaledata ();
	int pushPoint (double xpt, double ypt);
	int popPoint (double *xscaled, double *yscaled);
};

class BQPolyDraw {
public:
	BQScaledata *scale;
	int w, h;
	double x_last, y_last;
	int ret_last, zone_last;
	int first;
	double xgain, ygain;
	int xmode, ymode;

	BQPolyDraw (BqViewport *viewport, int ignoretran=0, double xgain=1.0, double ygain=1.0, 
							int xmode=BQ_POINT_MODE_WORLD_COORDINATES, int ymode=BQ_POINT_MODE_WORLD_COORDINATES);
	~BQPolyDraw ();
	QPainterPath *addPoints (QPainterPath *path, int npts, double *xpts, double *ypts, int start=0, int subtract=0);
	QPainterPath *addPoints (QPainterPath *path, int npts, double *xpts, double *ypts, double gap_value, int start=0);
	QPainterPath *addPoints (QPainterPath *path, QPainterPath *inpath, int start=0, int subtract=0);
};

class BQMapData : public BUMapData {
public:
	BQMapData (QString fname, int nthreads = 1);
	BQMapData (int nthreads = 1);
	~BQMapData ();
	int setResolution (int resolution);
	int setResolution (double degrees_per_pixel);
	void drawVector(QPainter &painter, BqViewport *viewport, 
				unsigned int levels, QList<BQ_Color> outline, 
				int *linewidth, int *resolution, BQ_Color &water, BQ_Color &land);
	QImage *drawImage(int width, int height, int tran, double xleft, double xright, 
				double ybottom, double ytop, double lonr, double latr, double gamma=0.8, double lightmin=0.0);
	unsigned char *getImage (int ibin);

	int resolution;
	int nthreads;
	unsigned char *image_buf;
	size_t image_buf_size;
	QImage *image;
	int width, height;
	double lightmin;
	double gamma;
	double xmin, xmax, ymin, ymax, dx, dy;
	int tran;
	double latr, lonr;

private:
	Pmtcomp *mtc;
	int isInside(double lat, double lon, int npts, double *latpts, double *lonpts);
};

struct BqNavigator {
	BqNavigator () {
		memset (this, 0, BQ_Offset (BqNavigator, fill_land));
	}

	BQViewport *viewport;
	BQMap *parent_map;
	BQMap *map;
	BQPolyline *map_outline;
	int qevent_state;
	double event_xstart, event_ystart;
	int show;

	BQ_Color fill_land;
	BQ_Color fill_water;
};

class BQNavigator : public BQConfigure {
public:
	BQNavigator (BQMap *parent_map);
	int refreshVisibility();
	void draw ();

	BqNavigator vars;

protected:
	int drawViewportItem (QPainter *painter) {return (0);};
	void setVisibility () {};

private:
	int init (BQMap *parent_map);
};

#define	BQ_MAP_MAXLEVELS		(10)
#define	BQ_MAP_MAXLEVELS_COASTS		(8)
#define	BQ_MAP_MAXLEVELS_RIVERS		(10)
#define	BQ_MAP_MAXLEVELS_POLITICAL	(4)


struct BqMapLayer {
	BqMapLayer () {
		memset (this, 0, BQ_Offset (BqMapLayer, shade_lonr));
		mapdata = NULL;
		shade_image = NULL;
		shade_image_buf = NULL;
		shade_lonr = 0.0;
	}

	int type;
	double opacity;
	double gamma;
	double lightmin;
	int antialias;
	BQMapData *mapdata;
	int nodata;
	unsigned int levels;
	int linewidth[BQ_MAP_MAXLEVELS];
	int resolution[BQ_MAP_MAXLEVELS];
	BQMap *map;
	double shade_lonl;
	double shade_latl;
	QImage *shade_image;
	unsigned char *shade_image_buf;
	size_t shade_image_buf_size;
	int shade_width;
	int shade_height;
	double shade_xmin;
	double shade_xmax;
	double shade_ymin;
	double shade_ymax;
	double shade_dx;
	double shade_dy;
	int shade_tran;
	double shade_latr;
	double shade_lonr;

	QString mapdata_source;
	QList<int> allowed_projections;
	BQ_Color fill_color;
	BQ_Color background_color;
	BQ_Color transparent_color;
	QList<BQ_Color> outline;
};

class BQMapLayer : public BQViewportItem, public BQConfigure {
	friend class BQMap;

public:
	BQMapLayer(BQViewport *master, BQMap *map);
	~BQMapLayer();

	BqMapLayer vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init (BQMap *map);
	int draw (QPainter *painter);
	QImage *drawShadeImage(int width, int height, int tran, double xleft, double xright, 
			double ybottom, double ytop, double lonr, double latr, 
			double lonl, double latl);
	int fillShadeImage ();
};

struct BqMap {
	BqMap () {
		memset (this, 0, BQ_Offset (BqMap, fill_land));
	}

	int nthreads;
	int cache_maxsize;
	BQTaskbar *taskbar;
	BQNavigator *navigator;
	BQAxes *grids;
	QWidget *parent;
	int make_navigator;
	int vector_only;

	BQ_Color fill_land;
	BQ_Color fill_water;
	BQ_Color fill_black;
	QHash<QString, QString > maplayer_definitions_string;
	QHash<QString, BQMapLayer * > maplayer_definitions;
	QList<QString> maplayers;
	QHash<QString, QString > maplayers_templates;
	QString taskbar_buttons_definitions_string;
	QString taskbar_buttons_string;
};

#define	BQ_MAP_TYPE_FILL_LAND		(0)
#define	BQ_MAP_TYPE_FILL_WATER		(1)
#define	BQ_MAP_TYPE_LINE_COASTS		(2)
#define	BQ_MAP_TYPE_LINE_RIVERS		(3)
#define	BQ_MAP_TYPE_LINE_POLITICAL	(4)
#define	BQ_MAP_TYPE_GRID		(5)
#define	BQ_MAP_TYPE_VECTOR		(6)
#define	BQ_MAP_TYPE_VECTOR_FILL		(7)
#define	BQ_MAP_TYPE_SHADE		(8)

class BQMap : public BQViewportItem, public BQConfigure {
public:
	BQMap(BQViewport *master, int make_navigator=1, int make_grids=1, QString *item_name=NULL);

	BQTaskbar *getTaskbar();
	void clearMaplayerDefinitions();
	void setMaplayersStale ();
	void setDefaults();
	int makeMaplayer (QString name);

	static void solarPosition (double epoch, double *latitude, double *longitude);

	BqMap vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init (int make_navigator, int make_grids);
};

#define BQ_MAPEVENTS_EVENTS_WINDOW_DISPLAY_MODE_INITIAL		0
#define BQ_MAPEVENTS_EVENTS_WINDOW_DISPLAY_MODE_SELECTED	1

struct BqDrawEvents {
	BqDrawEvents () {
		display_mode = BQ_MAPEVENTS_EVENTS_WINDOW_DISPLAY_MODE_INITIAL;
		interaction_mode = BQ_INTERACTION_MODE_SHOW;
		selection_opacity_reduction_factor = 0.5;
		selection_size_increase_factor = 4.0;
		map = NULL;
		mapevents = NULL;
		events_glyphs = NULL;
		detections_glyphs = NULL;
		popup = NULL;
	}
	~BqDrawEvents () {
		if (popup) delete popup;
		if (map) delete map;
	}

	int display_mode;
	int interaction_mode;
	double selection_opacity_reduction_factor;
	double selection_size_increase_factor;
	BQMap *map;
	BQMapevents *mapevents;
	BQGlyphs *events_glyphs;
	BQGlyphs *detections_glyphs;
	BQTextPopup *popup;

	QString layer;
};

class BQDrawEvents : public BQViewportItem, public BQConfigure {
	friend class BQMapevents;

public:
	BQDrawEvents (BQViewport *master, BQMapevents *mapevents, QString item_name, QString layer, 
				int make_navigator, int make_grids);
	~BQDrawEvents () {clear();}
	void makeEventsGlyphs ();
	void makeDetectionsGlyphs ();
	void clear ();
	void clearEvents ();
	void clearDetections ();
	void hide ();
	void hideEvents ();
	void hideDetections ();
	void show ();
	void showEvents ();
	void showDetections ();
	void refreshEvents ();
	void refreshDetections (int recenter=1);
	void showEventsWindow (int mode);
	void setInteractionMode (int interaction_mode);
	void changeDisplayMode (int mode);
	void hidePopup();
	int showPopup(EVEvent *event=NULL);
	int showPopup(EVDetection *detection=NULL, QString *color=NULL);

	BqDrawEvents vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	static int events_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	static int detections_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	QString makePopupText (EVEvent *event);
	QString makePopupText (EVDetection *detection);
	long evid = -1;
};

struct BqDrawEvent {
	BqDrawEvent () {
		map = NULL;
		mapevents = NULL;
		origins_glyphs = NULL;
		arrivals_glyphs = NULL;
		popup = NULL;
	}
	~BqDrawEvent () {
		if (popup) delete popup;
		if (map) delete map;
	}

	BQMap *map;
	BQMapevents *mapevents;
	BQGlyphs *origins_glyphs;
	BQGlyphs *arrivals_glyphs;
	BQTextPopup *popup;

	QString layer;
	QList<BQPolyline *> origin2sta_polylines;
};

class BQDrawEvent : public BQViewportItem, public BQConfigure {
	friend class BQMapevents;

public:
	BQDrawEvent (BQViewport *master, BQMapevents *mapevents, QString item_name, QString layer, 
				int make_navigator, int make_grids);
	~BQDrawEvent () {clear();}
	void makeOriginsGlyphs ();
	void makeArrivalsGlyphs ();
	void clear ();
	void show ();
	void hide ();
	void clearOrigins ();
	void showOrigins ();
	void hideOrigins ();
	void clearArrivals ();
	void showArrivals ();
	void hideArrivals ();
	void showOrigin (int update_origins_window=1);
	void refreshOrigins ();
	void refreshArrivals ();
	void showEventWindow (int update_origins_window=1);
	void setInteractionMode (int interaction_mode);
	void hidePopup();
	int showPopup(EVDetection *detection=NULL);
	int showPopup(EVAssociation *association=NULL);

	BqDrawEvent vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	static int origins_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	static int arrivals_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	QString makePopupText (EVDetection *detection);
	QString makePopupText (EVAssociation *association);
	static QList<BQ_XY> *makeGreatCirclePath (double lat1, double lon1, double lat2, double lon2, int n);
};

struct BqDrawOrigins {
	BqDrawOrigins () {
		map = NULL;
		ggc = NULL;
		mapevents = NULL;
		origins_glyphs = NULL;
	}
	~BqDrawOrigins () {
		if (popup) delete popup;
		if (map) delete map;
	}

	BQMap *map;
	BQGlyphsGraphicsContext *ggc;
	BQMapevents *mapevents;
	BQGlyphs *origins_glyphs;
	BQTextPopup *popup;

	QString layer;
	QList<BQPolyline *> error_ellipse_polylines;
	QHash<int, BQPolyline *> error_ellipse_hash;
	EVEvent event;
	EVOrigin preferred_origin;
};

class BQDrawOrigins : public BQViewportItem, public BQConfigure {
	friend class BQMapevents;

public:
	BQDrawOrigins (BQViewport *master, 
				BQGlyphsGraphicsContext *ggc, BQMapevents *mapevents, QString item_name, QString layer);
	~BQDrawOrigins () {clear();}
	void clear ();
	void makeGlyphs ();
	void hide ();
	void show ();
	void hidePopup();
	void showPopup(EVOrigin *origin=NULL);
	void showErrorEllipses (int index);

	BqDrawOrigins vars;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	static int glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	static QList<BQ_XY> *makeErrorEllipse (double lat, double lon, double semimajor, double semiminor, double azimuth, int n=0);
	QString makePopupText (EVOrigin *origin);
	long orid = -1;
};

#define BQ_GLYPHS_TYPE_EVENTS	1
#define BQ_GLYPHS_TYPE_ORIGINS	2

struct BqStation {
	BqStation () {
		memset (this, 0, BQ_Offset (BqStation, sta));
	}

	double lat;
	double lon;
	double elev;
	double distance;

	QString sta;
	BqGraphicsContext gc;
};

typedef void (BQ_glyphMapeventsCallback) (BQGlyph *glyph, BQEvent *event, int glyph_type, int show_mode, int window, BQ_ClientData client_data);

struct BqGlyphMapeventsCallback {
	BqGlyphMapeventsCallback () {
		memset (this, 0, sizeof(BqGlyphMapeventsCallback));
	}

	std::function<BQ_glyphMapeventsCallback> callback_proc;
	BQ_ClientData client_data;
};

typedef void (BQ_eventMapeventsCallback) (EVEvent *event, int show_mode, int window, BQ_ClientData client_data);

struct BqEventMapeventsCallback {
	BqEventMapeventsCallback () {
		memset (this, 0, sizeof(BqEventMapeventsCallback));
	}

	std::function<BQ_eventMapeventsCallback> callback_proc;
	BQ_ClientData client_data;
};

typedef void (BQ_eventviewMapeventsCallback) (BQEVCallbackEvent *event, BQ_ClientData client_data);

struct BqEventviewMapeventsCallback {
	BqEventviewMapeventsCallback () {
		memset (this, 0, sizeof(BqEventviewMapeventsCallback));
	}

	std::function<BQ_eventviewMapeventsCallback> callback_proc;
	BQ_ClientData client_data;
};

class BQEVClient : public EVClient {
public:
	BQEVClient (EVServer *ev, BQMapevents *bqme);

protected:
	void responseCallback (EVResponse response);
	BQMapevents *bqme;
};

#define BQ_MAPEVENTS_GLYPH_TYPE_EVENT		0
#define BQ_MAPEVENTS_GLYPH_TYPE_ORIGIN		1
#define BQ_MAPEVENTS_GLYPH_TYPE_ASSOCIATION	2
#define BQ_MAPEVENTS_GLYPH_TYPE_DETECTION	3

#define BQ_MAPEVENTS_GLYPH_WINDOW_EVENTS	0
#define BQ_MAPEVENTS_GLYPH_WINDOW_EVENT		1
#define BQ_MAPEVENTS_GLYPH_WINDOW_ORIGINS	2

struct BqMapeventsEvents {
	BqMapeventsEvents () {
		events_vector = NULL;
	}

	std::shared_ptr<Table<EVEvent>> events_vector_shared;
	Table<EVEvent> *events_vector;
	Table<EVDetection> events_detections_vector;
	EVEvent last_modified_event;
	EVEvent selected_event;
	EVOrigin selected_origin;
	EVOrigin preferred_origin;
	EVResponse last_event_view_response; 
};

struct BqMapevents {
	BqMapevents () {
		memset (this, 0, BQ_Offset (BqMapevents, events_definitions));
		selected_event_evid = -1;
		selected_origin_orid = -1;
	}
	~BqMapevents () {
		if (events_window) delete events_window;
		if (event_window) delete event_window;
		if (origins_window) delete origins_window;
		if (ggc) delete ggc;
		if (taskbar) delete taskbar;
	} 

	BQEVClient *ev;
	BQViewport *event_window_vp;
	BQViewport *origins_window_vp;
	BQDrawEvents *events_window;
	BQDrawEvent *event_window;
	BQDrawOrigins *origins_window;
	BQGlyphsGraphicsContext *ggc;
	int event_window_stale;
	BQTaskbar *taskbar;
	int show_popup_events_event;
	int show_popup_events_detection;
	int show_popup_event_event;
	int show_popup_event_detection;
	int show_popup_event_arrival;
	int show_popup_origins_origin;
	int show_events_detections;
	int auto_center;
	int auto_show_event;
	double update_interval;
	double last_update;
	QTimer *update_timer;
	double max_age_detections;
	double last_time;
	double age_ref;
	int age_ref_now;
	long selected_event_evid;
	long selected_origin_orid;
	double events_lat;
	double events_lon;
	double detections_lat;
	double detections_lon;
	double origins_lat;
	double origins_lon;
	double origins_delta_max;

	QString events_definitions[2];
	QString events_definitions_selected[2];
	QString origins_definitions[2];
	QString origins_definitions_pref[2];
	QList<BQPolyline *> origin2sta_polylines;
	QString taskbar_buttons_definitions_string;
	QString taskbar_buttons_string;
	QString glyphs_gcitem_definitions_string;
	QString glyphs_gc_definitions_string;
	QList<BqGlyphMapeventsCallback> glyphCallbacks; 
	QList<BqEventMapeventsCallback> eventCallbacks; 
	QList<BqEventviewMapeventsCallback> eventviewCallbacks; 
};

class BQMapevents : public BQViewportItem, public BQConfigure {
	friend class BQDrawEvent;
	friend class BQDrawEvents;
	friend class BQDrawOrigins;

public:
	BQMapevents(BQViewport *events_vp, BQViewport *event_vp, BQViewport *origins_vp, int make_navigator=1, int make_grids=1); 
	BQMap *getMap();
	BQTaskbar *getTaskbar();
	int setDefaults(int nort=0);
	void setEventView(EVServer *evserver);
	void refreshEventView();
	void selectEvent (long evid);
	void showEvent ();
	void selectOrigin (long orid);
	void showEvid (long evid);
	void setupTimerUpdate();
	void setGlyphCallback (std::function<BQ_glyphMapeventsCallback> glyphCallback, BQ_ClientData client_data);
	void setEventCallback (std::function<BQ_eventMapeventsCallback> eventCallback, BQ_ClientData client_data);
	void setEventviewCallback (std::function<BQ_eventviewMapeventsCallback> eventviewCallback, BQ_ClientData client_data);
	void processEVCallback (BQEVCallbackEvent *event);
	void updateDisplays (QString which);
	void linkBQEVEventsTableview (BQEVEventsTableview *evtableview);

	BqMapevents vars;
	BqMapeventsEvents events;
	BQEVEventsTableview *evtableview=NULL;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init(BQViewport *events_vp, BQViewport *event_vp, BQViewport *origins_vp, int make_navigator=1, int make_grids=1);
	void timerCallback();
	void makeOriginsWindow ();
	void makeEventWindow ();
	void showEventWindow (int update_origins_window=1);
	void getLastTime ();
	BQViewport *getEventViewport ();
	BQMap *getEventMap ();

	static int coding_callback (BQGlyph *glyph, QString definitions_name, QString coding_name, double *value, BQ_ClientData client_data);
};

struct BqTableviewParams {
	BqTableviewParams () {
		nrows = -1;
		ncols = -1;
		default_cell_height = -1;
		default_cell_width = -1;
		hover_interaction = 0;
		show_record_numbers = 0;
		auto_record_sort = 1;
		font = QFontDatabase::systemFont(QFontDatabase::FixedFont);
		font_header = QFontDatabase::systemFont(QFontDatabase::FixedFont);
		BQ_ParseColor ("", accentuated_foreground);
		BQ_ParseColor ("", accentuated_selected_foreground);
		BQ_ParseColor ("", accentuated_selected_background);
		BQ_ParseColor ("", record_numbers_background);
	}

	int nrows;
	int ncols;
	int default_cell_height;
	int default_cell_width;
	int hover_interaction;
	int show_record_numbers;
	int auto_record_sort;
	QFont font;
	QFont font_header;
	QList<QVariant> column_widths;
	BQ_Color accentuated_foreground;
	BQ_Color accentuated_selected_foreground;
	BQ_Color accentuated_selected_background;
	BQ_Color record_numbers_background;
};

class BQTableviewModel;
class BQTableviewSortProxy;

class BQTableviewHHeader : public QHeaderView {
	friend class BQTableview;
	friend class BQEVTableview;

public:
	BQTableviewHHeader (QWidget *parent, BQTableview *tableview);
	void setSortColumn (int col, int reverse);

	BQTableview *tableview;

protected:
	void mousePressEvent (QMouseEvent *event);
	void mouseReleaseEvent (QMouseEvent *event);
};

class BQTableview : public QTableView, public BQConfigure {
public:
	BQTableview (QWidget *parent);
	void setDimensions (int nrows, int ncols);
	void setDimensions ();
	void totalSize (int &width, int &height);
	void reset ();
	void setRecordTab (QList<long> tab);
	int getRecord (int row);
	int getRow (int record);
	void setColumnWidths();

	virtual QVariant callback (QStringList request) {return QVariant();}

	QString handle;
	BQTableviewModel *bqmodel;
	BQTableviewSortProxy *sort_proxy;
	BQTableviewHHeader *hheader;
	BqTableviewParams params;

protected:
	void selectionChanged (const QItemSelection &selected, const QItemSelection &deselected);
	void dataChanged (const QModelIndex &topLeft, const QModelIndex &bottomRight, const QVector<int> &roles);
	void mousePressEvent (QMouseEvent *event);
	void mouseMoveEvent (QMouseEvent *event);
	void mouseReleaseEvent (QMouseEvent *event);
	void setAutoRecordSort (int auto_record_sort);
	void selectionChangedCallback (int row);
};

class BQDbTableview : public BQTableview {
public:
	BQDbTableview (QWidget *parent, Dbptr db);
	void setTable();

	QVariant callback (QStringList request);

	Dbptr db;
	Dbptr dbs;
	int dbsfree = 0;
	QStringList collist;
	QHash<QString,QHash<QString,QString>> colhash;

protected:
	virtual void selectionCallback (int &row) {}
};

#define BQ_EVTABLEVIEW_TABLE_TYPE_EVENT			0
#define BQ_EVTABLEVIEW_TABLE_TYPE_ORIGIN		1

class BQTableviewModel : public QAbstractTableModel {
public:
	BQTableviewModel (QWidget *parent, BQTableview *tableview);
	void reset (int nrows, int ncols);

	BQTableview *tableview;
	QHash<QModelIndex, int> accentuated_hash;
	int last_hover_row = -1;
	int last_hover_col = -1;
	int sort_order_col = -1;
	int sort_order_reverse = 0;
	int noselect=0;
	QList<long> record_tab;

protected:
	int rowCount(const QModelIndex &parent = QModelIndex()) const ;
    int columnCount(const QModelIndex &parent = QModelIndex()) const;
    QVariant headerData(int section, Qt::Orientation orientation, int role = Qt::DisplayRole) const;
    QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const;
};

struct BqEVTableview {
	BqEVTableview () {
		show_err = 0;
		auto_select_event = 0;
	}

	int show_err;
	int auto_select_event;
	BQ_Color color_preferred_origin_foreground;
};

class BQEVTableview : public BQTableview {
public:
	BQEVTableview (QWidget *parent, EVClient *ev_client, int event_index=-1);
	BQEVTableview (QWidget *parent, EVServer *ev_server, int event_index=-1);
	BQEVTableview (QWidget *parent, BQEVTableview *sync_events_table, int selected_event_index=0);
	~BQEVTableview ();

	void setSelectedIndex (int selected_index);
	void setSelectedEvid (long selected_evid);
	void selectEvent (long evid, int nofeedback=0);
	void showEvent (long evid);
	void selectOrigin (long orid, int nofeedback=0);
	void showOrigin (long orid);
	int isOrigin ();
	EVClient *getEVClient();

	int parseColumnNames(Pf *pf);
	int parseColumnNames(QString pfstring);
	int parseColumnDefinitions(Pf *pf);
	int parseColumnDefinitions(QString pfstring);
	void refresh (EVResponse response);
	void refresh ();
	void mapEvid (long evid, int &record, int &row);
	void mapOrid (long orid, int &record, int &row);

	BqEVTableview vars;

	int event_index;
	EVClient *ev_client;
	std::shared_ptr<Table<EVEvent>> events_shared;
	Table<EVEvent> *events;
	QList<long> sort_tab;
	QStringList collist;
	QHash<QString,QHash<QString,QString>> colhash;
	BQEVTableview *sync_events_table=NULL;
	BQEVTableview *sync_origins_table=NULL;

protected:
	virtual void selectionCallback (int &row);
	QVariant callback (QStringList request);
	bool event (QEvent *event);

	EVEvent *selected_event=NULL;
	long selected_evid = -1;
	int selected_record = -1;

private:
	void init (int event_index);
	void setTable();
	int checkSource(QString source);
	int checkSort(QString Sort);
	void substituteIndex(QString &source, int index);
	void processEvent (BQEVCallbackEvent *ev);
	void setSelectedIndexPvt (int selected_index);
	int makeSortTab (int column=-1, int reverse=0);

	int free_ev_client = 0;
	int sort_column = -1;
	int sort_reverse = 0;
	int nofeedback = 0;
};

struct BqEVEventsTableview {
	BqEVEventsTableview () {
		show_err = 0;
		auto_select_event = 0;
	}

	int show_err;
	int auto_select_event;
};

class BQEVEventsTableview : public BQTableview {
	friend class BQEVOriginsTableview;
	friend class BQMapevents;
	friend class BQTraceview;

public:
	BQEVEventsTableview (QWidget *parent, EVClient *ev_client);
	BQEVEventsTableview (QWidget *parent, EVServer *ev_server);
	~BQEVEventsTableview ();

	EVClient *getEVClient();
	BQEVOriginsTableview *getBQEVOriginsTableview();
	Table<EVEvent> *getEvents();
	void selectEvid (long evid, int nofeedback=0);
	void showEvid (long evid);
	int sortByColumn (QString colname, int reverse=0);
	long selectedEvid();
	long selectedOrid();

	int parseColumnNames(Pf *pf);
	int parseColumnNames(QString pfstring);
	int parseColumnDefinitions(Pf *pf);
	int parseColumnDefinitions(QString pfstring);
	void refresh (EVResponse response);

	BqEVEventsTableview vars;
	QStringList collist;
	QHash<QString,QHash<QString,QString>> colhash;

protected:
	virtual void selectionCallback (int &row, long evid, long orid) {}
	QVariant callback (QStringList request);
	bool event (QEvent *event);

private:
	void init();
	void setTable(int update_columns=1);
	void setSelectedRow (int row);
	int checkSource(QString source);
	int checkSort(QString Sort);
	void mapEvid (long evid, int &record, int &row);
	int makeSortTab (int column=-1, int reverse=0);
	void substituteIndex(QString &source, int index);
	void processEvent (BQEVCallbackEvent *ev);

	EVClient *ev_client = NULL;
	BQEVOriginsTableview *ortv = NULL;
	BQMapevents *mapevents = NULL;
	BQTraceview *traceview = NULL;
	int free_ev_client = 0;
	std::shared_ptr<Table<EVEvent>> events_shared;
	Table<EVEvent> *events;
	long selected_evid = -1;
	EVEvent *selected_event = NULL;
	QList<long> sort_tab;
	int sort_column = -1;
	int sort_reverse = 0;
	int nofeedback = 0;
	int something_changed = 0;
	int force_selection_changed = 0;
};

struct BqEVOriginsTableview {
	BqEVOriginsTableview () {
		show_err = 0;
	}

	int show_err;
	BQ_Color color_preferred_origin_foreground;
};

class BQEVOriginsTableview : public BQTableview {
	friend class BQEVEventsTableview;

public:
	BQEVOriginsTableview (QWidget *parent, BQEVEventsTableview *evtv);
	~BQEVOriginsTableview ();

	BQEVEventsTableview *getBQEVEventsTableview();
	Table<EVOrigin> *getOrigins();
	void refresh (long orid=-1, int nofeedback=0);
	void selectOrid (long orid, int nofeedback=0);
	void showOrid (long orid);
	int sortByColumn (QString colname, int reverse=0);
	long selectedEvid();
	long selectedOrid();

	int parseColumnNames(Pf *pf);
	int parseColumnNames(QString pfstring);
	int parseColumnDefinitions(Pf *pf);
	int parseColumnDefinitions(QString pfstring);

	BqEVOriginsTableview vars;
	QStringList collist;
	QHash<QString,QHash<QString,QString>> colhash;

protected:
	virtual void selectionCallback (int &row, long evid, long orid) {}
	QVariant callback (QStringList request);

private:
	void init();
	void setTable(int update_columns=1);
	void setSelectedRow (int row);
	int checkSource(QString source);
	int checkSort(QString Sort);
	void mapOrid (long orid, int &record, int &row);
	int makeSortTab (int column=-1, int reverse=0);
	void substituteIndex(QString &source, int index);

	BQEVEventsTableview *evtv = NULL;
	long selected_orid = -1;
	EVOrigin *selected_origin = NULL;
	QList<long> sort_tab;
	int sort_column = -1;
	int sort_reverse = 0;
	int nofeedback = 0;
	int something_changed = 0;
	int force_selection_changed = 0;
};

struct BqPfTreeviewParams {
	BqPfTreeviewParams () {
		default_cell_height = -1;
		default_cell_width = -1;
		allow_value_edits = true;
		allow_key_edits = false;
		allow_constraints_edits = false;
		allow_comment_edits = false;
		show_constraints = false;
		show_comments = true;
		show_noninline_comments = true;
		show_sources = false;
		show_keys = true;
		show_values = true;
		font = QFontDatabase::systemFont(QFontDatabase::FixedFont);
		BQ_ParseColor ("black", key_foreground);
		BQ_ParseColor ("blue", key_arr_foreground);
		BQ_ParseColor ("green", key_tbl_foreground);
	}

	int default_cell_height;
	int default_cell_width;
	bool allow_value_edits;
	bool allow_key_edits;
	bool allow_constraints_edits;
	bool allow_comment_edits;
	bool show_constraints;
	bool show_comments;
	bool show_noninline_comments;
	bool show_sources;
	bool show_keys;
	bool show_values;
	QFont font;
	QList<int> column_widths;
	BQ_Color key_foreground;
	BQ_Color key_arr_foreground;
	BQ_Color key_tbl_foreground;
};

class BQPfTreeviewModel;

class BQPfTreeview : public QTreeView, public BQConfigure {
public:
	BQPfTreeview (QWidget *parent, BUPf *bupf);
	void redraw ();
	void iterate (const std::function<void(BQPfTreeview *tv, int row, const QModelIndex&, const QModelIndex&)> & function, 
			QModelIndex parent = QModelIndex());
	bool checkValue (int pfindex, QString *invalue=NULL, QList<int> *types=NULL);

	QString handle;
	BQPfTreeviewModel *bqmodel;
	BUPf *pf;
	BqPfTreeviewParams params;

protected:
	void dataChanged (const QModelIndex &topLeft = QModelIndex(), const QModelIndex &bottomRight = QModelIndex(), const QVector<int> &roles = QVector<int> ()) {
		// qDebug() << "dataChanged" << topLeft.row() << "x" << topLeft.column() << "to" << bottomRight.row() << "x" << bottomRight.column();
		QTreeView::dataChanged (topLeft, bottomRight, roles);
	}

private:
	void init (QWidget *parent, BUPf *bupf);
};

typedef void (BQ_commandCallback) (QString command, BQ_ClientData client_data);

struct BqCommandConsole {
	BQ_Color 	color_foreground;
	BQ_Color 	color_background;
	BQ_Color 	color_foreground_selection;
	BQ_Color 	color_background_selection;
	BQ_Color 	color_prompt;
	BQ_Color 	color_output;
	BQ_Color 	color_erroroutput;
	BQ_Font 	font;
	char		*prompt = NULL;
};

struct BqCommand {
	int position_prompt;
	int prompt_length;
	int position_command_end;
};

struct BqCommandQueue {
	BqCommandQueue (QString command, int echo_command, int echo_output) {
		this->command = command;
		this->echo_command = echo_command;
		this->echo_output = echo_output;
	}

	QString command;
	int echo_command;
	int echo_output;
};

struct BqCommandConsoleCommandCallback {
	BqCommandConsoleCommandCallback () {
		memset (this, 0, sizeof(BqCommandConsoleCommandCallback));
	}

	std::function<BQ_commandCallback> callback_proc;
	BQ_ClientData client_data;
};

class BQCommandConsole : public QTextEdit, public BQConfigure {
public:
	BQCommandConsole (QWidget *parent);

	void sendCommand (QString command, int echo_command=1, int echo_output=1);
	int importFromFile (QString file_name, int echo=1);
	void setCommandCallback (std::function<BQ_commandCallback> commandCallback, BQ_ClientData client_data=NULL);
	void removeCommandCallback (std::function<BQ_commandCallback> commandCallback);
	void printLineOutput (QString command, QString line, int error_output_flag, int output_done_flag);
	QString getCommand (int command_index);

	BqCommandConsole vars;
	QList<BqCommand> commands;
	QList<BqCommandQueue> command_queue;

protected:
    void keyPressEvent(QKeyEvent *e) override;
    void mousePressEvent(QMouseEvent *e) override;
    void mouseMoveEvent(QMouseEvent *e) override;
    void mouseReleaseEvent(QMouseEvent *e) override;
    void contextMenuEvent(QContextMenuEvent *e) override;
	void enterEvent(QEvent *event) override;

private:
	void setPrompt (QString prompt);
	void echoCommand (QString command);
	void setCommand (QString command);
	void startCommand (int call_callback=1);
	QString getCurrentCommand ();
	void setCommandEnd();
	int findCommand (int position);
	QString getCommand ();
	void showPrompt ();
	void clearSelection ();
	void cutAction ();
	void copyAction ();
	void pasteAction ();
	void deleteAction ();
	void commandCallback ();
	void processCommand ();

	int cursor_position = 0;
	int selection_in_progress = -1;
	int selection_command_index = -1;
	int current_command_index = -1;
	int last_newline = 1;
	QList<BqCommandConsoleCommandCallback> command_callbacks;
	QString original_command;
};

#define BQ_PIXELATOR_SOURCE_TYPE_TRACE	0
#define BQ_PIXELATOR_SOURCE_TYPE_EV 	1

struct BqPixelatorPixel {
	BqPixelatorPixel () {
		memset (this, 0, sizeof(BqPixelatorPixel));
	}

	int npoints;
	float min;
	float max;
};

struct BqPixelatorSegment {
	BqPixelatorSegment () {
		memset (this, 0, BQ_Offset (BqPixelatorSegment, pixels));
	}

	double time_first_pixel;
	double amin;
	double amax;
	double calib;
	char segtype[8];
	QList<BqPixelatorPixel> pixels;
};

struct BqPixelatorTraceSegment {
	int samp0;
	int nsamp;
	int data_type;
};

struct BqPixelatorTrace {	/* Nominally corresponds to a single wfdisc row */
	Dbptr db;
	double tstart;
	double tend;
	double dt;
	long nsamp;
	double calib;			/* Converts input raw data to output
					   units. (e.g. nm/count). */
	double calper;
	char data_type[32];		/* "V", "A", etc. */
	float *data;			/* Points to data in input units before applying calib. */
	int data_size;			/* for reallocating data */
	QList<BqPixelatorTraceSegment> segments;
};

struct BqPixelatorSource {
	BqPixelatorSource (Dbptr db, int free_when_done) {
		this->type = BQ_PIXELATOR_SOURCE_TYPE_TRACE;
		this->db = db;
		this->free_when_done = free_when_done;
	}
	BqPixelatorSource (EVClient *ev_client, QString stachan, int calib_from_db, int free_when_done) {
		this->type = BQ_PIXELATOR_SOURCE_TYPE_EV;
		this->free_when_done = free_when_done;
		strcpy (this->stachan, stachan.BQ_QString2char);
		auto args = stachan.split (QRegExp("\\s"), QString::SkipEmptyParts);
		if (args.size() > 0) this->sta = args[0].BQ_QString2char;
		if (args.size() > 1) this->chan = args[1].BQ_QString2char;
		this->ev_client = ev_client;
		this->calib_from_db = calib_from_db;
	}
	~BqPixelatorSource () {
		if (type == BQ_PIXELATOR_SOURCE_TYPE_TRACE) {
			if (free_when_done) {
				trdestroy (&db);
				return;
			}
		}
		if (type == BQ_PIXELATOR_SOURCE_TYPE_EV) {
			if (free_when_done) {
				delete ev_client;
			}
		}
	}

	int type;
	Dbptr db;
	int free_when_done;
	char stachan[64];
	std::string sta;
	std::string chan;
	EVClient *ev_client;
	Table<EVArrival> ev_sc_arrivals;
	Table<EVTableRecord> ev_sc_detections;
	std::string ev_wfdisc_dbpath;
	Table<EVTableRecord> ev_sc_wfdiscs;
	int calib_from_db;
};

class BQPixelator {
public:
	BQPixelator () {}
	~BQPixelator () {}

	int setPixels (double tstart, double dtperpixel, int npixels, int force=0);
	int setDbtrace (Dbptr dbtrace, int index=-1, int free_when_done=0);
	int setEv (EVClient *ev_client, QString stachan, int calib_from_db=1, int index=-1, int free_when_done=0);
	void shiftPixels (double time_shift);
	void shiftPixels (int pixel_shift);
	void getUtilization (int &nsamp, int &npixels);

	static QList<QPair<int,int>> ungapTraceDataArray (int nsamp, float *data);

protected:
	void addDbtrace (BqPixelatorSource *source);
	void addEv (BqPixelatorSource *source);
	void addSegment (double tstart, int nsamp, double dt, double calib, char *segtype, float *data);
	void copyFrom (BQPixelator *input);
	void pixelate();
	int getTraceEv (BqPixelatorSource *source, double tstart, double tend, int read_flags, QList<BqPixelatorTrace> &traces);

	double tstart=0.0;
	double dtperpixel=0.0;
	int npixels=0;
	double amin;
	double amax;
	QString filter_string;
	double filter_tpad=0.0;
	int npixels_used=0;
	int nsamp_total=0;
	double toffset=0.0;
	QList<BqPixelatorSource *> sources;
	QList<BqPixelatorSegment> segments;
};

#define BQ_TRACE_AMPSCALE_MODE_FIXED	0
#define BQ_TRACE_AMPSCALE_MODE_AUTO		1

#define BQ_TRACE_SOURCE_TYPE_DBTRACE	0
#define BQ_TRACE_SOURCE_TYPE_SCV		1

#define BQ_TIME_NULL					(-2.e30)

struct BqTrace {
	BqTrace () {
		atop = 1.0;
		abottom = -1.0;
	};

	BQ_Color 	color_foreground;
	BQ_Color 	color_background;
	BQ_Font 	font_trace_label;
	BQ_Color 	color_trace_label;
	BQ_Font 	font_trace_amp_label;
	BQ_Color 	color_trace_amp_label;
	BQ_Color 	color_trace_filter_label;
	BQ_Font 	font_trace_filter_label;
	BQ_Color 	color_trace_select_background;
	BQ_Color 	color_trace_select_order_label;
	BQ_Font 	font_trace_select_order_label;
	double 	 	linewidth;
	double 		xpixel_factor;
	int 		clip;
	int 		visible;
	int 		label_visible;
	int 		ampscale_visible;
	int 		ampscale_mode;
	int 		show_arrivals;
	int 		show_predicted_arrivals;
	int 		show_residuals;
	int 		show_detections;
	double 		stretch_height;
	double 		ybottom;
	double 		ytop;
	double 		abottom;
	double 		atop;
	double 		gain;
	double 		units_scale;
	QString 	channel_label;
	QString 	units;
	QString     filter_label;
	QString 	filter_string;
	double      filter_tpad;
};

struct BqTraceArrival {
	BqTraceArrival() {
		trace = NULL;
		index = -1;
		index_arrivals_hash = -1;
		index_residual = -1;
		index_deltime = -1;
		order = 0;
		arid = -1;
		time = BQ_TIME_NULL;
		residual = BQ_TIME_NULL;
		deltime = BQ_TIME_NULL;
		linewidth_deltime = 0.0;
		selected = 0;
		in_clipboard = 0;
	};

	BQTrace *trace;
	EVArrival arrival;
	BQ_XYL point;
	BQ_XYL point_residual;
	BQ_XYL point_deltime;
	int index;
	int index_arrivals_hash;
	int index_residual;
	int index_deltime;
	long order;
	long arid;
	double time;
	double residual;
	double deltime;
	double linewidth_deltime;
	int selected;
	int in_clipboard;
	QString status_msg;
	BQ_Color status_text;
	BQ_Color status_fill;
	QString gc;
};

struct BqTraceArrivalMenuItem {
	QString label;
	QString command;
};

typedef QHash<QString,QList<BqTraceArrivalMenuItem>> BqTraceArrivalMenus;

typedef void (BQTrace_arrivalCallback) (BqTraceArrival &trace_arrival);

class BQTrace : public BQPixelator, public BQConfigure {
	friend class BQTraceview;
	friend class BQTraceviewCommands;
	
public:
	BQTrace(BQTraceview *traceview);

	void makeArrivalsMenus();

	static int parseArrivalsMenus (QString pfstring, BqTraceArrivalMenus &menus_out);

	BqTrace vars;
	BqTraceArrivalMenus arrivals_menus;

private:
	void setFilterString (QString filter_string);
	void setFilterTpad (double tpad);
	void getStaChanCopy (QString &sta, QString &chan, int &copy);
	void copyFrom (BQTrace *input);
	long addArrival (double time, int &index);
	int showTrace (QString show);
	int showArrivals (QString show);
	int showDetections (QString show);
	int selectTrace (QString select);

	QPainterPath makePath (int force=0);
	void makeSegmentPath (QPainterPath &path, BqPixelatorSegment segment);
	void setAmpScale (int ampscale_mode, double ybottom, double ytop, double abottom=0.0, double atop=0.0);
	void getAmpScales (double &abottom, double &atop, double &ybottom, double &ytop, double &ybottom_pixel, double &ytop_pixel);
	double amp2Viewport (double amp);
	double viewport2Amp (double yvp);
	double amp2Pixel (double amp);
	double viewport2Pixel (double yvp);
	void setUnits();
	void drawLabel (QPainter *painter);
	void drawArrivals ();
	void drawPredictedArrivals ();
	void drawArrivalsEv (BqPixelatorSource *source);
	int arrivalsCallback (BQGlyph *glyph, BQEvent *event);
	void modifyArrivalPoint (BqTraceArrival &trace_arrival, QString mode);
	void drawDetections ();
	void drawDetectionsEv (BqPixelatorSource *source);
	void drawResiduals ();
	int makeResidualPoint (BqTraceArrival &trace_arrival);
	void moveResidualPoint (BqTraceArrival &trace_arrival);
	void drawDeltimes (long arid=-1);
	int makeDeltimePoint (BqTraceArrival &trace_arrival);
	void modifyDeltimePoint (BqTraceArrival &trace_arrival);
	void moveDeltimePoint (BqTraceArrival &trace_arrival);
	void hideDeltimes ();
	void arrivalCallbackAction (int index, std::function<BQTrace_arrivalCallback> callback_proc);
	void arrivalCallbackAction (long arid, std::function<BQTrace_arrivalCallback> callback_proc);
	int arrivalEditInteraction (BQGlyph *glyph, BQEvent *event);
	int arrivalTimeEditInteraction (BQGlyph *glyph, BQEvent *event);
	int arrivalDeltimeEditInteraction (BQGlyph *glyph, BQEvent *event);
	void moveArrivals (BQGlyph *glyph, BQEvent *event);
	void editArrival (long arid, double time, double deltime, QString phase, int &index);
	long editArrival (EVArrival arrival);
	int getVisibility ();
	void getArrivals();
	void getArrivalsEv(BqPixelatorSource *source);
	void getDetections();
	void getDetectionsEv(BqPixelatorSource *source);
	long addArrivalEv(BqPixelatorSource *source, double time, int &index);
	long addArrivalEv(BqPixelatorSource *source, BqTraceArrival &arrival_item, double tshift, int &index);
	void setToffset ();
	int setScale (QString ampscale_mode_str, double abottom, double atop);

	static long addArrival (BqTraceArrival &arrival_item, double tshift, int &index);

	QString units;
	double units_scale=0.0;
	int flags_stale;
	BQPolypoint *arrivals=NULL;
	QList<BqTraceArrival> trace_arrival_list;
	int arrivals_state;
	int narrivals=0;
	double arrivals_xoffset=0.0;
	BQPolypoint *detections=NULL;
	int ndetections=0;
	BQPolypoint *predicted_arrivals=NULL;
	BQPolypoint *residuals=NULL;
	BQPolypoint *deltimes=NULL;
	BQTraceview *traceview;
	QMenu *arrivals_menu_left=NULL;
	QMenu *arrivals_menu_middle=NULL;
	QMenu *arrivals_menu_right=NULL;
	int index;
	QString channel_label;
	QString sta;
	QString chan;
	double lat=-999.0;
	double lon=-999.0;
	double elev=-999.0;
	double distance=999.0;
	int copy=0;
	QList<BQTrace *> copies;
	BQTrace *original=NULL;
	QList<long> arids;
	long sort_order=0;
	long distance_sort_order=-1;
	int select_order=0;
	int force=0;
};

struct BqTraceviewCommandsSpec {
	QString usage;
	QString help;
};

typedef int (BQ_TraceviewCommands_object_callback) (BQTraceviewCommands *tvc, QString command, QString &output);

struct BqTraceviewCommandsObjectsSpec {
	QString object_name;
	std::function<BQ_TraceviewCommands_object_callback> proc;
};

typedef void (BQ_TraceviewCommand_substitute_callback) (BQTraceview *traceview, char *string, QString &output);

struct BqTraceviewCommandSubstituteParams {
	QString var;
	std::function<BQ_TraceviewCommand_substitute_callback> proc;
};

class BQTraceviewCommands {
	friend class BQTraceview;
	friend class BQTrace;

public:
	BQTraceviewCommands (BQTraceview *traceview);

	int processCommand (QString command, QString &output);
	int processHelpObject (QString command, QString &output);
	int processEchoObject (QString command, QString &output);
	int processAliasObject (QString command, QString &output);
	int processUnaliasObject (QString command, QString &output);
	int processAliasesObject (QString command, QString &output);
	int processHotkeyObject (QString command, QString &output);
	int processUnhotkeyObject (QString command, QString &output);
	int processHotkeysObject (QString command, QString &output);
	int processMainObject (QString command, QString &output);
	int processDisplayObject (QString command, QString &output);
	int processTracesObject (QString command, QString &output);
	int processArrivalsObject (QString command, QString &output);
	int processDetectionsObject (QString command, QString &output);
	int processEventObject (QString command, QString &output);
	int processOriginObject (QString command, QString &output);
	int processHotkeys (BQEvent *event, QString &command);

private:

	void printUsage ();
	int parseObject (QString object);
	int processAlias (QString command, QStringList &commands);
	int substitute (QString &output);
	void substituteHotkeysCommand (BQEvent *event, QString &command);

	BQTraceview *traceview = NULL;
	QHash<QString, BqTraceviewCommandsObjectsSpec *> objectsHash;
	QHash<QString, BqTraceviewCommandsSpec *> commandsHash;
	QHash<QString, BqTraceviewCommandSubstituteParams *> substituteHash;

	QString object_name;
	QString object_subset;
	QString action;
	QStringList params;
	QHash<QString, QStringList> aliases;
	QHash<QString, QString> hotkeys;
};

#define BQ_TRACEVIEW_INTERACTION_MODE_INITIAL			0
#define BQ_TRACEVIEW_INTERACTION_MODE_TIME_ZOOMIN		1
#define BQ_TRACEVIEW_INTERACTION_MODE_TIME_ZOOMOUT		2
#define BQ_TRACEVIEW_INTERACTION_MODE_TIME_PAN			3
#define BQ_TRACEVIEW_INTERACTION_MODE_TRACES_PAN		4
#define BQ_TRACEVIEW_INTERACTION_MODE_TRACES_ZOOMIN		5
#define BQ_TRACEVIEW_INTERACTION_MODE_TRACES_ZOOMOUT	6
#define BQ_TRACEVIEW_INTERACTION_MODE_BOTH_PAN			7
#define BQ_TRACEVIEW_INTERACTION_MODE_BOTH_ZOOMIN		8
#define BQ_TRACEVIEW_INTERACTION_MODE_BOTH_ZOOMOUT		9
#define BQ_TRACEVIEW_INTERACTION_MODE_EDIT_ARRIVALS		10
#define BQ_TRACEVIEW_INTERACTION_MODE_ADD_ARRIVALS		11

#define BQ_TRACEVIEW_EDIT_MODE_OFF						0
#define BQ_TRACEVIEW_EDIT_MODE_ON						1

struct BqArrivalGlyph {
	BQ_Color color_outline;
	BQ_Color color_fill;
	BQ_Color color_text;
	double linewidth;
};

struct BqArrivalEdit {
	BqArrivalEdit() {
		arid = -1;
		timestamp = 0.0;
		record_number = -1;
		trace = NULL;
	}

	BqArrivalEdit(BQTrace *trace, EVArrival arrival_before, EVArrival arrival_after) {
		this->arid = arrival_after.arid;
		this->timestamp = std_now();
		this->record_number = arrival_after.recno_arrival;
		this->trace = trace;
		this->arrival_before = arrival_before;
		this->arrival_after = arrival_after;
	}

	long arid;
	double timestamp;
	long record_number;
	BQTrace *trace;
	EVArrival arrival_before;
	EVArrival arrival_after;
};

struct BqArrivalEditQueueEntry {
	BqArrivalEditQueueEntry() {
	}

	QString state;
	QList<BqArrivalEdit> edits;
};

struct BqArrivalEditQueue {
	BqArrivalEditQueue() {
		current_index = -1;
	}

	void newEntry (BQTrace *trace, EVArrival arrival_before, EVArrival arrival_after) {
		BqArrivalEditQueueEntry entry;
		BqArrivalEdit edit(trace, arrival_before, arrival_after);
		entry.edits.append (edit);
		current_index++; 
		int i=0;
		while (entries.size() > 0 && current_index < entries.size()) {
			entries.removeLast();
			i++;
		}
		if (entries.size() == 0) entry.state = base_state;
		entries.append (entry);
		current_index = entries.size()-1;
		// qDebug() << "removed" << i << "size" << entries.size();
	}

	void clear () {
		current_index = -1;
		base_state.clear();
		entries.clear();
	}

	void appendEntry (BQTrace *trace, EVArrival arrival_before, EVArrival arrival_after) {
		BqArrivalEditQueueEntry entry = entries[current_index];
		BqArrivalEdit edit(trace, arrival_before, arrival_after);
		entry.edits.append (edit);
		entries[current_index] = entry;
	}

	int findState (QString state) {
		for (int i=0; i<entries.size(); i++) {
			auto entry = entries[i];
			if (entry.state == state) return (i);
		}
		return (-1);
	}

	void setState (QString state) {
		int index = findState (state);
		if (index >= 0) entries[index].state.clear();
		if (current_index < 0) {
			base_state = state;
			return;
		}
		if (state == base_state) base_state.clear();
		entries[current_index].state = state;
	}

	int current_index;
	QString base_state;
	QList<BqArrivalEditQueueEntry> entries;
};

struct BqTraceviewDisplayParameters {
	BQ_Color 		main_display_color_background;
	BQ_Color 		main_display_color_foreground;
	BQ_Color 		main_display_color_time_grids;
	BQ_Font 		main_display_font_trace_label;
	BQ_Color 		main_display_color_trace_label;
	BQ_Font 		main_display_font_trace_amp_label;
	BQ_Color 		main_display_color_trace_amp_label;
	BQ_Font 		main_display_font_trace_filter_label;
	BQ_Color 		main_display_color_trace_filter_label;
	BQ_Color 		main_display_color_trace_select_background;
	BQ_Font 		main_display_font_trace_select_order_label;
	BQ_Color 		main_display_color_trace_select_order_label;
	double 			main_display_linewidth;
	BqArrivalGlyph	main_display_detection;
	BqArrivalGlyph	main_display_predicted_arrival;
	BqArrivalGlyph	main_display_arrival_normal;
	BqArrivalGlyph	main_display_arrival_defining;
	BqArrivalGlyph	main_display_arrival_nondefining;
	BqArrivalGlyph	main_display_arrival_magnitude;
	BqArrivalGlyph	main_display_arrival_nonassociated;
	QString 		main_display_arrival_select_expr;
	QString 		main_display_arrival_reject_expr;
	QString 		main_display_arrival_status_format;
	QList<QList<QString>> main_display_arrival_status_arguments;
	BQ_Color 		edit_display_color_background;
	BQ_Color 		edit_display_color_foreground;
	BQ_Color 		edit_display_color_time_grids;
	double 			edit_display_linewidth;
	BqArrivalGlyph	edit_display_detection;
	BqArrivalGlyph	edit_display_predicted_arrival;
	BqArrivalGlyph	edit_display_arrival_normal;
	BqArrivalGlyph	edit_display_arrival_selected;
	BqArrivalGlyph	edit_display_arrival_edit;
	BqArrivalGlyph	edit_display_arrival_defining;
	BqArrivalGlyph	edit_display_arrival_nondefining;
	BqArrivalGlyph	edit_display_arrival_defining_tag;
	BqArrivalGlyph	edit_display_arrival_nondefining_tag;
	BqArrivalGlyph	edit_display_arrival_nonassociated;
	BqArrivalGlyph	edit_display_arrival_residual;
	BQ_Color        edit_display_arrival_deltime;
	QString 		edit_display_arrival_select_expr;
	QString 		edit_display_arrival_reject_expr;
	QString 		edit_display_arrival_status_format;
	QList<QList<QString>> edit_display_arrival_status_arguments;
	BQ_Color 		color_coordinates_status;
	BQ_Color 		color_origin_status;
	BQ_Color 		color_preferred_origin_status;
};

struct BqTraceview {
	BqTraceview () {
		display_expr = "*";
		max_traces = -1;
		min_pixels_traces = -1;
	}

	BqTraceviewDisplayParameters params;
	double 		xpixel_factor;
	int 		show_arrivals;
	int 		show_predicted_arrivals;
	int 		show_residuals;
	int 		show_detections;
	int 		show_associations;
	int 		hide_noarrivals;
	int 		hide_nodetections;
	int 		clip;
	int 		antialias;
	int 		label_visible;
	int 		ampscale_visible;
	int 		ampscale_mode;
	int 		palign;
	int 		use_wheel;
	int 		max_traces;
	int 		min_pixels_traces;
	int 		preserve_time_scale_on_resize;
	int 		preserve_traces_scale_on_resize;
	int 		disable_edits;
	double 		gain;
	double 		units_scale;
	QString 	predicted_phases;
	QString 	default_phase;
	QString 	filter_string;
	QString 	units;
	double      filter_tpad;
	int 		show_crosshairs;
	int 		show_coordinates_status;
	int 		show_arrival_status;
	int 		show_origin_status;
	int 		auto_distance_sort;
	int 		auto_traces_fit;
	QString 	display_expr;
	QString 	taskbar_buttons_definitions_string;
	QString 	taskbar_buttons_string;
};

class BQTraceview : public BQViewportItem, public BQConfigure {
	friend class BQTrace;
	friend class BQTraceviewCommands;
	friend class BQViewport;
	
public:
	BQTraceview(BQViewport *master, BQCommandConsole *cc=NULL, int make_taskbar=1, BQViewport *vptaskbar=NULL);

	void createTraces (EVServer *evserver, QString stachan_accept=QString());
	EVClient *getEVClient (QString dbname=QString());
	void setCommandConsole (BQCommandConsole *cc);
	int sendCommand (QString command, QString *output=NULL, int send_to_cc=1, int nofeedback=-1);
	void setNofeedback(int nofeedback=1);
	int importCommands (QString file_name);
	int exportCommands (QString file_name);
	void refresh (int traces=0);
	void linkBQEVEventsTableview (BQEVEventsTableview *evtableview);

	void setTimeWindow (double tstart=BQ_TIME_NULL, double twin=BQ_TIME_NULL);
	void panX (double tstart);
	void zoomX (double tstart, double tend);
	void panY (double ytop);
	void panY (BQTrace *trace);
	void panY (int itrace);
	void zoomY (double yref, double factor);
	void zoomY (BQTrace *trace1, BQTrace *trace2);
	void zoomY (int first, int number);
	void zoomY2 (double ytop, double ybottom);

	int duplicateTrace (QString &trace_label);
	BQTrace *getTrace (int trace_index);
	BQTrace *getTrace (QString trace_label);
	BQTrace *getTrace (double yviewport, int *trace_index=NULL);
	int resolveTracesExpressions (QString expressions, int copyok, QList<BQTrace *> &traces);
	void arrivalEditQueueClear ();
	int arrivalEditQueueUndo ();
	int arrivalEditQueueRedo ();
	void arrivalEditQueueSetState (QString state);
	int arrivalEditQueueGotoState (QString state);
	void getEvent (QString string_index, QString &output);
	void getOrigin (QString string_index, QString &output);
	int setPreserveTimeScaleOnResize (QString answer);
	int setPreserveTracesScaleOnResize (QString answer);
	int setEditMode (QString mode);
	int setSelectedArrival (long arid, QString selected);
	int setAddArrivalsMode (QString mode);
	void clearSelectedArrivals ();
	void clearAllTags ();
	void clearSelectedArrivalsTags ();
	void setTags (long arid, QString tag, int emptyfirst=1);
	void setSelectedArrivalsTags (QString tag, int emptyfirst=1);
	void setAssociatedArrivalsTags (int emptyfirst=1);
	int setPalign (QString palign);
	int setAutoDistanceSort (QString sort);
	int setAutoTracesFit (QString ans);
	int setShowPredictedArrivals (QString show);
	int setShowTracesWithArrivals (QString show);
	int setShowTracesWithDetections (QString show);
	int setShowTracesWithArrivalsOrDetections (QString show);
	int setBatch (QString batch);
	void autoTracesFit ();
	void outputSelectedArrivals (QString &output);
	void outputTaggedArrivals (QString &output);
	void setCopyClipboard ();
	void clearCopyClipboard ();
	int pasteCopyClipboard (QString time_string, int paste_tags);
	
	static void dbpickStachans2ScvStachans (QString stachans, QString &scv_stachans);

	void makeArrivalsMenus();
	int parseArrivalStatusParameters (Pf *pf, QString &format, QList<QList<QString>> &arguments);
	int parseArrivalStatusMsg (Dbptr dbarrival, QString format, QList<QList<QString>> arguments, QString &msg);
	int parseDisplayParameters (QString pfstring);
	int parseAliases (QString pfstring);
	int parseHotkeys (QString pfstring);
	void clearAliases ();
	void clearHotkeys ();
	int setAntialias (QString antialias);
	int setClip (QString clip);

	BqTraceview vars;

	double tstart=0.0;
	double tend=0.0;
	int make_taskbar = 1;
	BQTaskbar *taskbar=NULL;
	BQViewport *vptaskbar=NULL;
	BqTraceArrivalMenus arrivals_menus;

	int verbose = 0;

protected:
	int drawViewportItem (QPainter *painter);
	void setVisibility () {}
	virtual void callback (QString pfstring) {}

	double dtperpixel=0.0;
	int npixels=0;

private:
	void init();
	int draw ();
	void drawTraceLabels (QPainter *painter);
	void drawCrosshairs ();
	void drawCoordinatesStatus ();
	void drawArrivalStatus (long arid);
	void drawOriginStatus (QString status, int preferred=0);
	void setScroll(double ytop, double ybottom, int coords_set=1);
	void scanWfdisc ();
	int appendNewTrace (QString &channel_label);
	void parseLabel (QString label, QString &root, int &copy);
	void makeTracesSorttab ();
	void clearGlyphs ();
	void displayEvent (EVEvent event, QString &output);
	int getArrivalColors (long arid, int associated, BQ_Color &outline, BQ_Color &text, BQ_Color &fill, double &linewidth);
	int getArrivalColors (BqTraceArrival trace_arrival, int associated, BQ_Color &outline, BQ_Color &text, BQ_Color &fill, double &linewidth);
	int getArrivalResidual (long arid, double &residual);
	int getArrivalShow (char *phase);
	void getDetectionColors (BQ_Color &outline, BQ_Color &text, BQ_Color &fill, double &linewidth);
	void getPredictedArrivalColors (BQ_Color &outline, BQ_Color &text, BQ_Color &fill, double &linewidth);
	void clearArrivalCoords ();
	void addArrivalCoord (long arid, BQ_XY coord);
	int processFilterString (QString filter, QString &filter_string, QString &filter_label);
	void setSelectedArrivals (BQEvent *event, int clear=1);
	void print_arrivals_hash ();
	int editSelectedArrivals (double time, double deltime, QString phase);
	void refreshPost (int wfdiscs=0);
	void setAmpScale (int trace_index, int ampscale_mode, double ybottom, double ytop, double abottom=0.0, double atop=0.0);
	void setAmpScale (QString trace_label, int ampscale_mode, double ybottom, double ytop, double abottom=0.0, double atop=0.0);
	int setDisplayScales ();
	int createTraceEv (QString &trace_label, EVClient *ev_client, QString stachan, int calib_from_db=1, int index=-1, int free_when_done=0);
	void displayOrigin (EVOrigin origin, QString &output, int update_display);
	void distanceSort ();
	int layoutTraces ();
	void clearTags (long arid);
	void processEVCallback (BQEVCallbackEvent *event);
	int resolveTracesExpression (QString expression, 
			QString &sta_expression, QString &chan_expression, QString &copy_expression);
	int resolveTrace (QString sta_expression, QString chan_expression, QString copy_expression, int copyok, 
								BQTrace *trace);
	int orderTracesExpressions (QString expressions, QList<BQTrace *> &traces_output);
	void setAmpView (double ybottom_view, double ytop_view);
	void setXpixelFactor (double factor);
	void deleteTrace (int trace_index);
	void deleteTrace (QString trace_label);
	void displayTraces (QString traces_expr);
	void showhideTraces (QString traces_expr, int show);
	void showhideArrivals (QString traces_expr, int show);
	void showhideDetections (QString traces_expr, int show);
	int stretchTraces (QString traces_expr, double factor);
	int colorTraces (QString traces_expr, BQ_Color color);
	int filterTraces (QString traces_expr, QString filter);
	int unitsTraces (QString traces_expr, QString units);
	int scaleTraces (QString traces_expr, int ampscale_mode, double abottom = 0.0, double atop = 0.0);
	void clearArrivalsHash();
	QHash<long,QList<BqTraceArrival>> getSelectedArrivals();
	QStringList getArrivalTags(long arid);
	QStringList getArrivalTags(BqTraceArrival trace_arrival);
	bool arrivalIsSelected (long arid=-1);
	bool arrivalIsSelected (BqTraceArrival trace_arrival);
	bool arrivalIsInClipboard (long arid=-1);
	void setSelectedArrival (long arid);
	void clearSelectedArrival (long arid);
	QList<QList<BqTraceArrival>> getClipboard ();

	static void ccCommandCallback (QString command, BQ_ClientData client_data);
	static void eventCallback (void *item, BQEvent *event, BQ_ClientData client_data);
	static void resizeCallback (void *item, QResizeEvent *event, BQ_ClientData client_data);
	static int dbpickMatchString (char *spec, char *string);
	static int dbpickMatchStachans (char *stachans, char *sta, char *chan, int copy);

	int batch = 0;
	QString stachan_accept;
	QHash<QString,int> stachan_hash;
	QPainter *painter=NULL;
	int nofeedback = 0;
	int skip_render = 0;
	int interaction_mode = BQ_TRACEVIEW_INTERACTION_MODE_INITIAL;
	int edit_mode = BQ_TRACEVIEW_EDIT_MODE_OFF;
	QHash<long,double> toffset;
	int rubberband_xleft;
	double rubberband_xstart;
	int rubberband_ybottom;
	int rubberband_ytop;
	double rubberband_ystart;
	double ybottom_view_hold;
	double ytop_view_hold;
	double tstart_sources=BQ_TIME_NULL;
	double resize_xfactor = 1.0;
	double resize_yfactor = 1.0;
	double minimum_height = 0.0;
	QImage *image=NULL;
	QPainter *imagePainter=NULL;
	int image_w = 0;
	int image_h = 0;
	int image_xshift = 0;
	int image_yshift = 0;
	QScrollBar *scrollbar=NULL;
	QRubberBand *rubberband=NULL;
	double x_crosshairs = BQ_FLOAT_NULL;
	double y_crosshairs = BQ_FLOAT_NULL;
	BQAxes *axes;
	BQPolypoint *crosshairs1;
	BQPolypoint *crosshairs2;
	BQText *coordinates_text;
	BQText *arrival_text;
	BQText *origin_text;
	QMenu *traces_menu=NULL;
	int show_crosshairs = 1;
	int show_coordinates_status = 1;
	int show_arrival_status = 1;
	int show_origin_status = 1;
	int cursor_in_arrival = 0;
	BQCommandConsole *cc=NULL;
	BQTrace *last_trace_selected = NULL;
	QList<BQTrace *> traces;
	QList<int> traces_display_list;
	QList<int> traces_sorttab;
	QHash<QString,BQTrace *> channel_label_hash;
	QHash<long,QList<BqTraceArrival>> arrivals_hash;
	QHash<long,QList<BqTraceArrival>> arrivals_hash_existing;
	QHash<long,QStringList> arrivals_tags_hash;
	QHash<long,QStringList> arrivals_tags_hash_existing;
	int arrival_undo_position = -1;
	int temp_selection_arid = -1;
	QHash<long,BQ_XY> arrival_coords;
	QHash<QString,EVClient *> evs;
	EVClient *ev_client=NULL;
	int event_index=-1;
	int origin_index=-1;
	EVEvent current_event;
	double otime=-999.0;
	double olat=-999.0;
	double olon=-999.0;
	double odepth=-999.0;
	QHash<long,EVAssociation> current_associations;
	QHash<long,EVStamag> current_stamags;
	BQTraceviewCommands *commands=NULL;
	BqArrivalEditQueue edit_queue;
	BQEVEventsTableview *evtableview=NULL;

};

int BQ_TraceviewCommand_process (BQTraceview *traceview, QString command, QString &output);

struct BQ_GraphicsServer;

typedef void (BQ_GraphicsServer_run_thread_callback) (BQ_GraphicsServer *server, int argc, char **argv);

#define BQ_GRAPHICSSERVER_TYPE_VIEWPORT		0
#define BQ_GRAPHICSSERVER_TYPE_AXES			1
#define BQ_GRAPHICSSERVER_TYPE_TEXT			2
#define BQ_GRAPHICSSERVER_TYPE_POLYLINE		3
#define BQ_GRAPHICSSERVER_TYPE_POLYPOINT	4
#define BQ_GRAPHICSSERVER_TYPE_MAP			5
#define BQ_GRAPHICSSERVER_TYPE_MAPEVENTS	6
#define BQ_GRAPHICSSERVER_TYPE_TEXTPOPUP	7
#define BQ_GRAPHICSSERVER_TYPE_TASKBAR		8
#define BQ_GRAPHICSSERVER_TYPE_FRAME		9
#define BQ_GRAPHICSSERVER_TYPE_GRIDLAYOUT	10
#define BQ_GRAPHICSSERVER_TYPE_FORMLAYOUT	11
#define BQ_GRAPHICSSERVER_TYPE_SPLITTER		12
#define BQ_GRAPHICSSERVER_TYPE_DOCKWIDGET	13
#define BQ_GRAPHICSSERVER_TYPE_BUTTON		14
#define BQ_GRAPHICSSERVER_TYPE_MENUBAR		15
#define BQ_GRAPHICSSERVER_TYPE_MENU			16
#define BQ_GRAPHICSSERVER_TYPE_LINEEDIT		17
#define BQ_GRAPHICSSERVER_TYPE_ACTION		18
#define BQ_GRAPHICSSERVER_TYPE_FILEDIALOG	19
#define BQ_GRAPHICSSERVER_TYPE_SPREADSHEET	20
#define BQ_GRAPHICSSERVER_TYPE_MAINWINDOW	21
#define BQ_GRAPHICSSERVER_TYPE_TIMER		22
#define BQ_GRAPHICSSERVER_TYPE_SCROLLAREA	23
#define BQ_GRAPHICSSERVER_TYPE_DATAGRID		24
#define BQ_GRAPHICSSERVER_TYPE_CHECKBOX		25
#define BQ_GRAPHICSSERVER_TYPE_COMBOBOX		26
#define BQ_GRAPHICSSERVER_TYPE_DIALOG		27
#define BQ_GRAPHICSSERVER_TYPE_LABEL		28
#define BQ_GRAPHICSSERVER_TYPE_PFTREEVIEW	29
#define BQ_GRAPHICSSERVER_TYPE_TABLEVIEW	30
#define BQ_GRAPHICSSERVER_TYPE_DBTABLEVIEW	31
#define BQ_GRAPHICSSERVER_TYPE_EVTABLEVIEW	32
#define BQ_GRAPHICSSERVER_TYPE_EVEVENTSTABLEVIEW	33
#define BQ_GRAPHICSSERVER_TYPE_EVORIGINSTABLEVIEW	34
#define BQ_GRAPHICSSERVER_TYPE_COMMANDCONSOLE 35
#define BQ_GRAPHICSSERVER_TYPE_TRACEVIEW	36

struct BQ_GraphicsServer_type {
	BQ_GraphicsServer_type () {
		type = 0;
		widget = NULL;
	}
	BQ_GraphicsServer_type (int t, void *w) {
		type = t;
		widget = w;
	}

	int type;
	void *widget;
};

#define BQ_GRAPHICSSERVER_STATE_INACTIVE	0
#define BQ_GRAPHICSSERVER_STATE_RUNNING		1
#define BQ_GRAPHICSSERVER_STATE_QUIT		2
#define BQ_GRAPHICSSERVER_STATE_DONE		3

struct BQ_GraphicsServer;

typedef void (BQ_closeCallback) (void *widget, QCloseEvent *event, BQ_ClientData client_data);

struct BqCloseCallback {
	BqCloseCallback () {
		memset (this, 0, sizeof(BqCloseCallback));
	}

	std::function<BQ_closeCallback> callback_proc;
	void *widget;
	BQ_ClientData client_data;
};

class BQMainWindow : public QMainWindow {
public:
	BQMainWindow (BQ_GraphicsServer *server, QWidget *parent, int close_window=0);
	BQMainWindow (QWidget *parent, int close_window=0);
	QFrame *getToplevel();
	void addViewport (BQViewport *);
	void removeViewport (BQViewport *);
	void setCloseCallback (std::function<BQ_closeCallback> closeCallback, BQ_ClientData client_data=NULL);
	void removeCloseCallback (std::function<BQ_closeCallback> closeCallback);
	void print (QWidget *widget, int render_to_image=0, QPageLayout::Orientation *orientation=NULL, QString *fname=NULL);

	QString tl_handle;
	QFrame *tl;
	QPrinter printer;
	QList<BQViewport *> viewports;

protected:
	void paintEvent (QPaintEvent *event);
	void closeEvent (QCloseEvent *event);
private:
	BQ_GraphicsServer *server=NULL;
	QWidget *parent;
	int first;
	int close_window;
	QList<BqCloseCallback> close_callbacks;
};

class BQTopLevel : public QFrame {
public:
	BQTopLevel (BQ_GraphicsServer *server, QWidget *parent, int close_window=0);
	BQTopLevel (QWidget *parent=NULL, int close_window=0);
	void addViewport (BQViewport *);
	void removeViewport (BQViewport *);
	void setCloseCallback (std::function<BQ_closeCallback> closeCallback, BQ_ClientData client_data=NULL);
	void removeCloseCallback (std::function<BQ_closeCallback> closeCallback);

	QFrame *getToplevel();
	QPrinter printer;
	QList<BQViewport *> viewports;

protected:
	void paintEvent (QPaintEvent *event);
	void closeEvent (QCloseEvent *event);
private:
	BQ_GraphicsServer *server=NULL;
	QWidget *parent;
	int first;
	int close_window;
	QList<BqCloseCallback> close_callbacks;
};

typedef void (BQPfOptionsDialog_buttonCallback) ();
typedef void (BQPfOptionsDialog_checkboxCallback) (int state);

class BQoptions;

class BQPfOptionsDialog : public BQTopLevel {
public:
    BQPfOptionsDialog(QString pffile, int show_controls=1, int options_params=0);
    BQPfOptionsDialog(BUPf *bupf, int show_controls=1, int options_params=0);
    ~BQPfOptionsDialog();

    void setStatusMsg (QString msg);

    BQPfTreeview *tv=NULL;
    BQPfOptionsDialog *parent=NULL;

private:
    class OptionsBUPf : public BUPf {
    public:
        OptionsBUPf () : BUPf ("optionsdialog") {}

        BQPfOptionsDialog *od;

    protected:
        void valueChangedCallback (std::string key, int index, std::string from, std::string &to);
    };

    void init (QString pffile, BUPf *bupf);
    void addButton (QFrame *frame, QGridLayout *layout, int &column, QString text, 
                    std::function<BQPfOptionsDialog_buttonCallback> buttonCallback);
    void addCheckbox (QFrame *frame, QGridLayout *layout, int &column, 
                    QString text, Qt::CheckState state, int column_stretch,
                    std::function<BQPfOptionsDialog_checkboxCallback> checkboxCallback);
    void makeFrame();

    int show_controls;
    int options_params;
    int free_bupf = 1;
    QLabel *msg;

    QGridLayout *layout=NULL;
    BUPf *bupf=NULL;
    BQoptions *options=NULL;
};

struct BQ_GraphicsServer {
	BQ_GraphicsServer () {
		argc = 0;
		argv = NULL;
		qapp = NULL;
		first_toplevel = 1;
		send_mtfifo = NULL;
		receive_mtfifo = NULL;
		handle_instance = 0;
		run_thread_callback = NULL;
		run_thread = 0;
		server_thread = 0;
		setup = 0;
		state = BQ_GRAPHICSSERVER_STATE_INACTIVE;
		QPrinter printer ( QPrinter::HighResolution );
	}
	~BQ_GraphicsServer () {
		if (send_mtfifo) pmtfifo_destroy (send_mtfifo, NULL);
		if (receive_mtfifo) pmtfifo_destroy (receive_mtfifo, NULL);
		if (qapp) delete qapp;
	}

	int setup;
	int argc;
	char **argv;
	QApplication *qapp;
	QString toplevel_name;
	int first_toplevel;
	Pmtfifo *send_mtfifo;
	Pmtfifo *receive_mtfifo;
	QHash<QString, BQ_GraphicsServer_type> handles;
	int handle_instance;
	BQ_GraphicsServer_run_thread_callback *run_thread_callback;
	pthread_t run_thread;
	pthread_t server_thread;
	int state;
	QPrinter printer;
	QList<BQViewport *> viewports;
};

struct BQ_GraphicsServer_config_arg {
	BQ_GraphicsServer_config_arg (QString key, QString value) {
		this->key = key;
		this->value = value;
	}
	BQ_GraphicsServer_config_arg (char *key, char *value) {
		this->key = QString(key);
		this->value = QString(value);
	}

	QString key;
	QString value;
};

struct BQ_GraphicsServer_command {
	QString command;
	QString handle;
	QStringList args;
	QList<BQ_GraphicsServer_config_arg> config_args;
	QList<void *> extra_args;
};

#define BQ_GRAPHICSSERVER_RESPONSE_OK		0
#define BQ_GRAPHICSSERVER_RESPONSE_ERROR	1

struct BQ_GraphicsServer_response {
	BQ_GraphicsServer_response () {
		this->response = BQ_GRAPHICSSERVER_RESPONSE_OK;
	}
	BQ_GraphicsServer_response (int response, QString handle) {
		this->response = response;
		this->handle = handle;
	}

	int response;
	QString handle;
	QStringList return_params;
};

template <typename V> 
void BQ_showKeys(const char *pre, QHash<QString, V> *hash) {
	if (hash) return;
	QList<QString> keys = hash->keys();
	printf ("%s: keys: ", pre);
	foreach (QString key, keys) {
		printf ("%s ", key.toLocal8Bit().data());
	}
	printf ("\n");
}

extern void BQ_debug (const char *format, ...);
extern char *BQ_qevent_type_str (QEvent *event);
extern int BQ_Strcmp (QString str1, QString str2);
extern void BQ_edp2latlon (double latr, double lonr, double delx, double dely, double *lat, double *lon);
extern void BQ_edp2latlonr (double lat, double lon, double delx, double dely, double *latr, double *lonr);
extern void BQ_globe2latlon (double latr, double lonr, double delx, double dely, double *lat, double *lon);
extern void BQ_globe2latlonr (double lat, double lon, double delx, double dely, double *latr, double *lonr);
extern void BQ_latlon2edp (double latr, double lonr, double lat, double lon, double *delx, double *dely);
extern void BQ_latlon2edp2 (double latr, double lonr, double lat, double lon, double *delx, double *dely, double *north);
extern int BQ_latlon2globe (double latr, double lonr, double lat, double lon, double *delx, double *dely);
extern int BQ_latlon2globe2 (double latr, double lonr, double lat, double lon, double *delx, double *dely, double *north, double *az=NULL, double *dip=NULL);
extern void BQ_delbox2llbox (double xlatc, double xlonc, double xdelmn, double xdelmx, double ydelmn, double ydelmx, int globe,
			double *xlatmn, double *xlatmx, double *xlonmn, double *xlonmx);
extern int BQ_setup_line (QPainter *painter, BQ_Color *color, int linewidth, int linestyle, int joinstyle, int capstyle=-1);
extern int BQ_setup_line (QPainter *painter, BQ_Color *color, double linewidth, int linestyle, int joinstyle, int capstyle=-1);
extern int BQ_setup_fill (QPainter *painter, BQ_Color *color);
extern int BQ_setup_text (QPainter *painter, BQ_Color *color, BQ_Font *font);
extern int BQ_get_metrics (BQ_Font *font, char *text, int *w, int *h);
extern int BQ_get_fontmetrics (BQ_Font *font, char *text, int *str_w, int *str_h, int *ascent, int *descent);
extern int BQ_get_nice_fvals (double start, double end, double width, double minbigwidth,
			  double *start_big, int *n_big, double *incr_big,
			  double *start_sml, int *n_sml, double *incr_sml);
extern int BQ_nice_number (double number, char *format, char *string);
extern void BQ_InstallMessageHandler();
extern void BQ_MessageOutput(QtMsgType type, const QMessageLogContext &context, const QString &msg);
extern int BQ_pfcompile (char *pfstring, Pf **pf);
extern char *BQ_make_pfstring (QString pfstring, QString input_key, QString output_key);
extern int BQ_render_print (QPrinter *printer, QList<BQViewport *> viewports, QWidget *widget, int render_to_image, QPageLayout::Orientation *orientation=NULL, QString *fname=NULL);
extern BQ_GraphicsServer *BQ_GraphicsServer_create (int argc, char **argv, char *toplevel_name, BQ_GraphicsServer_run_thread_callback *run_thread_callback, int runqt=1);
extern void *BQ_GraphicsServer_get_item (BQ_GraphicsServer *server, QString handle, int *type=NULL, int idelete=0); 
extern int BQ_GraphicsServer_send (BQ_GraphicsServer *server, BQ_GraphicsServer_response **response, QString command, QString handle, QStringList *args, ...);
extern int BQ_GraphicsServer_send2 (BQ_GraphicsServer *server, BQ_GraphicsServer_response **response, QString command, QString handle, QStringList *args, QStringList *config_args);
extern int BQ_GraphicsServer_send_extra (BQ_GraphicsServer *server, BQ_GraphicsServer_response **response, QString command, QString handle, QList<void *> extra_args);
extern void BQ_GraphicsServer_release (BQ_GraphicsServer *server);
extern BQ_GraphicsServer *BQ_GraphicsServer_get_global ();

#endif
