
#ifndef _sysdata_
#define _sysdata_

#include <sys/time.h>

#include "stock.h"

typedef struct Pidstat {
    int             pid;		/* process id */
    int             ppid;		/* parent process id */
    int	            uid;		/* user id */
    unsigned long   size,		/* total size in kbytes */
                    rss;	        /* resident set size in kbytes */
    char            state[16];		/* run state -- ie, sleep, cpu, run */
    double	    started ;		/* process start time */
    double          pcpu ;		/* short term average % of all cpu's */
    double          pmem ;		/* short term average % of all mem */
    struct timeval  measured;		/* last measurement time */
    double 	    cputime;		/* total cpu time in seconds */
    double          ccputime;		/* total children cpu time in seconds */
    char	    user[32]; 		/* *Darwin only* user name */
    char 	    command[1024];	/* *Darwin only* command line */
} Pidstat ;

typedef struct CpuUsage { 
    float user ; 
    float kernel ; 
    float iowait ; 
    float idle ; 
    float swap ;
} CpuUsage ;

typedef struct SysMounted { 
    char mountpt[FILENAME_MAX] ; 
    char host[96] ;
    char source[FILENAME_MAX] ; 
    char fstype[32] ; 
} SysMounted ; 

typedef struct Pidrtsys {
	char	host[MAXHOSTNAMELEN];
	char	username[STRSZ];
	char	rtpath[FILENAME_MAX];
	char	antelope_version[STRSZ];
	int 	running;
	int 	pid;
} Pidrtsys;

#ifdef  __cplusplus
extern "C" {
#endif

extern char * pidexecname ( int pid );
extern char * pidpwd ( int pid );
extern char * pidcmdline ( int pid );
extern long sysboottime ( void );
extern int pidinfo ( int pid, Pidstat * process );
extern int pidrtsys( Pidrtsys **systems );

extern int sysncpu () ;
extern void syscpu ( int *ncpu, CpuUsage **cpu );
extern void sysloadavg ( double *avg, int *nproc );
extern void sysmem ( double *vm, double *physical, double *used );

extern int sysmounted ( Tbl **sysmnt );
extern int sysproc ( Tbl **procsp );

extern double cpu_utilization () ;

void syspaging ( int interval, double *rate_pagein, double *rate_pageout ) ;

#ifdef  __cplusplus
}
#endif
#endif
